/**
 * @file misclib.c
 * @brief Support Functions Library
 */

/*
 * Support Functions Library
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "misclib.h"
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <netdb.h>

/**
 * @brief accept() wrapping 함수
 * @param fd - Socket descriptor
 * @param sa - 소켓 구조체
 * @param salenptr - \a sa 의 length
 * @return
 *  성공 시 socket descriptor\n
 *  실패 시 -1
 */
int
Accept(int fd, struct sockaddr *sa, socklen_t *salenptr)
{
    int n;

again:
    if((n = accept(fd, sa, salenptr)) < 0) {
#ifdef  EPROTO
	if(errno == EPROTO || 
		errno == ECONNABORTED || 
		errno == EINTR) {
#else
	    if(errno == ECONNABORTED || 
		    errno == EINTR) {
#endif
		goto again;
	    }
	    else {
		return -1;
	    }
	}

    return n;
}


#ifdef HAVE_GETADDRINFO
/**
 * @brief 소켓 연결 
 * @param host - 호스트네임 또는 IP
 * @param port - 접속 포트
 * @param timeout - 최대 접속대기시간
 * @return
 *  성공 시 socket descriptor,\n
 *  실패 시 -1
 */
int 
Connect(const char *host, const char *port, int timeout)
{
    int	sockfd, n;
    int result;
    struct addrinfo hints, *res, *ressave;

    ASSERT(host != NULL && port != NULL);

    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    if((n = getaddrinfo(host, port, &hints, &res)) != 0)
	return -1;
    ressave = res;

    do {
	sockfd = socket(res->ai_family, 
		res->ai_socktype, 
		res->ai_protocol);
	/* ignore error */
	if(sockfd < 0) {
	    continue;	
	}

	alarm(timeout);
	result = connect(sockfd, res->ai_addr, res->ai_addrlen);
	if(result == 0) {
	    alarm(0);
	    break;		
	}
	alarm(0);
	if(result < 0 && errno == EINTR) {
	    close(sockfd);
	    freeaddrinfo(ressave);
	    return -1;
	}

	close(sockfd);		
    } while((res = res->ai_next) != NULL);

    if(res == NULL) {
	freeaddrinfo(ressave);
	return -1;
    }

    freeaddrinfo(ressave);

    return sockfd;
}
#else
/**
 * @brief 소켓 연결 
 * @param host - 호스트네임 또는 IP
 * @param port - 접속 포트
 * @param timeout - 최대 접속대기시간
 * @return
 *  성공 시 socket descriptor,\n
 *  실패 시 -1
 */
int 
Connect(const char *host, const char *port, int timeout)
{
    int fd;
    struct sockaddr_in servaddr;
    struct hostent *hptr = NULL;

    /* gethostbyname set h_errno if it failed */
    hptr = gethostbyname(host);
    if (!hptr)
	return -1;

    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0)
	return -1;

    memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_port = htons(atoi(port));
    memcpy(&servaddr.sin_addr, *hptr->h_addr_list, sizeof(servaddr.sin_addr));

    if (connect(fd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0)
	return -1;

    return fd;
}
#endif


/**
 * @brief 데몬화 
 * @param 없음
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
int
daemonize(void)
{
    int fd;
    pid_t pid;

    /* 1차 fork */
    if((pid = fork()) < 0)
	return -1;
    else if(pid != 0) {
	exit(EXIT_SUCCESS);
    }

    /* 2차 fork */
    if((pid = fork()) < 0) {
	return -1;
    }
    else if(pid != 0) {
	exit(EXIT_SUCCESS);
    }

    usleep(1000);
    setsid();
    chdir("/");
    umask(0);
    close(0);
    close(1);
    close(2);
    fd = open("/dev/null", O_RDONLY);
    if (fd < 0)
    {
	return -1;
    }
    dup2(fd, 1);
    dup2(fd, 2);

    return 0;
}


/**
 * @brief 특정 포트로 소켓생성 후 리슨하는 함수 
 * @param port - 리슨할 포트 번호
 * @return
 *  성공 시 socket descriptor,\n
 *  실패 시 -1
 */
int
socket_listen(int port)
{
    int fd;
    struct sockaddr_in servaddr;
    const int on = 1;

    if((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	return -1;
    }

    memset(&servaddr, 0, sizeof(servaddr));
    servaddr.sin_family = AF_INET;
    servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    servaddr.sin_port = htons(port);

    if(setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0) {
	return -1;
    }
    if(bind(fd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0) {
	return -1;
    }
    if(listen(fd, 1024) < 0) {
	return -1;
    }

    return fd;
}


/**
 * @brief Socket에서 \a n 만큼 read
 * @param fd - Socket descriptor
 * @param vptr - 수신한 데이터를 저장할 버퍼
 * @param n - 수신할 데이터의 사이즈
 * @return
 *  성공 시 수신한 데이터 사이즈,\n
 *  실패 시 -1
 */
ssize_t
readn(int fd, void *vptr, size_t n)
{
    size_t nleft;
    ssize_t nread;
    char *ptr = NULL;

    ptr = vptr;
    nleft = n;
    while(nleft > 0) {
	if((nread = read(fd, ptr, nleft)) < 0) {
	    if(errno == EINTR) {
		nread = 0;
	    }
	    else {
		return -1;
	    }
	}
	else if(nread == 0) {
	    break;
	}

	nleft -= nread;
	ptr += nread;
    }

    return (n - nleft);
}


/**
 * @brief Socket에 \a len 만큼 write
 * @param connfd - Socket descriptor
 * @param buf - write할 데이터
 * @param len - \a buf 의 길이
 * @return
 *  성공 시 write한 바이트수,\n
 *  실패 시 -1
 */
ssize_t
writen(int connfd, const char *buf, size_t len)
{
    int nwrite, twrite = 0, olen = len;
    const char *ptr = buf;

    ASSERT(buf != NULL);

    while(len > 0) {
again:
	if((nwrite = write(connfd, ptr, len)) < 0) {
	    if(errno == EINTR) {
		goto again;
	    }
	    else {
		return -1;
	    }
	}

	len -= nwrite;
	ptr += nwrite;
	twrite += nwrite;
    }

    return (twrite == olen) ? twrite : -1;
}

