/**
 * @file configure.c
 * @brief 설정파일 관리
 */

/*
 * 설정파일 관리
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <dirent.h>
#include <pthread.h>
#include <sys/stat.h>

#include "configure.h"
#include "support.h"
#include "../lib/slib.h"

/**
 * @brief 파라메터 존재여부 확인
 * @param list - 체크할 파라메터 리스트
 * @param err - 에러 발생 시 에러메세지를 저장할 버퍼
 * @param len - \a err 의 길이
 * @return: 
 *  모든 파라메터가 존재하면 0,\n 
 *  존재하지 않는 파라메터가 있으면 -1
 */
static int
is_exist_parm(char **list, char *err, size_t len)
{
    int i;

    ASSERT(list && err && len >= 0);

    for (i = 0; list[i]; i++)
    {
	if (config_check_parameter(list[i]) != 1 ||
		config_get_value(list[i]) == NULL)
	{
	    snprintf(err, len, "'%s' 파라메터가 존재하지 않음", 
		    list[i]);
	    return -1;
	}
    }

    return 0;
}

/**
 * @brief 설정파일 유효성 체크
 * @param err - 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
int
config_check(void **err)
{
    DIR *dir = NULL;
    char err_msg[BUFSIZ];
    /* 공통필수 파라메터 */
    char *param[] = {
	"ServerAlias", "Type", "ServerPort", "LogFile", "LogLevel", "DataDir", 
	NULL
    };
    /* 마스터 전용 */
    char *master[] = {
	"ControlPort", "ScanDelay", "Sendmail", "AdminEmail", NULL
    };
    /* 에이전트 전용 */
    char *agent[] = {
	"ServerIP", "ReconnectDelay", "CharSet", NULL
    };

    ASSERT(err);

    /* 공통 파라메터 확인 */
    if (is_exist_parm(param, err_msg, sizeof(err_msg)))
    {
	*err = strdup(err_msg);
	return -1;
    }

    /* 데이터 저장경로가 유효한지 확인 */
    if ((dir = opendir(config_get_value("DataDir"))) == NULL)
    {
	if (mkdir(config_get_value("DataDir"), S_IRWXU) == -1)
	{
	    *err = dup_str("can't open data directory (%s): %s", 
		    config_get_value("DataDir"), strerror(errno));
	    return -1;
	}
    }
    else
	closedir(dir);

    if (strcasecmp(config_get_value("Type"), "Master") == 0)
    {
	/* 마스터 전용 파라메터 확인 */
	if (is_exist_parm(master, err_msg, sizeof(err_msg)))
	{
	    *err = strdup(err_msg);
	    return -1;
	}
    }
    else if (strcasecmp(config_get_value("Type"), "Agent") == 0)
    {
	/* 에이전트 전용 파라메터 확인 */
	if (is_exist_parm(agent, err_msg, sizeof(err_msg)))
	{
	    *err = strdup(err_msg);
	    return -1;
	}
    }
    else
    {
	*err = dup_str("알 수 없는 서버 타입: %s", config_get_value("Type"));
	return -1;
    }

    return 0;
}

/**
 * @brief 시스템 설정에서 파라메터 값 확인
 * @param param - 가져올 파라메터명
 * @return
 *  파라메터가 있는 경우 해당 파라메터의 값 (호출자 사용 후 free),\n
 *  파라메터가 없는 경우 NULL
 */
char *
get_config(const char *param)
{
    char *value = NULL;

    value = config_get_value(param);
    return value ? strdup(value) : NULL;
}

/**
 * @brief 시스템 설정 모듈 초기화
 * @param config_file - 설정파일명
 * @return 없음
 */
void
init_configure(const char *config_file)
{
    void *err_msg = NULL;

    if (config_read(config_file, &config_check, &err_msg))
    {
	fprintf(stderr, "Failed to load %s configuration file: %s\n", 
		config_file, (char *) (err_msg ? err_msg : strerror(errno)));
	exit(EXIT_FAILURE);
    }
}
