/**
 * @file core.c
 * @brief Core (Frontend)
 */

/*
 * Core (Frontend)
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <pxe.h>
#include "core.h"
#include "core_backup.h"
#include "core_restore.h"
#include "core_delete.h"
#include "dbs.h"
#include "types.h"
#include "support.h"
#include "../lib/slib.h"

/**
 * @brief 요청작업 분석 / 해당 작업 실행
 * @param pxe - PXE 핸들러 포인터
 * @param bk - 백업 쓰레드 헤더 포인터
 * @param data - 수신한 백업모듈 패킷 (백업 쓰레드 헤더 제외)
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  작업 완료 시 R_FINISH,\n
 *  개별작업 성공 시 R_SUCCESS,\n
 *  에러 발생 시 R_ERROR
 */
static void
run_cmd(PXE *pxe, BK_Header *bk, void *data, void **msg)
{
    int result;

    ASSERT(pxe && bk && data && msg);

    if (bk->type == BACKUP)
	result = backup_cmd(pxe, bk->regno, data, NULL, msg);
    else if (bk->type == RESTORE)
	result = restore_cmd(pxe, bk->regno, data, NULL, msg);
    else if (bk->type == DELETE)
	result = delete_cmd(pxe, bk->regno, data, NULL, msg);
    else /* Never reached */
    {
	Log(ERROR, "[ASSERT] 유효하지 않은 작업 타입 (BACKUP, RESTORE, DELETE "
		"중 하나만 가능)");
	abort();
    }

    /* 내부에러 발생 */
    if(result < 0 && msg)
	Log(ERROR, "내부 에러 발생: 등록번호 - %" PRIu64 ", 에러 메세지 - %s", 
		bk->regno, msg);
}

/**
 * @brief 네트웍으로 수신한 패킷에서 BK 헤더 파싱
 * @param bk - 파싱된 BK 헤더 정보를 저장할 구조체 포인터
 * @param packet - 네트웍으로 수신한 패킷 (바이트 스트림 형태)
 * @return 없음
 */
void
parse_bk_header(BK_Header *bk, void *packet)
{
    uint32_t *type = NULL;
    REGNO *regno = NULL;

    type = (uint32_t *) packet;
    regno = (REGNO *) ((char *) type + sizeof(uint32_t));

    bk->type = ntohl(*type);
    bk->regno = ntohll(*regno);
}

/**
 * @brief 요청작업 파싱 후 해당 작업 실행
 * @param pxe - PXE 핸들러 포인터
 * @param packet - 수신한 패킷 포인터
 * @return 없음
 */
static void
switchboard(PXE *pxe, void *packet)
{
    BK_Header bk;
    void *data = NULL;
    void *msg = NULL;

    ASSERT(pxe && packet);

    parse_bk_header(&bk, packet);
    data = (char *) packet + BK_HEADER_SIZE;
    run_cmd(pxe, &bk, data, &msg);
    FREE(msg);
}

/**
 * @brief DBS Core Main
 * @param arg - PXE 쓰레드 로딩시 넘겨받는 파라메터
 * @return NULL 포인터
 */
void *
core(void *arg)
{
    PXE *pxe = NULL;
    PXELength cnt;
    void *packet = NULL;

    /* PXE 핸들러 생성 */
    pxe = pxe_open(arg);
    if (!pxe)
	exit(EXIT_FAILURE);

    /* Main loop */
    while (1)
    {
	cnt = pxe_read(pxe, &packet);
	if (cnt > 0)
	    switchboard(pxe, packet);

	pxe_free(packet);
    }

    /* Never reached */
    pxe_close(pxe);
    free(arg);

    return NULL;
}

/**
 * @brief BK 쓰레드 헤더 생성 (네트웍 전송을 위해 바이트 스트림 형태로 생성)
 * @param ptr - BK 쓰레드를 저장할 메모리 포인터
 * @param regno - 등록번호
 * @param type - 작업 타입(백업, 복구, 삭제)
 * @return 없음
 */
void
create_bk_header(void *ptr, REGNO regno, int type)
{
    uint32_t *type_ptr = (uint32_t *) ptr;
    REGNO *regno_ptr = (REGNO *) ((char *) type_ptr + sizeof(uint32_t));

    *type_ptr = htonl(type);
    *regno_ptr = htonll(regno);
}

/**
 * @brief PXE destination host 설정
 * @param pxe - PXE 핸들러 포인터
 * @param ip - 대상 호스트 IP
 * @return 
 *  성공 시 0,\n
 *  실패 시 -1
 */
int
set_dst_host(PXE *pxe, unsigned char *ip)
{
    char ip_str[16];

    ASSERT(pxe && ip);

    convert_ip_to_str(ip, ip_str, sizeof(ip_str));
    return pxe_set_destination(pxe, ip_str, BK_THREAD);
}
