/**
 * @file core_backup.c
 * @brief Backup Core (Backend)
 */

/*
 * Backup Core (Backend)
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <inttypes.h>
#include <sys/stat.h>

#include <pxe.h>
#include "core_backup.h"
#include "core.h"
#ifdef HAVE_LIBMYSQLCLIENT
#include "db_interface.h"
#endif
#include "tar_frontend.h"
#include "dbs.h"
#include "types.h"
#include "support.h"
#include "configure.h"
#include "../lib/slib.h"

#ifdef HAVE_LIBMYSQLCLIENT
static int open_cmd(PXE *, REGNO, BackupHeader *, void *, void *, void **);
static int data_cmd(PXE *, REGNO, BackupHeader *, void *, void *, void **);
static int end_cmd(PXE *, REGNO, BackupHeader *, void *, void *, void **);
#endif
static int start_cmd(PXE *, REGNO, BackupHeader *, void *, void *, void **);
static int error_cmd(PXE *, REGNO, BackupHeader *, void *, void *, void **);
#ifndef HAVE_LIBMYSQLCLIENT
static int dummy_cmd(PXE *, REGNO, BackupHeader *, void *, void *, void **);
#endif

/**
 * 백업 메인쓰레드 인자 구조체
 */
typedef struct
{
    REGNO regno;             /**< 등록번호 */
    BackupHeader header;     /**< 백업모듈 헤더 */
    unsigned char data[1];   /**< 사용자 데이터 */
} BK_ARG;

/**
 * 백업명령어 콜백 함수 구조체
 */
typedef struct
{
    int cmd;	/**< 명령어 */
    int (*func)(PXE *, REGNO, BackupHeader *,
	    void *, void *, void **);	/**< 콜백함수 포인터 */
} CMD_List;

/** 
 * 백업 명령어 콜백함수 테이블
 */
static CMD_List clist[] = {
#ifdef HAVE_LIBMYSQLCLIENT
    { BACKUP_OPEN, &open_cmd },
    { BACKUP_START, &start_cmd },
    { BACKUP_DATA, &data_cmd },
    { BACKUP_END, &end_cmd },
    { BACKUP_ERROR, &error_cmd }
#else
    { BACKUP_OPEN, &dummy_cmd },
    { BACKUP_START, &start_cmd },
    { BACKUP_DATA, &dummy_cmd },
    { BACKUP_END, &dummy_cmd },
    { BACKUP_ERROR, &error_cmd }
#endif
};

/*
 * 아래 함수들은 MySQL 라이브러리가 링크된 경우 마스터로 동작하기 위해 포함되는 
 * 함수들로서 MySQL 라이브러리가 링크되지 않은 경우에는 컴파일에 포함되지 않는다.
 */
#ifdef HAVE_LIBMYSQLCLIENT
/**
 * @brief 백업 파일 생성
 * @param name - 에이전트 이름
 * @param regno - 등록번호
 * @param options - 백업 옵션
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 파일 디스크립터,\n
 *  실패 시 -1
 */
static int
create_backup_file(const char *name, REGNO regno, int options, void **msg)
{
    int fd;
    char *data_dir = NULL;
    char path[PATH_MAX];
    char filename[PATH_MAX];

    ASSERT(name);
    ASSERT(msg);

    data_dir = get_config("DataDir");
    if (!data_dir)
    {
	*msg = dup_str("'DataDir' 파라메터가 존재하지 않음");
	return -1;
    }

    snprintf(path, sizeof(path), "%s/%s", data_dir, name);
    if (mkdir(path, S_IRWXU) == -1 && errno != EEXIST)
    {
	*msg = dup_str("백업 디렉토리(%s) 생성 실패: %s", path, strerror(errno));
	FREE(data_dir);
	return -1;
    }

    snprintf(filename, sizeof(filename), "%s/%" PRIu64 ".%s", path, regno, 
	    IS_COMPRESS(options) ? GZ_EXT : TAR_EXT);
    fd = open(filename, O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
    if (fd < 0)
    {
	*msg = dup_str("백업파일(%s) 생성 실패: %s", filename, strerror(errno));
	FREE(data_dir);
	return -1;
    }

    FREE(data_dir);
    return fd;
}

/**
 * @brief 에이전트에게 BACKUP_START 명령 전송
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 백업모듈 헤더 포인터
 * @param data - 사용자 데이터
 * @param fd - 마스터 백업 파일 디스크립터
 * @return 
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
send_start_cmd(PXE *pxe, REGNO regno, 
	BackupHeader *hdr, void *data, int fd)
{
    int ret_val;
    size_t size;
    void *ptr = NULL;
    BackupHeader header;

    ASSERT(pxe && hdr);

    /* 기본 리턴값 */
    ret_val = -1;

    /* 
     * 백업모듈 헤더 복사 후 명령어와 FD 변경 
     */
    memcpy(&header, hdr, sizeof(BackupHeader));
    header.cmd = BACKUP_START;
    header.fd = fd;

    size = create_backup_packet(&ptr, regno, 
	    &header, data, hdr->len);
    if (size == 0)
	goto cleanup;
    if (set_dst_host(pxe, header.ip))
	goto cleanup;
    if (pxe_write(pxe, ptr, size))
	goto cleanup;

    ret_val = 0;
    
cleanup:
    FREE(ptr);

    return ret_val;
}

/**
 * @brief 헤더 분석 후 스위치보드에 전달 
 * @param pxe - PXE 핸들러 포인터
 * @param packet - 백업 쓰레드 해더 + 백업모듈 헤더 + 사용자 데이터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 처리한 명령어의 번호,\n
 *  실패 시 -1
 */
static int
process_packet(PXE *pxe, void *packet, void *sql, void **msg)
{
    BK_Header bk;

    ASSERT(pxe && packet);

    /* 이 시점에서 헤더는 바이트오더 변환이 안되어있음 */
    parse_bk_header(&bk, packet);
    return backup_cmd(pxe, bk.regno, (char *) packet + BK_HEADER_SIZE, sql, msg); 
}

/**
 * @brief 백업파일 제거
 * @param name - 에이전트 이름
 * @param regno - 등록번호
 * @param options - 백업 옵션
 * @return 없음
 *
 * 백업 진행 중 에러가 발생했을때 작성중이던 백업파일을 제거하기
 * 위해 사용하는 함수이다.
 */
static void
remove_backup_file(const char *name, REGNO regno, int options)
{
    char *data_dir = NULL;
    char pathname[PATH_MAX];

    data_dir = get_config("DataDir");
    if (!data_dir)
    {
	Log(ERROR, "'DataDir' 파라메터가 존재하지 않음");
	return;
    }
    snprintf(pathname, sizeof(pathname), "%s/%s/%" PRIu64 ".%s", data_dir, name,
	    regno, IS_COMPRESS(options) ? GZ_EXT : TAR_EXT);
    unlink(pathname);
    FREE(data_dir);
}

/**
 * @brief 백업 메인쓰레드
 * @param arg - BK_ARG 포인터 
 * @return NULL 포인터
 */
static void *
backup_main(void *arg)
{
    int result;
    int fd;
    sql_t sql;
    int sql_stat;
    PXE *pxe = NULL;
    BK_ARG *info = NULL;
    void *err_msg = NULL;
    void *packet = NULL;
    char *name = NULL;
    char *delete = NULL;

    /* 
     * For cleanup routine
     */
    fd = -1;
    result = -1;
    sql_stat = FALSE;

    pxe = pxe_open(arg);
    if (!pxe)
    {
	Log(ERROR, "PXE 핸들러 생성 실패: %s", strerror(errno));
	goto cleanup;
    }

    info = pxe_get_argument(arg);
    ASSERT(info);

    if (dbms_connect(&sql, &err_msg))
    {
	Log(ERROR, "%s", err_msg);
	report_result(info->regno, FAIL_STAT, "%s", err_msg);
	goto cleanup;
    }
    sql_stat = TRUE;

    name = get_agent_name(&sql, info->regno);
    if (!name || *name == '\0')
    {
	Log(ERROR, "에이전트 이름을 확인 할 수 없습니다.");
	report_result(info->regno, FAIL_STAT, "에이전트 이름을 확인 할 수 없습니다.");
	goto cleanup;
    }

    fd = create_backup_file(name, info->regno, info->header.options, &err_msg);
    if (fd < 0)
    {
	Log(ERROR, "%s", err_msg);
	report_result(info->regno, FAIL_STAT, "%s", err_msg);
	goto cleanup;
    }

    if (send_start_cmd(pxe, info->regno, &info->header, info->data, fd))
    {
	err_msg = dup_str("에이전트로 백업 시작 명령을 전송 할 수 없음: %s", 
		strerror(errno));
	Log(ERROR, "%s", err_msg);
	report_result(info->regno, FAIL_STAT, "%s", err_msg);
	goto cleanup;
    }

    /* Main loop */
    while (1)
    {
	if (pxe_read_timedwait(pxe, &packet, DBS_READ_TIMEOUT) == -1)
	{
	    err_msg = dup_str("%d초 동안 백업 데이터가 수신되지 않아서 종료되었음. "
		    "에이전트의 비정상 종료 가능성 높음!!", DBS_READ_TIMEOUT);
	    report_result(info->regno, FAIL_STAT, "%s", err_msg);
	    Log(ERROR, "%s", err_msg);
	    result = BACKUP_ERROR;  /* for removing backup file */
	    break;
	}
	result = process_packet(pxe, packet, &sql, &err_msg);
	pxe_free(packet);
	if (result < 0)
	{
	    /* 내부오류 발생 */
	    Log(ERROR, "%s", err_msg);
	    break;
	}
	if (result == BACKUP_ERROR || 
		result == BACKUP_END)
	    /* 정상 종료 */
	    break;
    }

cleanup:
    if (sql_stat == TRUE)
	db_close(&sql);
    if (fd >= 0)
	close(fd);
    if (result < 0 || result == BACKUP_ERROR)
    {
	delete = get_config("DeleteOnError");
	if (strcasecmp(delete, "YES") == 0)
	    remove_backup_file(name, info->regno, info->header.options);
	free(delete);
    }
    if (pxe)
	pxe_close(pxe);
    FREE(err_msg);
    FREE(arg);
    FREE(name);

    return NULL;
}

/**
 * @brief 백업 메인쓰레드 인자 생성
 * @param ptr - (OUT) 생성된 메인쓰레드 인자 포인터 (호출자 사용 후 FREE)
 * @param regno - 등록번호
 * @param hdr - 백업모듈 헤더
 * @param data - 데이터
 * @param len - \a data 의 길이
 * @return 
 *  성공 시 생성된 메인쓰레드 인자의 길이,\n
 *  실패 시 0
 */
static size_t
create_backup_arg(void **ptr, REGNO regno, 
	BackupHeader *hdr, void *data, size_t len)
{
    size_t total_len;
    BK_ARG *arg = NULL;

    ASSERT(ptr && hdr);

    total_len = sizeof(BK_ARG) + len;
    *ptr = calloc(1, total_len);
    if (!*ptr)
	return 0;

    arg = (BK_ARG *) *ptr;
    arg->regno = regno;
    memcpy(&arg->header, hdr, sizeof(BackupHeader));
    if (data && len > 0)
	memcpy(arg->data, data, len);

    return total_len;
}

/**
 * @brief BACKUP_OPEN 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 백업모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
open_cmd(PXE *pxe, REGNO regno, BackupHeader *hdr, 
	void *data, void *sql, void **msg)
{
    int ret_val;
    size_t size;
    int result;
    void *ptr = NULL;

    ASSERT(pxe && hdr && data && msg);

    /* 기본 리턴값 */
    ret_val = -1;

    Log(INFO, "%" PRIu64 "번 백업 요청", regno);
    size = create_backup_arg(&ptr, regno, hdr, data, hdr->len);
    if (size == 0)
    {
	*msg = dup_str("백업 메인 쓰레드에 필요한 인자 생성 실패: %s", 
		strerror(errno));
	report_result(regno, FAIL_STAT, "%s", *msg);
	goto cleanup;
    }

    result = pxe_load_thread(PXE_AUTO_ASSIGN, &backup_main, ptr, size, 
	    PXE_DETACH_THREAD);
    if (result < 0)
    {
	*msg = dup_str("백업 메인 쓰레드 로딩 실패: %s", strerror(errno));
	report_result(regno, FAIL_STAT, "%s", *msg);
	goto cleanup;
    }

    ret_val = 0;

cleanup:
    FREE(ptr);

    return ret_val;
}

/**
 * @brief BACKUP_DATA 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 백업모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
data_cmd(PXE *pxe, REGNO regno, BackupHeader *hdr, 
	void *data, void *sql, void **msg)
{
    int cnt;
    off_t offset;

    ASSERT(pxe && hdr && data && sql && msg);

    /* 압축 옵션이 없을때만 파일 정보 DB에 기록 */
    if (!IS_COMPRESS(hdr->options))
    {
	offset = lseek(hdr->fd, 0, SEEK_CUR);
	if (offset == (off_t) -1)
	{
	    *msg = dup_str("백업되는 파일의 오프셋을 계산 할 수 없음: %s", 
		    strerror(errno));
	    ret_backup_error(pxe, regno, hdr, "%s", *msg);
	    report_result(regno, ERROR_STAT, "%s", *msg);
	    return -1;
	}
	write_tar_header(sql, regno, data, hdr->len, offset);
    }

    errno = 0;
    if ((cnt = write(hdr->fd, data, hdr->len)) != hdr->len)
    {
	*msg = dup_str("백업 데이터를 디스크에 기록 할 수 없음: %s", 
		errno != 0 ? strerror(errno) : "디스크 용량 초과 확인 요망");
	ret_backup_error(pxe, regno, hdr, "%s", *msg);
	report_result(regno, ERROR_STAT, "%s", *msg);
	return -1;
    }

    return 0;
}

/**
 * @brief BACKUP_END 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 백업모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
end_cmd(PXE *pxe, REGNO regno, BackupHeader *hdr, 
	void *data, void *sql, void **msg)
{
    ASSERT(pxe && hdr && data && msg);

    report_result(regno, DONE_STAT, NULL);
    Log(INFO, "%" PRIu64 "번 백업 종료", regno);
    return 0;
}
/*
 * 여기까지가 MySQL 라이브러리가 링크된 경우에만 컴파일 되는 함수들
 */
#endif

/**
 * @brief BACKUP_ERROR 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 백업모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
error_cmd(PXE *pxe, REGNO regno, BackupHeader *hdr, 
	void *data, void *sql, void **msg)
{
    ASSERT(pxe && hdr && data && msg);

    *msg = dup_str("[원격호스트에서 전송된 에러 메세지] %s", data);
#ifdef HAVE_LIBMYSQLCLIENT
    report_result(regno, ERROR_STAT, "%s", *msg);
#endif
    return 0;
}

/**
 * @brief 네트웍으로 수신한 패킷에서 백업모듈 헤더 파싱
 * @param hdr - 파싱된 백업모듈 헤더 정보를 저장할 구조체 포인터
 * @param data - 네트웍으로 수신한 패킷 (바이트 스트림 형태)
 * @return 없음
 */
void
parse_backup_header(BackupHeader *hdr, void *data)
{
    uint8_t *ip = (uint8_t *) data;
    int32_t *cmd = (int32_t *) ((char *) ip + sizeof(uint8_t) * 4);
    TIME *ltime = (TIME *) ((char *) cmd + sizeof(int32_t));
    TIME *mtime = (TIME *) ((char *) ltime + sizeof(TIME));
    int32_t *options = (int32_t *) ((char *) mtime + sizeof(TIME));
    int32_t *fd = (int32_t *) ((char *) options + sizeof(int32_t));
    uint32_t *len = (uint32_t *) ((char *) fd + sizeof(int32_t));

    memcpy(hdr->ip, ip, sizeof(hdr->ip));
    hdr->cmd = ntohl(*cmd);
    hdr->ltime = ntohl(*ltime);
    hdr->mtime = ntohl(*mtime);
    hdr->options = ntohl(*options);
    hdr->fd = ntohl(*fd);
    hdr->len = ntohl(*len);
}

/**
 * @brief 백업모듈 스위치보드
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param data - 수신한 백업모듈 패킷 (BK 쓰레드 헤더 제외)
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 처리한 명령어의 종류,\n
 *  실패 시 -1
 */
int
backup_cmd(PXE *pxe, REGNO regno, void *data, void *sql, void **msg)
{
    int result;
    void *data_ptr = NULL;
    BackupHeader hdr;

    ASSERT(pxe && data && msg);

    parse_backup_header(&hdr, data);
    data_ptr = (char *) data + BACKUP_HEADER_SIZE;

    /* 잘못된 명령어가 입력되면 ASSERT */
    ASSERT(hdr.cmd >= 0 && hdr.cmd < ARRAY_SIZE(clist));
    
    /* Switchboard */
    result = (*clist[hdr.cmd].func)(pxe, regno, &hdr, data_ptr, sql, msg);
    if (result < 0)
	return -1;

    return hdr.cmd;
}

/**
 * @brief 백업모듈 헤더 생성 (네트웍 전송을 위해 바이트 스트림 형태로 생성)
 * @param ptr - 백업모듈 헤더를 저장할 메모리 포인터
 * @param header - 생성할 백업모듈 헤더 정보
 * @param data_len - 실제 데이터의 사이즈
 * @return 없음
 */
static void
create_backup_header(void *ptr, BackupHeader *header, unsigned int data_len)
{
    uint8_t *ip = (uint8_t *) ptr;
    int32_t *cmd = (int32_t *) ((char *) ip + sizeof(uint8_t) * 4);
    TIME *ltime = (TIME *) ((char *) cmd + sizeof(int32_t));
    TIME *mtime = (TIME *) ((char *) ltime + sizeof(TIME));
    int32_t *options = (int32_t *) ((char *) mtime + sizeof(TIME));
    int32_t *fd = (int32_t *) ((char *) options + sizeof(int32_t));
    uint32_t *len = (uint32_t *) ((char *) fd + sizeof(int32_t));

    memcpy(ip, header->ip, sizeof(header->ip));
    *cmd = htonl(header->cmd);
    *ltime = htonl(header->ltime);
    *mtime = htonl(header->mtime);
    *options = htonl(header->options);
    *fd = htonl(header->fd);
    *len = htonl(data_len);
}

/**
 * @brief 백업모듈 패킷 생성
 * @param ptr - (OUT) 생성된 백업모듈 패킷 포인터 (호출자 사용 후 FREE)
 * @param regno - 등록번호
 * @param header - 백업모듈 헤더 포인터
 * @param data - 패킷에 적재할 데이터
 * @param len - \a data 길이
 * @return
 *  성공 시 생성된 패킷의 사이즈,\n
 *  실패 시 0
 *
 * 백업모듈 패킷 = BK 쓰레드 헤더 + 백업모듈 헤더 + 데이터 + null char
 */
size_t
create_backup_packet(void **ptr, REGNO regno,
	BackupHeader *header, void *data, unsigned int len)
{
    size_t total_len;
    void *data_ptr = NULL;

    ASSERT(ptr && header);

    /* +1 is null character */
    total_len = BK_HEADER_SIZE + BACKUP_HEADER_SIZE + len + 1;
    *ptr = malloc(total_len);
    if (!*ptr)
	return 0;

    /* BK 쓰레드 헤더 생성(바이트 스트림 형태) */
    create_bk_header(*ptr, regno, BACKUP);

    /* 백업모듈 헤더 생성(바이트 스트림 형태) */
    create_backup_header((char *) *ptr + BK_HEADER_SIZE, header, len);

    /* 데이터 복사 */
    if (len > 0 && data)
    {
	data_ptr = (char *) *ptr + BK_HEADER_SIZE + BACKUP_HEADER_SIZE;
	memcpy(data_ptr, data, len);
	((unsigned char *) data_ptr)[len] = '\0';
    }

    return total_len;
}

/**
 * @brief Peer host에 에러메세지 전송 / 로컬호스트에 로그 기록
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param header - 백업모듈 헤더 포인터
 * @param format - 메세지 포멧
 * @param ... - \a format 의 가변인자
 * @return 없음
 */
void
ret_backup_error(PXE *pxe, REGNO regno, const BackupHeader *header, 
	const char *format, ...)
{
    size_t len;
    va_list ap;
    void *ptr = NULL;
    char msg[8192];
    BackupHeader hdr;

    ASSERT(pxe && header && format);

    va_start(ap, format);
    vsnprintf(msg, sizeof(msg), format, ap);
    va_end(ap);

    memcpy(&hdr, header, sizeof(BackupHeader));
    hdr.cmd = BACKUP_ERROR;
    hdr.len = strlen(msg);

    Log(ERROR, "%s", msg);
    len = create_backup_packet(&ptr, regno, &hdr, msg, hdr.len);
    if (len == 0)
    {
	Log(ERROR, "원격호스트에 전송할 에러 메세지 생성 실패: %s", 
		strerror(errno));
	goto cleanup;
    }
    if (pxe_write(pxe, ptr, len))
    {
	Log(ERROR, "원격호스트로 에러 메세지를 전송 할 수 없음: %s", 
		strerror(errno));
	goto cleanup;
    }

cleanup:
    FREE(ptr);
}

/**
 * @brief BACKUP_START 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 백업모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
start_cmd(PXE *pxe, REGNO regno, BackupHeader *hdr, 
	void *data, void *sql, void **msg)
{
    int ret_val;
    int result;
    size_t len;
    void *info = NULL;

    ASSERT(pxe && hdr && data && msg);

    /* 기본 리턴값 */
    ret_val = -1;

    Log(INFO, "%" PRIu64 "번 백업 시작", regno);
    len = create_ltar_backup_info(&info, regno, 
	    pxe_get_src_thread(pxe), hdr, data, hdr->len);
    if (len == 0)
    {
	*msg = dup_str("TAR 라이브러리로 전달할 정보 생성 실패: %s", 
		strerror(errno));
	ret_backup_error(pxe, regno, hdr, "%s", *msg);
	goto cleanup;
    }
    
    result = pxe_load_thread(PXE_AUTO_ASSIGN, &ltar_backup, info, len, 
	    PXE_DETACH_THREAD);
    if (result < 0)
    {
	*msg = dup_str("TAR 라이브러리 메인 쓰레드 로딩 실패: %s", 
		strerror(errno));
	ret_backup_error(pxe, regno, hdr, "%s", *msg);
	goto cleanup;
    }

    ret_val = 0;

cleanup:
    FREE(info);

    return ret_val;
}

#ifndef HAVE_LIBMYSQLCLIENT
/**
 * @brief Dummy Function
 */
static int 
dummy_cmd(PXE *pxe, REGNO regno, BackupHeader *hdr, 
	void *data, void *sql, void **msg)
{
    return 0;
}
#endif
