/**
 * @file core_delete.c
 * @brief Delete Core (Backend)
 */

/*
 * Delete Core (Backend)
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>

#include "core_delete.h"
#include "core.h"
#ifdef HAVE_LIBMYSQLCLIENT
#include "db_interface.h"
#endif
#include "types.h"
#include "support.h"
#include "configure.h"
#include "../lib/slib.h"

#ifdef HAVE_LIBMYSQLCLIENT
static int open_cmd(PXE *, REGNO, DeleteHeader *, void *, void *, void **);
static int end_cmd(PXE *, REGNO, DeleteHeader *, void *, void *, void **);
static int error_cmd(PXE *, REGNO, DeleteHeader *, void *, void *, void **);
#endif
static int start_cmd(PXE *, REGNO, DeleteHeader *, void *, void *, void **);

/**
 * 삭제 명령어 콜백함수 구조체
 */
typedef struct
{
    int cmd;	/**< 명령어 */
    int (*func)(PXE *, REGNO, DeleteHeader *, 
	    void *, void *, void **);	/**< 콜백 함수 포인터 */
} CMD_List;

/**
 * 삭제 쓰레드 인자 구조체
 */
typedef struct
{
    REGNO sid;            /**< 삭제 작업 자체의 등록번호 */
    REGNO bk_sid;         /**< 삭제 대상의 등록번호 */
    DeleteHeader hdr;     /**< 삭제 명령 헤더 (옵션) */
    PXEThread src_tid;    /**< 상대 호스트의 PXE 쓰레드 번호 */
    char src_ip[16];      /**< 상대 호스트의 IP */
} delete_info_t;

/**
 * 삭제 명령 콜백 함수 테이블
 */
static CMD_List clist[] = {
#ifdef HAVE_LIBMYSQLCLIENT
    { DELETE_OPEN, &open_cmd },
    { DELETE_END, &end_cmd },
    { DELETE_ERROR, &error_cmd },
#endif
    { DELETE_START, &start_cmd }
};

/**
 * @brief 삭제모듈 헤더 생성 (네트웍 전송을 위해 바이트 스트림 형태로 생성)
 * @param ptr - 삭제모듈 헤더를 저장할 메모리 포인터
 * @param header - 생성할 삭제모듈 헤더 정보
 * @param data_len - 실제 데이터의 사이즈
 * @return 없음
 */
static void
create_delete_header(void *ptr, DeleteHeader *header, unsigned int data_len)
{
    REGNO *bk_sid = (REGNO *) ptr;
    uint8_t *ip = (uint8_t *) ((char *) bk_sid + sizeof(REGNO));
    int32_t *cmd = (int32_t *) ((char *) ip + sizeof(uint8_t) * 4);
    int32_t *options = (int32_t *) ((char *) cmd + sizeof(int32_t));
    uint32_t *len = (uint32_t *) ((char *) options + sizeof(int32_t));

    *bk_sid = htonll(header->bk_sid);
    memcpy(ip, header->ip, sizeof(header->ip));
    *cmd = htonl(header->cmd);
    *options = htonl(header->options);
    *len = htonl(data_len);
}

/**
 * @brief 삭제모듈 패킷 생성
 * @param ptr - (OUT) 생성된 삭제모듈 패킷 포인터 (호출자 사용 후 FREE)
 * @param regno - 등록번호
 * @param header - 삭제모듈 헤더 포인터
 * @param data - 패킷에 적재할 데이터
 * @param len - \a data 길이
 * @return
 *  성공 시 생성된 패킷의 사이즈,\n
 *  실패 시 0
 *
 * 삭제모듈 패킷 = BK 쓰레드 헤더 + 삭제모듈 헤더 + 결과 메세지 + null char
 */
size_t
create_delete_packet(void **ptr, REGNO regno, 
	DeleteHeader *header, void *data, unsigned int len)
{
    size_t total_len;
    void *data_ptr = NULL;

    ASSERT(ptr && header);

    /* +1 is null character */
    total_len = BK_HEADER_SIZE + DELETE_HEADER_SIZE + len + 1;
    *ptr = malloc(total_len);
    if (!*ptr)
	return 0;

    /* BK 쓰레드 헤더 생성(바이트 스트림 형태) */
    create_bk_header(*ptr, regno, DELETE);

    /* 삭제모듈 헤더 생성(바이트 스트림 형태) */
    create_delete_header((char *) *ptr + BK_HEADER_SIZE, header, len);

    /* 데이터 복사 */
    if (len > 0 && data)
    {
	data_ptr = (char *) *ptr + BK_HEADER_SIZE + DELETE_HEADER_SIZE;
	memcpy(data_ptr, data, len);
	((unsigned char *) data_ptr)[len] = '\0';
    }

    return total_len;
}

/**
 * @brief 네트웍으로 수신한 패킷에서 삭제모듈 헤더 파싱
 * @param hdr - 파싱된 삭제모듈 헤더 정보를 저장할 구조체 포인터
 * @param data - 네트웍으로 수신한 패킷 (바이트 스트림 형태)
 * @return 없음
 */
void
parse_delete_header(DeleteHeader *hdr, void *data)
{
    REGNO *bk_sid = (REGNO *) data;
    uint8_t *ip = (uint8_t *) ((char *) bk_sid + sizeof(REGNO));
    int32_t *cmd = (int32_t *) ((char *) ip + sizeof(uint8_t) * 4);
    int32_t *options = (int32_t *) ((char *) cmd + sizeof(int32_t));
    uint32_t *len = (uint32_t *) ((char *) options + sizeof(int32_t));

    hdr->bk_sid = ntohll(*bk_sid);
    memcpy(hdr->ip, ip, sizeof(hdr->ip));
    hdr->cmd = ntohl(*cmd);
    hdr->options = ntohl(*options);
    hdr->len = ntohl(*len);
}

/**
 * @brief 삭제모듈 스위치보드
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param data - 수신한 삭제모듈 패킷 (BK 쓰레드 헤더 제외)
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 처리한 명령어의 종류,\n
 *  실패 시 -1
 */
int
delete_cmd(PXE *pxe, REGNO regno, void *data, void *sql, void **msg)
{
    int i;
    int result;
    void *data_ptr = NULL;
    DeleteHeader hdr;

    ASSERT(pxe && data && msg);

    parse_delete_header(&hdr, data);
    data_ptr = (char *) data + DELETE_HEADER_SIZE;

    /* Switchboard */
    for (i = 0; i < ARRAY_SIZE(clist); i++)
    {
	if (clist[i].cmd == hdr.cmd)
	{
	    result = (*clist[i].func)(pxe, regno, &hdr, data_ptr, sql, msg);
	    if (result < 0)
		return -1;
	    else
		break;
	}
    }
    /* 잘못된 명령어가 입력되면 ASSERT */
    ASSERT(i != ARRAY_SIZE(clist));

    return clist[i].cmd;
}

/*
 * 아래 함수들은 MySQL 라이브러리가 링크된 경우 마스터로 동작하기 위해 포함되는 
 * 함수들로서 MySQL 라이브러리가 링크되지 않은 경우에는 컴파일에 포함되지 않는다.
 */
#ifdef HAVE_LIBMYSQLCLIENT
/**
 * @brief TAR 아카이브 파일 삭제
 * @param sql - SQL 핸들러
 * @param regno - 등록번호
 * @param err - (OUT) 에러 발생 시 에러메세지 저장할 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
remove_tar_archive(sql_t *sql, REGNO regno, void **err)
{
    int compressed;
    char *datadir = NULL;
    char filename[PATH_MAX];
    char *name = NULL;

    datadir = get_config("DataDir");
    if (!datadir)
    {
	*err = dup_str("'DataDir' 파라메터가 존재하지 않음");
	return -1;
    }

    name = get_agent_name(sql, regno);
    if (!name || *name == '\0')
    {
	*err = dup_str("에이전트 이름을 확인 할 수 없습니다.");
	return -1;
    }

    compressed = is_compressed(sql, regno, err);
    if (compressed < 0)
    {
	FREE(datadir);
	FREE(name);
	return -1;
    }

    snprintf(filename, sizeof(filename), "%s/%s/%" PRIu64 ".%s", datadir, name,
	    regno, compressed == 1 ? GZ_EXT : TAR_EXT);
    if (unlink(filename) == -1)
    {
	*err = dup_str("백업된 파일(%s) 삭제 실패: %s", filename, 
		strerror(errno));
	FREE(datadir);
	FREE(name);
	return -1;
    }

    FREE(datadir);
    FREE(name);

    return 0;
}

/**
 * @brief 백업 관련 DB 엔트리 삭제
 * @param sql - SQL 핸들러
 * @param regno - 등록번호
 * @return 없음
 */
static void
remove_db_entry(sql_t *sql, REGNO regno)
{
    char query[8192];

    /* 현재 상태를 삭제로 변경 */
    snprintf(query, sizeof(query), "UPDATE job SET status='%c' WHERE id='%llu'", 
	    CANCEL_STAT, regno);
    (void) db_query(sql, query, NULL);
}

/**
 * @brief 백업 데이터 삭제
 * @param sql - SQL 핸들러
 * @param regno - 등록번호
 * @param err - (OUT) 에러 발생 시 에러메세지 저장할 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 0,\n
 *  실패 시 -1
 *
 * 백업된 TAR 아카이브 파일을 삭제하고 JOB, RECORD 테이블에서 관련 내용을 모두 
 * 삭제한다.
 */
int
delete_backup_data(sql_t *sql, REGNO regno, void **err)
{
    if (remove_tar_archive(sql, regno, err))
	return -1;
    remove_db_entry(sql, regno);

    return 0;
}

/**
 * @brief 에이전트에게 DELETE_START 명령 전송
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 삭제모듈 헤더 포인터
 * @return 
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
send_start_cmd(PXE *pxe, REGNO regno, DeleteHeader *hdr)
{
    int ret_val;
    size_t size;
    void *ptr = NULL;
    DeleteHeader header;

    ASSERT(pxe && hdr);

    /* 기본 리턴값 */
    ret_val = -1;

    /* 
     * 삭제모듈 헤더 복사 후 명령어 변경 
     */
    memcpy(&header, hdr, sizeof(DeleteHeader));
    header.cmd = DELETE_START;

    size = create_delete_packet(&ptr, regno, &header, NULL, 0);
    if (size == 0)
	goto cleanup;
    if (set_dst_host(pxe, header.ip))
	goto cleanup;
    if (pxe_write(pxe, ptr, size))
	goto cleanup;

    ret_val = 0;
    
cleanup:
    FREE(ptr);

    return ret_val;
}

/**
 * @brief 헤더 분석 후 스위치보드에 전달 
 * @param pxe - PXE 핸들러 포인터
 * @param packet - 백업 쓰레드 해더 + 삭제모듈 헤더 + 사용자 데이터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 처리한 명령어의 번호,\n
 *  실패 시 -1
 */
static int
process_packet(PXE *pxe, void *packet, void *sql, void **msg)
{
    BK_Header bk;

    ASSERT(pxe && packet);

    /* 이 시점에서 헤더는 바이트오더 변환이 안되어있음 */
    parse_bk_header(&bk, packet);
    return delete_cmd(pxe, bk.regno, (char *) packet + BK_HEADER_SIZE, sql, msg); 
}

/**
 * @brief JOB 테이블 백업 기록에서 로컬 복사본 옵션 플래그 제거
 * @param sql - SQL 핸들러
 * @param regno - 대상 등록번호
 * @return 없음
 *
 * 로컬 복사본을 삭제하면 JOB 테이블의 백업 기록에서 로컬 복사본 옵션을 제거해서
 * 에이전트에 복사본 파일이 없다는것을 나타낸다. JOB 테이블에 로컬 복사본 유무를
 * 나타내는 필드를 따로 만드려고 했으나, 이미 옵션으로 로컬 복사본 유무를 나타내고
 * 있기 때문에 따로 만들지 않고 옵션에서 제거하는 방식을 선택했다.
 */
static void
unset_local_copy_flag(sql_t *sql, REGNO regno)
{
    char query[8192];

    snprintf(query, sizeof(query), "UPDATE job SET options = options & ~%d "
	    "WHERE id = '%" PRIu64 "'", LOCAL_COPY, regno);
    (void) db_query(sql, query, NULL);
}

/**
 * @brief 삭제 메인쓰레드
 * @param arg - DEL_ARG 포인터 
 * @return NULL 포인터
 */
static void *
delete_main(void *arg)
{
    sql_t sql;
    int result;
    int sql_stat = FALSE;
    void *err_msg = NULL;
    PXE *pxe = NULL;
    DEL_ARG *info = NULL;
    void *packet = NULL;

    pxe = pxe_open(arg);
    if (!pxe)
    {
	Log(ERROR, "PXE 핸들러 생성 실패: %s", strerror(errno));
	goto cleanup;
    }

    info = pxe_get_argument(arg);
    ASSERT(info);

    if (dbms_connect(&sql, &err_msg))
    {
	Log(ERROR, "%s", err_msg);
	report_result(info->regno, FAIL_STAT, "%s", err_msg);
	goto cleanup;
    }
    sql_stat = TRUE;

    /* 마스터쪽 파일을 삭제하는 경우 에이전트로 패킷 전달없이 실행 */
    if (!IS_REMOVE_LOCAL_COPY(info->header.options))
    {
	if (delete_backup_data(&sql, info->header.bk_sid, &err_msg))
	{
	    report_result(info->regno, FAIL_STAT, "%s", err_msg);
	    Log(ERROR, "%s", err_msg);
	}
	else
	    report_result(info->regno, DONE_STAT, NULL);
	goto cleanup;
    }

    /* 로컬복사본 삭제를 위해 패킷을 에이전트로 전송 */
    if (send_start_cmd(pxe, info->regno, &info->header))
    {
	err_msg = dup_str("에이전트로 삭제 시작 명령을 전송 할 수 없음: %s", 
		strerror(errno));
	report_result(info->regno, FAIL_STAT, "%s", err_msg);
	Log(ERROR, "%s", err_msg);
	goto cleanup;
    }

    /* 삭제 결과 수신 */
    if (pxe_read_timedwait(pxe, &packet, DBS_READ_TIMEOUT) == -1)
    {
	err_msg = dup_str("%d초 동안 삭제 결과가 수신되지 않아서 종료되었음. "
		"에이전트의 비정상 종료 가능성 높음!!", DBS_READ_TIMEOUT);
	report_result(info->regno, FAIL_STAT, "%s", err_msg);
	Log(ERROR, "%s", err_msg);
	goto cleanup;
    }
    result = process_packet(pxe, packet, &sql, &err_msg);
    pxe_free(packet);
    if (result < 0)
    {
	/* 내부오류 발생 */
	Log(ERROR, "%s", err_msg);
	goto cleanup;
    }

    /* 에이전트 백업 복사본이 삭제되었기 때문에 JOB 테이블 row에서 복사본 옵션을
     * 제거한다.
     */
    unset_local_copy_flag(&sql, info->header.bk_sid);

cleanup:
    if (sql_stat == TRUE)
	db_close(&sql);
    if (pxe)
	pxe_close(pxe);
    FREE(err_msg);
    FREE(arg);

    return NULL;
}

/**
 * @brief 삭제 메인쓰레드 인자 생성
 * @param ptr - (OUT) 생성된 메인쓰레드 인자 포인터 (호출자 사용 후 FREE)
 * @param regno - 등록번호
 * @param hdr - 삭제모듈 헤더
 * @param data - 데이터
 * @param len - \a data 의 길이
 * @return 
 *  성공 시 생성된 메인쓰레드 인자의 길이,\n
 *  실패 시 0
 */
static size_t
create_delete_arg(void **ptr, REGNO regno, 
	DeleteHeader *hdr, void *data, size_t len)
{
    size_t total_len;
    DEL_ARG *arg = NULL;

    ASSERT(ptr && hdr);

    total_len = sizeof(DEL_ARG) + len;
    *ptr = calloc(1, total_len);
    if (!*ptr)
	return 0;

    arg = (DEL_ARG *) *ptr;
    arg->regno = regno;
    memcpy(&arg->header, hdr, sizeof(DeleteHeader));
    if (data && len > 0)
	memcpy(arg->data, data, len);

    return total_len;
}

/**
 * @brief DELETE_OPEN 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 삭제모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
open_cmd(PXE *pxe, REGNO regno, DeleteHeader *hdr, 
	void *data, void *sql, void **msg)
{
    int ret_val;
    size_t size;
    int result;
    void *ptr = NULL;

    ASSERT(pxe && hdr && data && msg);

    /* 기본 리턴값 */
    ret_val = -1;

    Log(INFO, "%" PRIu64 "번 삭제 요청", regno);
    size = create_delete_arg(&ptr, regno, hdr, data, hdr->len);
    if (size == 0)
    {
	*msg = dup_str("삭제 메인 쓰레드에 필요한 인자 생성 실패: %s", 
		strerror(errno));
	report_result(regno, FAIL_STAT, "%s", *msg);
	goto cleanup;
    }

    result = pxe_load_thread(PXE_AUTO_ASSIGN, &delete_main, ptr, size, 
	    PXE_DETACH_THREAD);
    if (result < 0)
    {
	*msg = dup_str("삭제 메인 쓰레드 로딩 실패: %s", strerror(errno));
	report_result(regno, FAIL_STAT, "%s", *msg);
	goto cleanup;
    }

    ret_val = 0;

cleanup:
    FREE(ptr);

    return ret_val;
}

/**
 * @brief DELETE_END 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 삭제모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
end_cmd(PXE *pxe, REGNO regno, DeleteHeader *hdr, 
	void *data, void *sql, void **msg)
{
    report_result(regno, DONE_STAT, NULL);
    Log(INFO, "%" PRIu64 "번 삭제 종료", regno);

    return 0;
}

/**
 * @brief DELETE_ERROR 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 삭제모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
error_cmd(PXE *pxe, REGNO regno, DeleteHeader *hdr, 
	void *data, void *sql, void **msg)
{
    *msg = dup_str("[원격호스트에서 전송된 에러 메세지] %s", data);
    report_result(regno, ERROR_STAT, "%s", *msg);
    return 0;
}
/*
 * 여기까지가 MySQL 라이브러리가 링크된 경우에만 컴파일 되는 함수들
 */
#endif

/**
 * @brief Peer host에 에러메세지 전송 / 로컬호스트에 로그 기록
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param header - 삭제모듈 헤더 포인터
 * @param format - 메세지 포멧
 * @param ... - \a format 의 가변인자
 * @return 없음
 */
static void
ret_delete_error(PXE *pxe, REGNO regno, const DeleteHeader *header, 
	const char *format, ...)
{
    size_t len;
    va_list ap;
    void *ptr = NULL;
    char msg[8192];
    DeleteHeader hdr;

    ASSERT(pxe && header && format);

    va_start(ap, format);
    vsnprintf(msg, sizeof(msg), format, ap);
    va_end(ap);

    memcpy(&hdr, header, sizeof(DeleteHeader));
    hdr.cmd = DELETE_ERROR;
    hdr.len = strlen(msg);

    Log(ERROR, "%s", msg);
    len = create_delete_packet(&ptr, regno, &hdr, msg, hdr.len);
    if (len == 0)
    {
	Log(ERROR, "원격호스트로 전송할 에러메세지 생성 실패: %s", 
		strerror(errno));
	goto cleanup;
    }
    if (pxe_write(pxe, ptr, len))
    {
	Log(ERROR, "원격호스트로 에러메세지를 전송 할 수 없음: %s", 
		strerror(errno));
	goto cleanup;
    }

cleanup:
    FREE(ptr);
}

/**
 * @brief Peer host에 삭제 성공 패킷 전송
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param header - 삭제모듈 헤더 포인터
 * @return 없음
 */
static void
ret_delete_success(PXE *pxe, REGNO regno, const DeleteHeader *header) 
{
    size_t len;
    void *ptr = NULL;
    DeleteHeader hdr;

    ASSERT(pxe && header);

    memcpy(&hdr, header, sizeof(DeleteHeader));
    hdr.cmd = DELETE_END;

    len = create_delete_packet(&ptr, regno, &hdr, NULL, 0);
    if (len == 0)
    {
	Log(ERROR, "원격호스트로 전송할 메세지 생성 실패: %s", 
		strerror(errno));
	goto cleanup;
    }
    if (pxe_write(pxe, ptr, len))
    {
	Log(ERROR, "원격호스트로 메세지를 전송 할 수 없음: %s", 
		strerror(errno));
	goto cleanup;
    }

cleanup:
    FREE(ptr);
}

/**
 * @brief 에이전트에 저장된 로컬복사본 파일 삭제
 * @param regno - 삭제할 파일의 작업번호
 * @param err - (OUT) 에러 발생시 에러 메세지를 저장 (사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
remove_client_tar_archive(REGNO regno, void **err)
{
    int result, gz_result;
    char *datadir = NULL;
    char filename[PATH_MAX];

    datadir = get_config("DataDir");
    if (!datadir)
    {
	*err = dup_str("'DataDir' 파라메터가 존재하지 않음");
	return -1;
    }

    /* FIXME -
     * 에이전트에서는 백업 데이터베이스에 접근 할 수 없기 때문에 해당 백업 작업이
     * 압축이 된건지 아닌지 알 수 없다. 따라서 마스터에서 삭제 명령을 에이전트에게
     * 보낼때부터 압축 여부를 알려줘야 하는데, 현재는 구현되지 않아서 압축/비압축
     * 확장자를 모두 사용해서 삭제를 시도한다.
     */
    snprintf(filename, sizeof(filename), "%s/%" PRIu64 ".%s", datadir, regno, 
	    TAR_EXT);
    result = unlink(filename);

    snprintf(filename, sizeof(filename), "%s/%" PRIu64 ".%s", datadir, regno, 
	    GZ_EXT);
    gz_result = unlink(filename);

    if (result == -1 && gz_result == -1)
    {
	*err = dup_str("%s", strerror(errno));
	FREE(datadir);
	return -1;
    }

    FREE(datadir);
    return 0;
}

/**
 * @brief 삭제 쓰레드의 인자 생성
 * @param info - 생성되는 인자를 저장할 포인터
 * @param pxe - PXE 핸들러
 * @param sid - 삭제 명령 자체의 등록번호
 * @param bk_sid - 삭제 대상 작업의 등록번호
 * @param hdr - 삭제 모듈 헤더
 * @return 없음
 */
static void 
make_delete_arg(delete_info_t *info, PXE *pxe, REGNO sid, REGNO bk_sid, 
	DeleteHeader *hdr)
{
    info->sid = sid;
    info->bk_sid = bk_sid;
    memcpy(&info->hdr, hdr, sizeof(DeleteHeader));
    info->src_tid = pxe_get_src_thread(pxe);
    pxe_get_src_ip(pxe, info->src_ip, sizeof(info->src_ip));
}

/**
 * @brief 삭제대상 파일 제거
 * @param arg - PXE 쓰레드 인자
 * @return NULL
 */
static void *
remove_job(void *arg)
{
    PXE *pxe = NULL;
    void *err_msg = NULL;
    delete_info_t *info = NULL;

    pxe = pxe_open(arg);
    if (!pxe)
    {
	Log(ERROR, "PXE 핸들러 생성 실패: %s", strerror(errno));
	goto cleanup;
    }

    info = pxe_get_argument(arg);
    if (pxe_set_destination(pxe, info->src_ip, info->src_tid))
    {
	Log(ERROR, "삭제 실패: 유효하지 않은 PXE 핸들러");
	goto cleanup;
    }

    Log(INFO, "%" PRIu64 "번 삭제 시작", info->sid);
    if (remove_client_tar_archive(info->bk_sid, &err_msg))
    {
	ret_delete_error(pxe, info->sid, &info->hdr, "%" PRIu64 "번 백업 파일 "
		"삭제 실패: %s", info->bk_sid, err_msg);
	goto cleanup;
    }
    ret_delete_success(pxe, info->sid, &info->hdr);
    Log(INFO, "%" PRIu64 "번 삭제 완료", info->sid);

cleanup:
    pxe_close(pxe);
    FREE(err_msg);
    FREE(arg);
    return NULL;
}

/**
 * @brief DELETE_START 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 삭제모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
start_cmd(PXE *pxe, REGNO regno, DeleteHeader *hdr, 
	void *data, void *sql, void **msg)
{
    int result;
    delete_info_t info;

    ASSERT(pxe && hdr && data && msg);
    make_delete_arg(&info, pxe, regno, hdr->bk_sid, hdr);
    result = pxe_load_thread(PXE_AUTO_ASSIGN, &remove_job, &info, 
	    sizeof(delete_info_t), PXE_DETACH_THREAD);
    if (result < 0)
	goto error;

    return 0;

error:
    ret_delete_error(pxe, regno, hdr, "%" PRIu64 "번 백업 파일 삭제 실패: %s", 
	    hdr->bk_sid, strerror(errno));
    /* 삭제 모듈의 경우 마스터/에이전트 양쪽 모두 core()를 통해서 처리가 
     * 되기 때문에 에러가 발생한 경우 여기서 로그를 기록하고 상위 루틴에서
     * 다시 로그가 기록되지 않도록 메세지 변수를 NULL 처리한다.
     */
    *msg = NULL;
    return -1;
}
