/**
 * @file core_restore.c
 * @brief Restore Core (Backend)
 */

/*
 * Restore Core (Backend)
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태(superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <inttypes.h>
#include <unistd.h>
#include <fnmatch.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <regex.h>
#include <limits.h>
#include <stdbool.h>

#include <pxe.h>
#include <libtar.h>
#include "core_restore.h"
#include "core.h"
#ifdef HAVE_LIBMYSQLCLIENT
#include "db_interface.h"
#endif
#include "types.h"
#include "tar_frontend.h"
#include "configure.h"
#include "support.h"

/**
 * 복구 데이터 전송 후 처리결과 대기 시간 
 */
#define FINISH_WAIT_DELAY    120   

static int open_cmd(PXE *, REGNO, RestoreHeader *, void *, void *, void **);
#ifdef HAVE_LIBMYSQLCLIENT
static int start_cmd(PXE *, REGNO, RestoreHeader *, void *, void *, void **);
#endif
static int end_cmd(PXE *, REGNO, RestoreHeader *, void *, void *, void **);
static int error_cmd(PXE *, REGNO, RestoreHeader *, void *, void *, void **);

/**
 * 복구 메인쓰레드 인자 구조체
 */
typedef struct
{
    REGNO regno;             /**< 등록번호 */
    PXEThread src;           /**< Peer's Thread ID */
    RestoreHeader header;    /**< 복구모듈 헤더 */
    char data[1];            /**< 사용자 데이터 */
} RT_ARG;

/**
 * 복구 명령어 콜백함수 구조체
 */
typedef struct
{
    int cmd;	/**< 명령어 */
    int (*func)(PXE *, REGNO, RestoreHeader *, 
	    void *, void *, void **);	/**< 콜백 함수 포인터 */
} CMD_List;

/**
 * 복구 명령 콜백 함수 테이블
 */
static CMD_List clist[] = {
    { RESTORE_OPEN, &open_cmd },
#ifdef HAVE_LIBMYSQLCLIENT
    { RESTORE_START, &start_cmd },
#endif
    { RESTORE_END, &end_cmd },
    { RESTORE_ERROR, &error_cmd }
};

/*
 * 아래 함수들은 MySQL 라이브러리가 링크된 경우 마스터로 동작하기 위해 포함되는 
 * 함수들로서 MySQL 라이브러리가 링크되지 않은 경우에는 컴파일에 포함되지 않는다.
 */
#ifdef HAVE_LIBMYSQLCLIENT
/**
 * @brief 복구 메인쓰레드 인자 생성
 * @param ptr - (OUT) 생성된 메인쓰레드 인자 포인터 (호출자 사용 후 FREE)
 * @param regno - 등록번호
 * @param src - Peer 호스트의 쓰레드 인덱스
 * @param hdr - 복구모듈 헤더
 * @param data - 데이터
 * @param len - \a data 의 길이
 * @return
 *  성공 시 생성된 메인쓰레드 인자의 길이,\n
 *  실패 시 0
 */
static size_t
create_restore_arg(void **ptr, REGNO regno, PXEThread src,
	RestoreHeader *hdr, void *data, size_t len)
{
    size_t total_len;
    RT_ARG *arg = NULL;

    ASSERT(ptr && hdr);

    total_len = sizeof(RT_ARG) + len;
    *ptr = calloc(1, total_len);
    if (!*ptr)
	return 0;

    arg = (RT_ARG *) *ptr;
    arg->regno = regno;
    arg->src = src;
    memcpy(&arg->header, hdr, sizeof(RestoreHeader));
    if (data && len > 0)
	memcpy(arg->data, data, len);

    return total_len;
}

/**
 * @brief 백업 파일 오픈
 * @param name - 에이전트 이름
 * @param regno - 복구 대상 백업 등록번호
 * @param options - 복구 옵션
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 파일 디스크립터,\n
 *  실패 시 -1
 */
static int
open_backup_file(const char *name, REGNO regno, int options, void **msg)
{
    int fd;
    char *data_dir = NULL;
    char filename[PATH_MAX];

    ASSERT(msg);

    data_dir = get_config("DataDir");
    if (!data_dir)
    {
	*msg = dup_str("'DataDir' 파라메터가 존재하지 않음");
	return -1;
    }
    
    snprintf(filename, sizeof(filename), "%s/%s/%" PRIu64 ".%s", data_dir, name, 
	    regno, IS_COMPRESS(options) ? GZ_EXT : TAR_EXT);

    fd = open(filename, O_RDONLY);
    if (fd < 0)
    {
	*msg = dup_str("백업된 파일(%s)을 열 수 없음: %s", filename, 
		strerror(errno));
	FREE(data_dir);
	return -1;
    }

    FREE(data_dir);
    return fd;
}

/**
 * @brief 복구 작업 데이터(RESTORE_DATA) 전송
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param buf - 전송할 데이터 포인터
 * @param size - \a buf 의 사이즈
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
send_restore_data(PXE *pxe, REGNO regno, RestoreHeader *hdr, 
	void *buf, size_t size, void **msg)
{
    size_t len;
    void *ptr = NULL;
    RestoreHeader header;

    ASSERT(pxe && hdr && buf && msg);

    memcpy(&header, hdr, sizeof(RestoreHeader));
    header.cmd = RESTORE_DATA;

    len = create_restore_packet(&ptr, regno, &header, buf, size);
    if (len == 0 || !ptr)
    {
	*msg = dup_str("복구용 패킷 생성 실패: %s", 
		strerror(errno));
	return -1;
    }

    if (pxe_write(pxe, ptr, len))
    {
	FREE(ptr);
	*msg = dup_str("복구용 패킷 전송 실패: %s", 
		strerror(errno));
	return -1;
    }

    FREE(ptr);
    return 0;
}

/**
 * @brief 복구 에러 통보 패킷인지 확인
 * @param packet - 수신한 패킷 포인터
 * @param msg - (OUT) 에러 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  에러 통보 패킷인 경우 TRUE,\n
 *  아닌 경우 FALSE
 */
static BOOL
is_restore_error(void *packet, void **msg)
{
    char *data = NULL;
    RestoreHeader restore_hdr;

    ASSERT(packet && msg);

    parse_restore_header(&restore_hdr, (char *) packet + BK_HEADER_SIZE);
    data = (char *) packet + BK_HEADER_SIZE + RESTORE_HEADER_SIZE;

    if (restore_hdr.cmd != RESTORE_ERROR)
	return FALSE;

    *msg = calloc(1, restore_hdr.len + 1);
    if (*msg)
	memcpy(*msg, data, restore_hdr.len);

    return TRUE;
}

/**
 * @brief 전체 TAR 파일 모두 전송
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 복구 모듈 헤더 포인터
 * @param fd - TAR 파일 디스크립터
 * @param msg - (OUT) 에러 발생 시 에러 메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 DBS_ERR_OK,\n
 *  실패 시 DBS_ERR_PEER or DBS_ERR_LOCAL
 */
static int
send_whole_file(PXE *pxe, REGNO regno, 
	RestoreHeader *hdr, int fd, void **msg)
{
    ssize_t cnt;
    int pxe_options;
    void *packet = NULL;
    unsigned char buf[8192];
    
    ASSERT(pxe && hdr && msg);

    while ((cnt = read(fd, buf, sizeof(buf))) != 0)
    {
	if (cnt < 0)
	{
	    *msg = dup_str("백업 파일을 읽을 수 없음: %s", 
		    strerror(errno));
	    return DBS_ERR_LOCAL;
	}
	
	/* 
	 * 자신한테 수신된 패킷이 있는지 확인하고 에러 패킷이면 에러 처리
	 */
	pxe_options = pxe_get_options(pxe);
	pxe_set_options(pxe, pxe_options | PXE_NON_BLOCK);
	if (pxe_read(pxe, &packet))
	{
	    if (is_restore_error(packet, msg))
	    {
		pxe_free(packet);
		/* 상대방 호스트에서 에러 발생 */
		return DBS_ERR_PEER;
	    }
	    pxe_free(packet);
	}
	pxe_set_options(pxe, pxe_options);

	if (IS_TIME_LIMIT(hdr->options))
	{
	    if (check_timeout(hdr->ltime))
	    {
		*msg = strdup("작업 취소!! 제한시간이 만료되었음.");
		return DBS_ERR_LOCAL;
	    }
	}

	if (send_restore_data(pxe, regno, hdr, buf, cnt, msg))
	    return DBS_ERR_LOCAL;
    }

    return DBS_ERR_OK;
}

/**
 * @brief 문자열 배열 생성
 * @param filename - \\n을 구분자로 절대경로를 저장하고 있는 문자열 포인터
 * @return 생성된 문자열 배열 (호출자 사용 후 FREE)
 */
static char **
make_glob_list(const char *filename)
{
    int i;
    int cnt;
    char *dup_str = NULL;
    char **glob = NULL;

    ASSERT(filename);

    dup_str = strdup(filename);
    if (!dup_str)
	return NULL;

    /* 마지막 문자열과 리스트의 끝을 나타내는 NULL을 위해 2부터 시작 */
    cnt = 2;
    for (i = 0; i < strlen(dup_str); i++)
    {
	if (dup_str[i] == '\n')
	    cnt++;
    }

    glob = malloc(cnt * sizeof(char *));
    if (!glob)
    {
	FREE(dup_str);
	return NULL;
    }

    {
	int j;
	char *ptr = NULL;
	char *end = NULL;

	ptr = dup_str;
	for (j = 0; (end = strchr(ptr, '\n')) != NULL; j++)
	{
	    *end++ = '\0';  
	    glob[j] = strdup(ptr);
	    ptr = end;
	}
	glob[j++] = strdup(ptr);
	glob[j] = NULL;
    }

    FREE(dup_str);
    return glob;
}

/**
 * @brief make_glob_list에 의해 생성된 문자열 배열 해제
 * @param glob - 해재할 문자열 배열 포인터
 * @return 없음
 */
static void
free_glob_list(char **glob)
{
    int i;

    ASSERT(glob);

    for (i = 0; glob[i] != NULL; i++)
	FREE(glob[i]);

    FREE(glob);
}

/**
 * @brief TAR 아카이브에서 해당 파일이 사용하는 TAR BLOCK 수 측정
 * @param tar - TAR 헤더 포인터
 * @return 측정된 TAR BLOCK 수
 */
static size_t 
get_block_size(struct tar_header *tar)
{
    size_t size;
    size_t block;

    size = oct_to_int(tar->size);
    block = (size / T_BLOCKSIZE) + (size % T_BLOCKSIZE ? 1 : 0);

    return block;
}

/**
 * @brief TAR 파일에서 헤더 부분을 읽어들임
 * @param fd - TAR 파일 디스크립터
 * @param buf - 헤더를 저장할 버퍼 포인터
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 읽어들인 헤더 바이트 수,\n
 *  실패 시 -1
 */
static ssize_t
read_tar_header(int fd, void *buf, void **msg)
{
    int num_zero;
    ssize_t cnt;

    num_zero = 0;
    while ((cnt = read(fd, buf, T_BLOCKSIZE)))
    {
	if (cnt == 0)
	    return 0;
	else if (cnt < 0)
	{
	    *msg = dup_str("백업 파일을 읽을 수 없음: %s", 
		    strerror(errno));
	    return -1;
	}
	else if (cnt != T_BLOCKSIZE)
	{
	    *msg = dup_str("백업 파일을 읽을 수 없음: 잘못된 TAR 포맷"); 
	    return -1;
	}
	else
	{
	    if (((unsigned char *) buf)[0] == '\0')
	    {
		/* 
		 * EOF 블럭인지 확인 
		 */
		num_zero++;
		if (num_zero >= 2)
		    return 0;
		else 
		    continue;
	    }
	    else
	    {
		/* 일반 블럭을 read 한 경우 */
		break;
	    }
	}
    }

    return cnt;
}

/**
 * @brief TAR 파일이 일반파일인지 확인
 * @param tar - TAR 헤더 포인터
 * @return
 *  일반 파일인 경우 TRUE,\n
 *  아닌 경우 FALSE
 */
static BOOL
is_tar_regfile(struct tar_header *tar)
{
    if (tar->typeflag == REGTYPE || 
	    tar->typeflag == AREGTYPE || 
	    tar->typeflag == CONTTYPE || 
	    (S_ISREG((mode_t) oct_to_int(tar->mode)) && tar->typeflag != LNKTYPE))
	return TRUE;

    return FALSE;
}

/**
 * @brief TAR 아카이브에 있는 실제 파일 내용을 전송
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param fd - TAR 파일 디스크립터
 * @param block - 해당 파일이 사용하고 있는 TAR BLOCK 수
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
send_real_data(PXE *pxe, REGNO regno, RestoreHeader *hdr, 
	int fd, size_t block, void **msg)
{
    size_t len;
    ssize_t cnt;
    unsigned char buf[8192];
    uint64_t filesize;

    filesize = block * T_BLOCKSIZE;

    while (filesize > 0)
    {
	len = filesize > sizeof(buf) ? sizeof(buf) : filesize;
	cnt = read(fd, buf, len);
	if (cnt < 0)
	{
	    *msg = dup_str("백업 파일을 읽을 수 없음: %s", strerror(errno));
	    return -1;
	}
	else if (cnt != len)
	{
	    *msg = dup_str("백업 파일을 읽을 수 없음: 잘못된 TAR 포맷");
	    return -1;
	}

	if (send_restore_data(pxe, regno, hdr, buf, cnt, msg))
	    return -1;

	filesize -= cnt;
    }

    return 0;
}

/**
 * Peer host로부터 수신된 에러 알림이 있는지 확인한다.
 *
 * @param pxe - PXE 핸들러
 * @param msg - 에러 발생 시 에러 메세지를 저장할 포인터
 * @return - TRUE; 수신된 에러가 없는 경우는 FALSE
 * @warning - \a msg 는 사용 후 FREE() 필요
 */
static BOOL
check_error_notify(PXE *pxe, void **msg)
{
    void *packet = NULL;
    int pxe_options;

    ASSERT(pxe != NULL);
    ASSERT(msg != NULL);

    pxe_options = pxe_get_options(pxe);
    pxe_set_options(pxe, pxe_options | PXE_NON_BLOCK);
    if (pxe_read(pxe, &packet))
    {
	if (is_restore_error(packet, msg))
	{
	    pxe_free(packet);
	    return TRUE;
	}
	pxe_free(packet);
    }
    pxe_set_options(pxe, pxe_options);

    return FALSE;
}

/**
 * 백업된 데이터를 읽어서 에이전트로 전송한다.
 *
 * @param pxe - PXE 핸들러
 * @param fd - TAR 아카이브 파일 디스크립터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더
 * @param msg - 에러 발생 시 에러메세지 포인터
 * @return - 0; 에러 발생 시 -1
 * @warning - \a msg 는 사용 후 FREE() 필요
 */
static int
send_backup_data(PXE *pxe, int fd, REGNO regno, RestoreHeader *hdr, void **msg)
{
    ssize_t cnt;
    char buf[8192];
    struct tar_header *tar = NULL;
    size_t size;

    /* 최초 TAR 헤더 읽어들임 */
    cnt = read_tar_header(fd, buf, msg);
    /* TAR EOF가 나온 경우 */
    if (cnt == 0)
	/* 정상 종료 처리 */
	return 0;
    else if (cnt < 0)
	return -1;

    tar = (struct tar_header *) buf;

    while (tar->typeflag == GNU_LONGNAME_TYPE || 
	    tar->typeflag == GNU_LONGLINK_TYPE)
    {
	/* Longname/Longlink 헤더 블럭 전송 */
	if (send_restore_data(pxe, regno, hdr, buf, T_BLOCKSIZE, msg))
	    return -1;

	/* Longname/Longlink 데이터 블럭 전송 */
	size = get_block_size(tar);
	if (send_real_data(pxe, regno, hdr, fd, size, msg))
	    return -1;

	/* 다음 헤더를 읽어들임 */
	cnt = read_tar_header(fd, buf, msg);
	/* TAR EOF가 나온 경우 */
	if (cnt == 0)
	    /* 정상 종료 처리 */
	    return 0;
	else if (cnt < 0)
	    return -1;
    }

    /* 실제 파일의 헤더 전송 */
    if (send_restore_data(pxe, regno, hdr, buf, T_BLOCKSIZE, msg))
	return -1;

    /*
     * 실제 파일의 내용 전송
     */
    size = get_block_size(tar);
    if (is_tar_regfile(tar) && 
	    send_real_data(pxe, regno, hdr, fd, size, msg))
	return -1;

    return 0;
}

/**
 * 복구 대상 파일명의 조건에 일치하는 데이터를 찾아서 에이전트로 전송한다.
 *
 * @param glob - 복구 대상 파일명
 * @param sql - SQL 핸들러
 * @param pxe - PXE 핸들러
 * @param fd - TAR 아카이브 파일 디스크립터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param msg - 에러 발생 시 에러메세지 포인터
 * @return - 0; 에러 발생 시 -1
 * @warning - \a msg 는 사용 후 FREE() 필요
 */
static int
retrieve_data(sql_t *sql, const char *glob, PXE *pxe, int fd, REGNO regno, 
	RestoreHeader *hdr, void **msg)
{
    int ret_val = -1;
    char query[8192];
    sql_res_t *res = NULL;
    sql_row_t row;
    char *escape_glob = NULL;
    off_t offset;
    size_t len;
    char *path = NULL;
    char *name = NULL;

    if (db_escape_string(sql, &escape_glob, glob, strlen(glob)))
    {
	*msg = dup_str("복구 파일 리스트 생성을 위한 Escape String 실패: %s", 
		strerror(errno));
	goto cleanup;
    }

    len = strlen(escape_glob);
    path = escape_glob;

    /* 디렉토리 전체 복원인 경우 */
    if (path[len - 1] == '*' && path[len - 2] == '/')
    {
	char *fullpath = NULL;

	/* 맨 뒤의 slash와 asterisk 제거 */
	path[len - 2] = '\0';

	if ((name = strrchr(path, '/')) == NULL)
	{
	    *msg = dup_str("잘못된 파일 형식입니다.");
	    goto cleanup;
	}

	fullpath = strdup(path);
	if (!fullpath)
	{
	    *msg = dup_str("복구 파일 리스트 생성을 위한 메모리 할당 실패: %s", 
		    strerror(errno));
	    goto cleanup;
	}

	if (name != &path[0])
	    *name++ = '\0';
	/* 대상 경로가 최상위 디렉토리인 '/' 인 경우 */
	else
	{
	    path = "/";
	    name++;
	}

	snprintf(query, sizeof(query), "SELECT offset FROM `records` WHERE "
		"job_id = '%" PRIu64 "' AND ((path = '%s' AND filename = '%s') "
		"OR (path LIKE '%s%%'))", hdr->bk_sid, path, name, fullpath);
	FREE(fullpath);
    }
    /* 파일 또는 디렉토리 지정 복원인 경우 */
    else
    {
	if ((name = strrchr(path, '/')) == NULL)
	{
	    *msg = dup_str("잘못된 파일 형식입니다.");
	    goto cleanup;
	}

	if (name != &path[0])
	    *name++ = '\0';
	/* 대상 경로가 최상위 디렉토리인 '/' 인 경우 */
	else
	{
	    path = "/";
	    name++;
	}

	snprintf(query, sizeof(query), "SELECT offset FROM `records` WHERE "
		"job_id = '%" PRIu64 "' AND path = '%s' AND filename = '%s'", 
		hdr->bk_sid, path, name);
    }

    if (db_query(sql, query, &res))
    {
	*msg = dup_str("복구 파일 리스트 생성을 위한 쿼리 실패: %s", 
		db_errmsg(sql));
	goto cleanup;
    }

    if (db_num_rows(res) == 0)
    {
	*msg = dup_str("매칭되는 파일이 없습니다.");
	goto cleanup;
    }

    while ((row = db_fetch_row(res)))
    {
	errno = 0;
	offset = (off_t) strtoumax(row[0], (char **) NULL, 10);
	if (errno != 0)
	{
	    *msg = dup_str("복구 대상 파일의 오프셋을 계산할 수 없음: %s", 
		    strerror(errno));
	    goto cleanup;
	}

	if (lseek(fd, offset, SEEK_SET) == (off_t) -1)
	{
	    *msg = dup_str("복구 대상 파일의 오프셋으로 이동 할 수 없음: %s", 
		    strerror(errno));
	    goto cleanup;
	}

	if (send_backup_data(pxe, fd, regno, hdr, msg))
	    goto cleanup;
    }

    ret_val = 0; /* success */

cleanup:
    if (res)
	db_free_result(res);
    FREE(escape_glob);

    return ret_val;
}

/**
 * @brief TAR 아카이브에서 특정 파일만 추출해서 전송
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param fd - TAR 파일 디스크립터
 * @param filename - 추출할 파일 및 디렉토리 (\n을 구분자로 나열된 문자열)
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 DBS_ERR_OK,\n
 *  실패 시 DBS_ERR_PEER or DBS_ERR_LOCAL
 */
static int
send_glob_file(PXE *pxe, REGNO regno, RestoreHeader *hdr, 
	int fd, const char *filename, void **msg)
{
    int i;
    int ret_val;
    char **glob = NULL;
    char buf[8192];
    sql_t sql;

    ASSERT(pxe && hdr && filename && msg);

    /* for cleanup routine */
    ret_val = DBS_ERR_LOCAL;

    if (dbms_connect(&sql, msg))
	goto cleanup;

    glob = make_glob_list(filename);
    if (!glob)
    {
	*msg = dup_str("glob 리스트 생성 실패: %s", strerror(errno));
	goto cleanup;
    }

    for (i = 0; glob[i]; i++)
    {
	/* 자신한테 수신된 패킷이 있는지 확인하고 에러 패킷이면 에러 처리 */
	if (check_error_notify(pxe, msg))
	{
	    /* 상대방 호스트에서 에러 발생 */
	    ret_val = DBS_ERR_PEER;
	    goto cleanup;
	}

	/* 제한 시간이 만료되었는지 확인 */
	if (IS_TIME_LIMIT(hdr->options))
	{
	    if (check_timeout(hdr->ltime))
	    {
		*msg = strdup("작업 취소!! 제한시간이 만료되었음.");
		goto cleanup;
	    }
	}

	if (retrieve_data(&sql, glob[i], pxe, fd, regno, hdr, msg))
	    goto cleanup;
    }

    /* TAR EOF 패킷 전송 */
    memset(buf, 0, sizeof(buf));
    if (send_restore_data(pxe, regno, hdr, buf, T_BLOCKSIZE * 2, msg))
	goto cleanup;

    /* Success */
    ret_val = DBS_ERR_OK;

cleanup:
    if (glob)
	free_glob_list(glob);
    db_close(&sql);

    return ret_val;
}

/**
 * @brief 파일 내용을 PEER에게 전송
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param filename - TAR 아카이브에서 추출 할 파일 또는 디렉토리 리스트
 * @param fd - 파일 디스크립터
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
send_file_contents(PXE *pxe, REGNO regno, RestoreHeader *hdr, 
	const char *filename, int fd, void **msg)
{
    int result;

    ASSERT(pxe && hdr && msg && filename);

    /* 압축되었거나 복구할 개별 파일을 지정하지 않은 경우 */
    if (IS_COMPRESS(hdr->options) || strlen(filename) == 0)
	result = send_whole_file(pxe, regno, hdr, fd, msg);
    else
	result = send_glob_file(pxe, regno, hdr, fd, filename, msg);

    return result;
}

/**
 * @brief 복구 메인쓰레드
 * @param arg - RT_ARG 포인터 
 * @return NULL 포인터
 */
void *
restore_main(void *arg)
{
    int fd;
    int result;
    PXE *pxe = NULL;
    RT_ARG *info = NULL;
    void *packet = NULL;
    void *err_msg = NULL;
    char ip[16];
    sql_t sql;
    bool sql_stat = false;
    char *name = NULL;

    /* for cleanup routine */
    fd = -1;

    pxe = pxe_open(arg);
    if (!pxe)
    {
	Log(ERROR, "PXE 핸들러 생성 실패: %s", strerror(errno));
	goto cleanup;
    }

    info = pxe_get_argument(arg);
    ASSERT(info);

    if (dbms_connect(&sql, &err_msg))
    {
	ret_restore_error(pxe, info->regno, &info->header, "%s", err_msg);
	report_result(info->regno, FAIL_STAT, "%s", err_msg);
	goto cleanup;
    }
    sql_stat = true;

    name = get_agent_name(&sql, info->header.bk_sid);
    if (!name || *name == '\0')
    {
	ret_restore_error(pxe, info->regno, &info->header, 
		"에이전트 이름을 확인 할 수 없습니다.");
	report_result(info->regno, FAIL_STAT, "에이전트 이름을 확인 할 수 없습니다.");
	goto cleanup;
    }

    /* Peer IP와 담당 쓰레드 아이디 설정 */
    convert_ip_to_str(info->header.ip, ip, sizeof(ip));
    pxe_set_destination(pxe, ip, info->src);

    fd = open_backup_file(name, info->header.bk_sid, info->header.options, &err_msg);
    if (fd < 0)
    {
	ret_restore_error(pxe, info->regno, &info->header, "%s", err_msg);
	report_result(info->regno, FAIL_STAT, "%s", err_msg);
	goto cleanup;
    }

    result = send_file_contents(pxe, info->regno, 
	    &info->header, info->data, fd, &err_msg);
    if (result != DBS_ERR_OK)
    {
	/* 로컬호스트에서 발생한 에러만 상대방 호스트에 통보 */
	if (result == DBS_ERR_LOCAL)	
	    ret_restore_error(pxe, info->regno, &info->header, "%s", err_msg);
	report_result(info->regno, FAIL_STAT, "%s", err_msg);
	goto cleanup;
    }

    /* 백업파일 내용 전송이 끝난 후에 상대방에서 처리결과(에러)를 확인하기위해
     * 지정된 시간동안 대기한다. 상대방이 복구 작업을 완료했을 경우의 통보는
     * 코어 쓰레드로 전송되기 때문에 여기서 신경쓸 필요는 없다. 그러나 에러통보는
     * 코어로 전송되지 않고 이 쓰레드로 전송되기 때문에 백업파일의 내용을 상대방에게
     * 모두 전송했다고해서 바로 종료하게 되면 상대방에서 발생한 에러 통보를 받지
     * 못하는 경우가 생길 수 있다. 따라서 일정시간 대기한 후에 수신된 에러통보가
     * 있는지 확인하도록 한다.
     */
    if (pxe_read_timedwait(pxe, &packet, FINISH_WAIT_DELAY) >= 0)
    {
	if (is_restore_error(packet, &err_msg))
	{
	    report_result(info->regno, FAIL_STAT, "%s", err_msg);
	    pxe_free(packet);
	    goto cleanup;
	}
	pxe_free(packet);
    }

cleanup:
    if (sql_stat)
	db_close(&sql);
    if (fd >= 0)
	close(fd);
    if (pxe)
	pxe_close(pxe);
    FREE(err_msg);
    FREE(arg);
    FREE(name);

    return NULL;
}

/**
 * @brief RESTORE_START 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
start_cmd(PXE *pxe, REGNO regno, RestoreHeader *hdr, 
	void *data, void *sql, void **msg)
{
    int ret_val;
    size_t size;
    int result;
    void *ptr = NULL;

    ASSERT(pxe && hdr && data && msg);

    /* 기본 리턴값 */
    ret_val = -1;

    Log(INFO, "%" PRIu64 "번 백업에 대한 %" PRIu64 "번 복구 시작", 
	    hdr->bk_sid, regno);
    size = create_restore_arg(&ptr, regno, 
	    pxe_get_src_thread(pxe), hdr, data, hdr->len);
    if (size == 0)
    {
	*msg = dup_str("복구 쓰레드용 인자 생성 실패: %s", 
		strerror(errno));
	report_result(regno, FAIL_STAT, "%s", *msg);
	goto cleanup;
    }

    result = pxe_load_thread(PXE_AUTO_ASSIGN, &restore_main, ptr, size, 
	    PXE_DETACH_THREAD);
    if (result < 0)
    {
	*msg = dup_str("복구 메인 쓰레드 로딩 실패: %s", 
		strerror(errno));
	report_result(regno, FAIL_STAT, "%s", *msg);
	goto cleanup;
    }

    ret_val = 0;

cleanup:
    FREE(ptr);

    return ret_val;
}
/*
 * 여기까지가 MySQL 라이브러리가 링크된 경우에만 컴파일 되는 함수들
 */
#endif

/**
 * @brief RESTORE_END 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
end_cmd(PXE *pxe, REGNO regno, RestoreHeader *hdr, 
	void *data, void *sql, void **msg)
{
    ASSERT(pxe && hdr && data && msg);

#ifdef HAVE_LIBMYSQLCLIENT
    report_result(regno, DONE_STAT, NULL);
#endif
    Log(INFO, "%" PRIu64 "번 복구 종료", regno);
    return 0; 
}

/**
 * @brief RESTORE_ERROR 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
error_cmd(PXE *pxe, REGNO regno, RestoreHeader *hdr, 
	void *data, void *sql, void **msg)
{
    ASSERT(pxe && hdr && data && msg);

    *msg = dup_str("[원격호스트에서 전송된 에러 메세지] %s", data);
#ifdef HAVE_LIBMYSQLCLIENT
    report_result(regno, ERROR_STAT, "%s", *msg);
#endif
    return 0;
}

/**
 * @brief 네트웍으로 수신한 패킷에서 복구모듈 헤더 파싱
 * @param hdr - 파싱된 복구모듈 헤더 정보를 저장할 구조체 포인터
 * @param data - 네트웍으로 수신한 패킷 (바이트 스트림 형태)
 * @return 없음
 */
void
parse_restore_header(RestoreHeader *hdr, void *data)
{
    REGNO *bk_sid = (REGNO *) data;
    uint8_t *ip = (uint8_t *) ((char *) bk_sid + sizeof(REGNO));
    int32_t *cmd = (int32_t *) ((char *) ip + sizeof(uint8_t) * 4);
    TIME *ltime = (TIME *) ((char *) cmd + sizeof(int32_t));
    int32_t *options = (int32_t *) ((char *) ltime + sizeof(TIME));
    uint32_t *len = (uint32_t *) ((char *) options + sizeof(int32_t));

    hdr->bk_sid = ntohll(*bk_sid);
    memcpy(hdr->ip, ip, sizeof(hdr->ip));
    hdr->cmd = ntohl(*cmd);
    hdr->ltime = ntohl(*ltime);
    hdr->options = ntohl(*options);
    hdr->len = ntohl(*len);
}

/**
 * @brief 복구모듈 스위치보드
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param data - 수신한 복구모듈 패킷 (BK 쓰레드 헤더 제외)
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  성공 시 처리한 명령어의 종류,\n
 *  실패 시 -1
 */
int
restore_cmd(PXE *pxe, REGNO regno, void *data, void *sql, void **msg)
{
    int i;
    int result;
    void *data_ptr = NULL;
    RestoreHeader hdr;

    ASSERT(pxe && data && msg);

    parse_restore_header(&hdr, data);
    data_ptr = (char *) data + RESTORE_HEADER_SIZE;

    /* Switchboard */
    for (i = 0; i < ARRAY_SIZE(clist); i++)
    {
	if (clist[i].cmd == hdr.cmd)
	{
	    result = (*clist[i].func)(pxe, regno, &hdr, data_ptr, sql, msg);
	    if (result < 0)
		return -1;
	    else
		break;
	}
    }
    /* 잘못된 명령어가 입력되면 ASSERT */
    ASSERT(i != ARRAY_SIZE(clist));

    return clist[i].cmd;
}

/**
 * @brief 복구모듈 헤더 생성 (네트웍 전송을 위해 바이트 스트림 형태로 생성)
 * @param ptr - 복구모듈 헤더를 저장할 메모리 포인터
 * @param header - 생성할 복구모듈 헤더 정보
 * @param data_len - 실제 데이터의 사이즈
 * @return 없음
 */
static void
create_restore_header(void *ptr, RestoreHeader *header, unsigned int data_len)
{
    REGNO *bk_sid = (REGNO *) ptr;
    uint8_t *ip = (uint8_t *) ((char *) bk_sid + sizeof(REGNO));
    int32_t *cmd = (int32_t *) ((char *) ip + sizeof(uint8_t) * 4);
    TIME *ltime = (TIME *) ((char *) cmd + sizeof(int32_t));
    int32_t *options = (int32_t *) ((char *) ltime + sizeof(TIME));
    uint32_t *len = (uint32_t *) ((char *) options + sizeof(int32_t));

    *bk_sid = htonll(header->bk_sid);
    memcpy(ip, header->ip, sizeof(header->ip));
    *cmd = htonl(header->cmd);
    *ltime = htonl(header->ltime);
    *options = htonl(header->options);
    *len = htonl(data_len);
}

/**
 * @brief 복구모듈 패킷 생성
 * @param ptr - (OUT) 생성된 복구모듈 패킷 포인터 (호출자 사용 후 FREE)
 * @param regno - 등록번호
 * @param header - 복구모듈 헤더 포인터
 * @param data - 패킷에 적재할 데이터
 * @param len - \a data 길이
 * @return
 *  성공 시 생성된 패킷의 사이즈,\n
 *  실패 시 0
 *
 * 복구모듈 패킷 = BK 쓰레드 헤더 + 복구모듈 헤더 + 데이터 + null char
 */
size_t
create_restore_packet(void **ptr, REGNO regno, 
	RestoreHeader *header, void *data, unsigned int len)
{
    size_t total_len;
    void *data_ptr = NULL;

    ASSERT(ptr && header);

    /* +1 is null character */
    total_len = BK_HEADER_SIZE + RESTORE_HEADER_SIZE + len + 1;
    *ptr = malloc(total_len);
    if (!*ptr)
	return 0;

    /* BK 쓰레드 헤더 생성(바이트 스트림 형태) */
    create_bk_header(*ptr, regno, RESTORE);

    /* 복구모듈 헤더 생성(바이트 스트림 형태) */
    create_restore_header((char *) *ptr + BK_HEADER_SIZE, header, len);

    /* 데이터 복사 */
    if (len > 0 && data)
    {
	data_ptr = (char *) *ptr + BK_HEADER_SIZE + RESTORE_HEADER_SIZE;
	memcpy(data_ptr, data, len);
	((unsigned char *) data_ptr)[len] = '\0';
    }

    return total_len;
}

/**
 * @brief Peer host에 에러메세지 전송 / 로컬호스트에 로그 기록
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param header - 복구모듈 헤더 포인터
 * @param format - 메세지 포멧
 * @param ... - \a format 의 가변인자
 * @return 없음
 */
void
ret_restore_error(PXE *pxe, REGNO regno, const RestoreHeader *header, 
	const char *format, ...)
{
    size_t len;
    va_list ap;
    void *ptr = NULL;
    char msg[8192];
    RestoreHeader hdr;

    ASSERT(pxe && header && format);

    va_start(ap, format);
    vsnprintf(msg, sizeof(msg), format, ap);
    va_end(ap);

    memcpy(&hdr, header, sizeof(RestoreHeader));
    hdr.cmd = RESTORE_ERROR;
    hdr.len = strlen(msg);

    Log(ERROR, "%s", msg);
    len = create_restore_packet(&ptr, regno, &hdr, msg, hdr.len);
    if (len == 0)
    {
	Log(ERROR, "원격호스트로 전송할 에러메세지 생성 실패: %s", 
		strerror(errno));
	goto cleanup;
    }
    if (pxe_write(pxe, ptr, len))
    {
	Log(ERROR, "원격호스트로 에러메세지를 전송 할 수 없음: %s", 
		strerror(errno));
	goto cleanup;
    }

cleanup:
    FREE(ptr);
}

/**
 * @brief RESTORE_OPEN 명령어 처리
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param data - 실제 데이터 포인터
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 리턴 메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  에러 발생 시 -1
 */
static int 
open_cmd(PXE *pxe, REGNO regno, RestoreHeader *hdr, 
	void *data, void *sql, void **msg)
{
    int ret_val;
    int result;
    size_t len;
    void *info = NULL;

    ASSERT(pxe && hdr && data && msg);

    /* 기본 리턴값 */
    ret_val = -1;

    Log(INFO, "%" PRIu64 "번 백업에 대한 %" PRIu64 "번 복구 요청", 
	    hdr->bk_sid, regno);
    len = create_ltar_restore_info(&info, regno, hdr, data, hdr->len);
    if (len == 0)
    {
	*msg = dup_str("TAR 라이브러리에 전달할 인자 생성 실패: %s", 
		strerror(errno));
	ret_restore_error(pxe, regno, hdr, "%s", *msg);
	goto cleanup;
    }
    
    result = pxe_load_thread(PXE_AUTO_ASSIGN, &ltar_restore, info, len, 
	    PXE_DETACH_THREAD);
    if (result < 0)
    {
	*msg = dup_str("TAR 라이브러리 메인 쓰레드 로딩 실패: %s", 
		strerror(errno));
	ret_restore_error(pxe, regno, hdr, "%s", *msg);
	goto cleanup;
    }

    ret_val = 0;

cleanup:
    FREE(info);

    return ret_val;
}
