/**
 * @file db_interface.c
 * @brief DBMS Interface
 */

/*
 * DBMS Interface
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태(superkkt@sds.co.kr)


 * REFERENCE


mysql> desc jobs;
+----------+-------------------------------+------+-----+---------+----------------+
| Field    | Type                          | Null | Key | Default | Extra          |
+----------+-------------------------------+------+-----+---------+----------------+
| id       | bigint(20) unsigned           | NO   | PRI | NULL    | auto_increment |
| schedule_id | bigint(20) unsigned        | NO   | MUL |         |                |
| gid      | bigint(20) unsigned           | NO   |     |         |                |
| bk_sid   | bigint(20) unsigned           | YES  |     | NULL    |                |
| master   | varchar(255)                  | NO   | MUL |         |                |
| agent    | varchar(255)                  | NO   |     |         |                |
| argument | mediumtext                    | NO   |     |         |                |
| wk_type  | enum('B','R','D')             | NO   |     |         |                |
| options  | int(11)                       | YES  |     | 0       |                |
| ip       | varchar(16)                   | NO   |     |         |                |
| ldate    | datetime                      | YES  |     | NULL    |                |
| sdate    | datetime                      | YES  |     | NULL    |                |
| edate    | datetime                      | YES  |     | NULL    |                |
| status   | enum('W','R','D','F','E','C') | NO   |     |         |                |
| msg      | text                          | YES  |     | NULL    |                |
| email    | varchar(255)                  | NO   |     |         |                |
| wkid     | varchar(16)                   | YES  |     | system  |                |
| ts       | datetime                      | NO   |     |         |                |
+----------+-------------------------------+------+-----+---------+----------------+
18 rows in set (0.00 sec)

mysql> desc records;
+----------+---------------------+------+-----+---------+----------------+
| Field    | Type                | Null | Key | Default | Extra          |
+----------+---------------------+------+-----+---------+----------------+
| id       | bigint(20) unsigned | NO   | PRI | NULL    | auto_increment |
| job_id   | bigint(20) unsigned | NO   | MUL |         |                |
| path     | varchar(8192)       | NO   | MUL |         |                |
| filename | varchar(256)        | NO   | MUL |         |                |
| size     | bigint(20) unsigned | NO   |     |         |                |
| mtime    | varchar(30)         | NO   |     |         |                |
| offset   | bigint(20) unsigned | NO   |     |         |                |
| typeflag | char(1)             | NO   |     |         |                |
+----------+---------------------+------+-----+---------+----------------+
8 rows in set (0.00 sec)

mysql> desc schedules;
+---------------+---------------------+------+-----+---------------------+----------------+
| Field         | Type                | Null | Key | Default             | Extra          |
+---------------+---------------------+------+-----+---------------------+----------------+
| id            | bigint(20) unsigned | NO   | PRI | NULL                | auto_increment |
| bk_sid        | bigint(20) unsigned | YES  |     | NULL                |                |
| master        | varchar(255)        | NO   | MUL |                     |                |
| agent         | varchar(255)        | NO   |     |                     |                |
| datetimes     | char(19)            | YES  |     | 0000-00-00 00:00:00 |                |
| weekday       | int(11)             | YES  |     | 0                   |                |
| argument      | mediumtext          | NO   |     |                     |                |
| wk_type       | enum('B','R','D')   | NO   |     |                     |                |
| options       | int(11)             | YES  |     | 0                   |                |
| ip            | varchar(16)         | NO   |     |                     |                |
| master_expire | int(11)             | YES  |     | 0                   |                |
| agent_expire  | int(11)             | YES  |     | 0                   |                |
| ldate         | int(5)              | YES  |     | NULL                |                |
| email         | varchar(255)        | NO   |     |                     |                |
| wkid          | varchar(16)         | YES  |     | system              |                |
| ts            | datetime            | NO   |     |                     |                |
+---------------+---------------------+------+-----+---------------------+----------------+
16 rows in set (0.00 sec)

 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <errno.h>
#include <stdarg.h>
#include <inttypes.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>

#include <pxe.h>
#include "db_interface.h"
#include "dbs.h"
#include "core.h"
#include "core_backup.h"
#include "core_restore.h"
#include "core_delete.h"
#include "types.h"
#include "support.h"
#include "configure.h"
#include "../lib/slib.h"

/**
 * JOB 테이블 스캔 주기 (초)
 *
 * 여기 지정된 스캔 주기는 스캐너가 이 주기마다 JOB 테이블을 스캔하는것이 아니라
 * 각각의 스캐닝 사이에 얼만큼의 DELAY를 넣어줄지를 지정하는 것이다. 각 스캐닝에
 * 소요되는 시간은 JOB 테이블에 등록된 ROW의 수에 따라 달라진다.
 */
#define DEFAULT_SCAN_DELAY 1

/**
 * Job DB SQL 쿼리 필드 리스트
 *
 * 쿼리에서 리턴된 결과를 사용할때 이 필드 리스트에 있는 멤버 네임을
 * 사용해서 접근한다. 따라서 쿼리문이 변경되는 경우 멤버 리스트도 
 * 같이 변경해줘야 한다.
 */
typedef enum
{
    SID,         /**< 등록번호 */
    SC_SID,      /**< 스케줄러 등록번호 */
    BK_SID,      /**< 복구 작업 시 대상 백업 등록번호 */
    ARGUMENT,    /**< 대상 파일 및 디렉토리 */
    WK_TYPE,     /**< 작업 타입(백업, 복구, 삭제) */
    OPTIONS,     /**< 작업 옵션 */
    IP,          /**< 대상 IP */
    LDATE,       /**< 제한 시간 */
    EMAIL        /**< 에러 발생시 공지할 이메일 주소 */
} DB_FieldName;

/**
 * 현재 실행중인 요청의 갯수 
 */
static int request_count = 0;

/**
 * 결과 레포팅용 SQL 핸들러
 */
static sql_t report_sql;

/**
 * 로컬 전역변수에 접근하기 위해 사용하는 뮤텍스
 */
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * @brief 새로운 요청이 있는지 확인
 * @param sql - SQL 핸들러
 * @param res - (OUT) 쿼리 결과 (호출자가 사용 후 FREE)
 * @return
 *  성공 시 새로운 요청의 갯수 (>= 0)\n
 *  실패 시 -1
 */
static int
db_scan(sql_t *sql, sql_res_t **res)
{
    size_t cnt;
    char *alias = NULL;
    char *escape_alias = NULL;
    char query[2048];

    ASSERT(sql && res);

    alias = get_config("ServerAlias");
    if (!alias)
    {
	Log(ERROR, "'ServerAlias' 파라메터가 존재하지 않음");
	return -1;
    }
    if (db_escape_string(sql, &escape_alias, alias, strlen(alias)) == -1)
    {
	FREE(alias);
	return -1;
    }

    /* Select filed 순서가 바뀌면 DB_FieldName 타입도 수정요망 */
    snprintf(query, sizeof(query), 
	    "SELECT id, schedule_id, bk_sid, argument, wk_type, options, ip, "
	    "ldate, email FROM jobs WHERE master = '%s' AND status = '%c' "
	    "ORDER BY id ASC", escape_alias, WAIT_STAT);
    if (db_query(sql, query, res))
    {
	Log(ERROR, "신규 작업 스캔을 위한 쿼리 실패: %s", db_errmsg(sql));
	FREE(alias);
	FREE(escape_alias);
	return -1;
    }

    cnt = db_num_rows(*res);
    if (cnt == 0)
	db_free_result(*res);

    FREE(alias);
    FREE(escape_alias);
    return cnt;
}

/**
 * @brief 에러 발생시 공지 메일을 전송할 관리자 메일 주소 확인
 * @param sql - SQL 핸들러
 * @param regno - 관리자 메일 주소를 확인할 작업번호
 * @return
 *  성공 시 관리자 이메일 문자열 (사용 후 FREE),\n
 *  실패 시 NULL
 */
static char *
get_admin_email(sql_t *sql, REGNO regno)
{
    char *email = NULL;
    char query[8192];
    sql_row_t row;
    sql_res_t *res = NULL;

    snprintf(query, sizeof(query), "SELECT email FROM job WHERE id = '%" 
	    PRIu64 "'", regno);
    if (db_query(sql, query, &res))
	return NULL;
    if ((row = db_fetch_row(res)))
	email = strdup(row[0]);

    db_free_result(res);
    return email;
}

/**
 * @brief 관리자에게 에러 상황 통보
 * @param sql - SQL 핸들러
 * @param regno - 등록번호
 * @param status - 상태
 * @param msg - 메세지
 * @return 없음
 */
static void
send_alert(sql_t *sql, REGNO regno, unsigned char status, const char *msg)
{
    int result;
    FILE *fp = NULL;
    char buf[8192];
    char *email = NULL;
    char *sendmail = NULL;

    sendmail = get_config("Sendmail");
    if (!sendmail) 
    {
	Log(ERROR, "'Sendmail' 파라메터가 존재하지 않음");
	goto cleanup;
    }

    if (access(sendmail, F_OK) < 0)
    {
	Log(ERROR, "센드메일 명령어(%s)에 접근 할 수 없음: %s", 
		sendmail, strerror(errno));
	goto cleanup;
    }

    email = get_admin_email(sql, regno);
    if (!email)
    {
	Log(ERROR, "관리자 이메일 주소를 확인 할 수 없음");
	goto cleanup;
    }

    snprintf(buf, sizeof(buf), "%s %s", sendmail, email);
    fp = popen(buf, "w");
    if (!fp)
    {
	Log(ERROR, "센드메일 명령어 실행 실패: %s", buf);
	goto cleanup;
    }

    snprintf(buf, sizeof(buf), "Subject: [Direct Backup System] Error "
	    "notification mail\nFailed to backup, restore or delete request "
	    "for %" PRIu64 "\n.\n", regno);
    fputs(buf, fp);
    result = pclose(fp);
    if (result == -1)
    {
	Log(ERROR, "%s로 이메일 발송 실패", email);
	goto cleanup;
    }

cleanup:
    FREE(sendmail);
    FREE(email);
}

/** 
 * @brief 요청 처리 상황 업데이트
 * @param sql - SQL 핸들러 포인터
 * @param regno - 등록번호
 * @param status - 처리 상태
 * @param msg - 에러 메시지
 * @return 없음
 */
static void
update_status(sql_t *sql, REGNO regno, 
	unsigned char status, const char *msg)
{
    char where[2048];
    char query[8192];

    ASSERT(sql);

    snprintf(query, sizeof(query), "UPDATE jobs SET status = '%c'", status); 

    /* 
     * 에러 메세지 추가 
     */
    if (msg)
    {
	char buf[8192];
	char *escape_msg = NULL;
	
	db_escape_string(sql, &escape_msg, msg, strlen(msg));
	snprintf(buf, sizeof(buf), ", msg = '%s'", escape_msg);
	strncat(query, buf, sizeof(query) - strlen(query));
	FREE(escape_msg);
    }

    /* 
     * 처리 시간 추가 
     */
    if (status == RUN_STAT)
    {
	strncat(query, ", sdate = now()", sizeof(query) - strlen(query));
    }
    else if (status == DONE_STAT || 
	    status == FAIL_STAT ||
	    status == ERROR_STAT)
    {
	strncat(query, ", edate = now()", sizeof(query) - strlen(query));
    }
    else
    {
	/* Never reached */
	Log(ERROR, "[ASSERT] 알 수 없는 상태 지시자 (RUN, DONE, FAIL, ERROR 중 "
		"하나만 사용 가능)");
	abort();
    }

    /* WHERE 구문 추가 (DONE/FAIL/ERROR 상태에서는 상태 변경 불가) */
    snprintf(where, sizeof(where), " WHERE id = '%" PRIu64 "' AND status NOT "
	    "IN('%c', '%c', '%c')", regno, DONE_STAT, FAIL_STAT, ERROR_STAT);
    strncat(query, where, sizeof(query) - strlen(query));

    if (db_query(sql, query, NULL))
	Log(ERROR, "작업 상태 업데이트 실패: %s", db_errmsg(sql));
    else
	Log(DEBUG, "Updated jobs status: %s", query);

    if (status == FAIL_STAT || status == ERROR_STAT)
	send_alert(sql, regno, status, msg);
}

/**
 * @brief 이전 작업의 실행 시간을 확인
 * @param sql - SQL 핸들러 포인터
 * @param query - 작업 시점 확인을 위해 실행할 쿼리문
 * @param hdr - 백업모듈 헤더 포인터
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
get_previous_time(sql_t *sql, const char *query, BackupHeader *hdr, void **msg)
{
    int ret_val;
    size_t cnt;
    time_t tm;
    sql_row_t row;
    sql_res_t *res = NULL;

    /* for cleanup routine */
    ret_val = -1;

    if (db_query(sql, query, &res))
    {
	*msg = strdup(db_errmsg(sql));
	goto cleanup;
    }
    cnt = db_num_rows(res);
    if (cnt == 0)
    {
	*msg = strdup("마지막 백업에 대한 정보를 찾을 수 없음");
	goto cleanup;
    }

    if ((row = db_fetch_row(res)))
    {
	if (convert_time_to_epoch(row[0], "%Y-%m-%d %H:%M:%S", &tm))
	{
	    *msg = dup_str("현재 시간을 유닉스 타임스탬프로 변환 할 수 없음");
	    goto cleanup;
	}
	hdr->mtime = (TIME) tm;

	/* Success */
	ret_val = 0;
    }

cleanup:
    if (res)
	db_free_result(res);

    return ret_val;
}

/**
 * @brief 마지막 백업 작업을 찾아서 Incremental 시점을 지정
 * @param sql - SQL 핸들러 포인터
 * @param arg - 백업 대상 문자열 포인터
 * @param ip - 대상 IP
 * @param hdr - 백업모듈 헤더 포인터
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
set_incremental_time(sql_t *sql, const char *arg, const char *ip,
	BackupHeader *hdr, void **msg)
{
    char query[8192];
    char *escape_ip = NULL;
    char *escape_arg = NULL;

    if (db_escape_string(sql, &escape_ip, ip, strlen(ip)))
	return -1;
    if (db_escape_string(sql, &escape_arg, arg, strlen(arg)))
	return -1;

    /* 마지막 백업(Differential을 제외한 나머지 작업) 검색 (백업 대상이 정확하게
     * 일치하는 엔트리만 검색) 
     */
    snprintf(query, sizeof(query), "SELECT ts FROM jobs WHERE ip = '%s' AND "
	    "argument = '%s' AND wk_type = 'B' AND (options & %d) = 0 AND "
	    "status = '%c' ORDER BY id DESC LIMIT 1", escape_ip, escape_arg, 
	    DIFFERENTIAL, DONE_STAT);
    FREE(escape_ip);
    FREE(escape_arg);

    return get_previous_time(sql, query, hdr, msg);
}

/**
 * @brief 마지막 풀 백업 작업을 찾아서 Differential 시점을 지정
 * @param sql - SQL 핸들러 포인터
 * @param arg - 백업 대상 문자열 포인터
 * @param ip - 대상 IP
 * @param hdr - 백업모듈 헤더 포인터
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
set_differential_time(sql_t *sql, const char *arg, const char *ip,
	BackupHeader *hdr, void **msg)
{
    char query[8192];
    char *escape_ip = NULL;
    char *escape_arg = NULL;

    if (db_escape_string(sql, &escape_ip, ip, strlen(ip)))
	return -1;
    if (db_escape_string(sql, &escape_arg, arg, strlen(arg)))
	return -1;

    /* 마지막 풀 백업 작업을 검색 (백업 대상이 정확하게 일치하는 엔트리만 검색) 
     */
    snprintf(query, sizeof(query), "SELECT ts FROM jobs WHERE ip = '%s' AND "
	    "argument = '%s' AND wk_type = 'B' AND (options & %d) = 0 AND "
	    "status = '%c' ORDER BY id DESC LIMIT 1", escape_ip, escape_arg,
	    INCREMENTAL | DIFFERENTIAL, DONE_STAT);
    FREE(escape_ip);
    FREE(escape_arg);

    return get_previous_time(sql, query, hdr, msg);
}

/**
 * @brief 백업모듈 패킷 생성
 * @param sql - SQL 핸들러 포인터
 * @param ptr - (OUT) 생성된 백업모듈 패킷을 저장할 포인터 (호출자 사용 후 FREE)
 * @param row - JOB 테이블 SQL row 포인터
 * @param msg - (OUT) 에러 발생 시 에러 메세지 포인터 (호출자가 사용 후 NORMAL FREE)
 * @return
 *  성공 시 생성한 백업모듈 패킷의 사이즈,\n
 *  실패 시 0
 */
static size_t
make_backup(sql_t *sql, void **ptr, sql_row_t row, void **msg)
{
    time_t tm;
    int options;
    REGNO regno;
    BackupHeader header;

    ASSERT(sql && ptr && row);

    /* 사용자 입력값 유효성 체크 */
    if (!row[OPTIONS] || !row[SID] || !row[IP])
    {
	*msg = strdup("옵션, 등록번호, IP 중 일부가 데이터베이스에 입력되지 않았음");
	return 0;
    }

    options = atoi(row[OPTIONS]);
    regno = (REGNO) strtoll(row[SID], (char **) NULL, 10);

    /*
     * 백업모듈 헤더 작성
     */
    memset(&header, 0, sizeof(BackupHeader));

    if (IS_INCREMENTAL(options))
    {
	if (set_incremental_time(sql, row[ARGUMENT], row[IP], &header, msg))
	    return 0;
    }
    else if (IS_DIFFERENTIAL(options))
    {
	if (set_differential_time(sql, row[ARGUMENT], row[IP], &header, msg))
	    return 0;
    }

    /* 제한시간 설정 */
    if (IS_TIME_LIMIT(options) && row[LDATE])
    {
	if (convert_time_to_epoch(row[LDATE], "%Y-%m-%d %H:%M:%S", &tm))
	{
	    *msg = dup_str("현재 시간을 유닉스 타임스탬프로 변환 할 수 없음");
	    return 0;
	}
    }
    else 
	tm = 0;

    if (convert_ip_to_int(row[IP], header.ip))
    {
	*msg = dup_str("문자열 IP를 정수형으로 변환 할 수 없음: %s", 
		strerror(errno));
	return 0;
    }
    header.ltime = (TIME) tm;
    header.options = options;
    header.cmd = BACKUP_OPEN;
    header.fd = -1;

    return create_backup_packet(ptr, regno, &header, 
	    row[ARGUMENT], strlen(row[ARGUMENT]));
}

/**
 * @brief 백업 당시 압축을 했었는지 확인
 * @param sql - SQL 핸들러 포인터
 * @param regno - 등록번호
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  압축을 한 경우 1,\n
 *  압축을 하지 않은 경우 0,\n
 *  에러 발생 시 -1
 */
int
is_compressed(sql_t *sql, REGNO regno, void **msg)
{
    int options;
    sql_row_t row;
    sql_res_t *res = NULL;
    char query[2048];

    ASSERT(sql);

    snprintf(query, sizeof(query), "SELECT options FROM jobs WHERE id = '%" 
	    PRIu64 "'", regno);

    if (db_query(sql, query, &res))
    {
	*msg = strdup(db_errmsg(sql));
	return -1;
    }

    if ((row = db_fetch_row(res)))
    {
	int ret_val;

	options = atoi(row[0]);
	if (IS_COMPRESS(options))
	    ret_val = 1;
	else
	    ret_val = 0;

	db_free_result(res);
	return ret_val;
    }

    db_free_result(res);
    *msg = dup_str("%" PRIu64 "번 백업에 대한 정보를 찾을 수 없음", regno);
    return -1;
}

/**
 * @brief 복구모듈 패킷 생성
 * @param sql - SQL 핸들러 포인터
 * @param ptr - (OUT) 생성된 복구모듈 패킷을 저장할 포인터 (호출자 사용 후 FREE)
 * @param row - JOB 테이블 SQL row 포인터
 * @param msg - (OUT) 에러 발생 시 에러 메세지 포인터 (호출자가 사용 후 NORMAL FREE)
 * @return
 *  성공 시 생성한 복구모듈 패킷의 사이즈,\n
 *  실패 시 0
 */
static size_t
make_restore(sql_t *sql, void **ptr, sql_row_t row, void **msg)
{
    time_t tm;
    int result;
    int options;
    REGNO regno;
    REGNO bk_regno;
    RestoreHeader header;

    ASSERT(sql && ptr && row);

    /* 사용자 입력값 유효성 체크 */
    if (!row[OPTIONS] || !row[SID] || !row[IP] || !row[BK_SID])
    {
	*msg = strdup("옵션, 등록번호, IP 중 일부가 데이터베이스에 입력되지 않았음");
	return 0;
    }

    options = atoi(row[OPTIONS]);
    regno = (REGNO) strtoll(row[SID], (char **) NULL, 10);
    
    /* 복구대상 등록번호 */
    bk_regno = (REGNO) strtoll(row[BK_SID], (char **) NULL, 10);

    /*
     * 복구모듈 헤더 작성
     */
    memset(&header, 0, sizeof(RestoreHeader));

    /* 제한시간이 있으면 시간 설정 */
    if (IS_TIME_LIMIT(options) && row[LDATE])
    {
	if (convert_time_to_epoch(row[LDATE], "%Y-%m-%d %H:%M:%S", &tm))
	{
	    *msg = dup_str("현재 시간을 유닉스 타임스탬프로 변환 할 수 없음");
	    return 0;
	}
    }
    else 
	tm = 0;

    if (convert_ip_to_int(row[IP], header.ip))
    {
	*msg = dup_str("문자열 IP를 정수형으로 변환 할 수 없음: %s", 
		strerror(errno));
	return 0;
    }
    header.ltime = (TIME) tm;

    /* regno의 백업 당시 옵션을 확인해서 Compress가 있으면 복구 옵션에도
     * Compress를 추가해줘야만 복구모듈이 로딩될때 압축을 풀 수 있다. */
    result = is_compressed(sql, bk_regno, msg);
    if (result == 1)
	options |= COMPRESS;
    else if (result < 0)
	return 0;

    header.options = options;
    header.bk_sid = bk_regno;
    header.cmd = RESTORE_OPEN;

    return create_restore_packet(ptr, regno, &header, 
	    row[ARGUMENT], strlen(row[ARGUMENT]));
}

/**
 * @brief 해당 등록번호의 작업이 실행된 대상 에이전트의 IP를 확인
 * @param sql - SQL 핸들러 포인터
 * @param regno - 작업번호
 * @return
 *  성공 시 에이전트 IP 문자열 (사용 후 FREE),\n
 *  실패 시 NULL
 */
static char *
get_target_ip(sql_t *sql, REGNO regno)
{
    int cnt;
    char query[8192];
    sql_row_t row;
    sql_res_t *res = NULL;
    char *ip = NULL;

    snprintf(query, sizeof(query), "SELECT ip FROM jobs WHERE id='%" PRIu64 
	    "'", regno);
    if (db_query(sql, query, &res))
	goto cleanup;
    cnt = db_num_rows(res);
    if (cnt == 0)
	goto cleanup;

    if ((row = db_fetch_row(res)))
	ip = strdup(row[0]);

cleanup:
    if (res)
	db_free_result(res);

    return ip;
}

/**
 * @brief 삭제모듈 패킷 생성
 * @param sql - SQL 핸들러 포인터
 * @param ptr - (OUT) 생성된 삭제모듈 패킷을 저장할 포인터 (호출자 사용 후 FREE)
 * @param row - JOB 테이블 SQL row 포인터
 * @param msg - (OUT) 에러 발생 시 에러 메세지 포인터 (호출자가 사용 후 NORMAL FREE)
 * @return
 *  성공 시 생성한 삭제모듈 패킷의 사이즈,\n
 *  실패 시 0
 */
static size_t
make_delete(sql_t *sql, void **ptr, sql_row_t row, void **msg)
{
    int options;
    REGNO regno;
    REGNO bk_regno;
    DeleteHeader header;
    char *target_ip = NULL;

    ASSERT(sql && ptr && row);

    /* 사용자 입력값 유효성 체크 */
    if (!row[OPTIONS] || !row[SID] || !row[BK_SID])
    {
	*msg = strdup("옵션, 등록번호, 삭제 대상 중 일부가 데이터베이스에 "
		"입력되지 않았음");
	return 0;
    }

    options = atoi(row[OPTIONS]);
    regno = (REGNO) strtoll(row[SID], (char **) NULL, 10);
    /* 삭제대상 등록번호 */
    bk_regno = (REGNO) strtoll(row[BK_SID], (char **) NULL, 10);
    /* 삭제대상이 어떤 에이전트에서 백업했는지 확인 */
    target_ip = get_target_ip(sql, bk_regno);
    if (!target_ip)
    {
	*msg = dup_str("삭제 대상이 작업했던 에이전트의 IP를 찾을 수 없음"); 
	return 0;
    }

    /*
     * 삭제모듈 헤더 작성
     */
    memset(&header, 0, sizeof(DeleteHeader));
    if (convert_ip_to_int(target_ip, header.ip))
    {
	*msg = dup_str("문자열 IP를 정수형으로 변환 할 수 없음: %s", 
		strerror(errno));
	FREE(target_ip);
	return 0;
    }
    header.options = options;
    header.bk_sid = bk_regno;
    header.cmd = DELETE_OPEN;

    FREE(target_ip);
    return create_delete_packet(ptr, regno, &header, NULL, 0);
}

/**
 * @brief 요청 분류 / 패킷 생성 / 대상 호스트 설정
 * @param sql - SQL 핸들러 포인터
 * @param pxe - PXE 핸들러 포인터
 * @param row - JOB table의 SQL row
 * @param ptr - (OUT) 대상 호스트로 전송할 패킷 포인터 (호출자가 사용 후 FREE)
 * @param cnt - (OUT) \a ptr 의 사이즈
 * @param err - (OUT) 에러 발생 시 에러 메세지 포인터 (호출자가 사용 후 NORMAL FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
switchboard(sql_t *sql, PXE *pxe, sql_row_t row, 
	void **ptr, ssize_t *cnt, void **err)
{
    ASSERT(sql && pxe && row && ptr && cnt && err);

    if (strcmp(row[WK_TYPE], BACKUP_TYPE) == 0)
    {
	/* 백업모듈 패킷 생성 */
	*cnt = make_backup(sql, ptr, row, err);
	if (*cnt == 0)
	    return -1;

	pxe_set_destination(pxe, PXE_LOCALHOST, BK_THREAD);
    }
    else if (strcmp(row[WK_TYPE], RESTORE_TYPE) == 0)
    {
	/* 복구모듈 패킷 생성 */
	*cnt = make_restore(sql, ptr, row, err);
	if (*cnt == 0)
	    return -1;

	pxe_set_destination(pxe, row[IP], BK_THREAD);
    }
    else if (strcmp(row[WK_TYPE], DELETE_TYPE) == 0)
    {
	/* 삭제모듈 패킷 생성 */
	*cnt = make_delete(sql, ptr, row, err);
	if (*cnt == 0)
	    return -1;

	pxe_set_destination(pxe, PXE_LOCALHOST, BK_THREAD);
    }
    else
    {
	*err = dup_str("유효하지 않은 작업 지시자 (BACKUP, RESTORE, DELETE 중 "
	    "하나만 사용가능)");
	return -1;
    }

    return 0;
}

/**
 * 현재 실행중인 작업의 개수 카운터를 증가시킨다.
 *
 * @param 없음
 * @return 현재 실행중인 작업의 개수 (방금 증가된 작업 포함)
 */
static int
increase_request_counter(void)
{
    int result;
    int cnt;

    if ((result = pthread_mutex_lock(&mutex)))
	Log(FATAL, "뮤텍스 잠금 실패: %s", strerror(result));

    cnt = ++request_count;

    if ((result = pthread_mutex_unlock(&mutex)))
	Log(FATAL, "뮤텍스 해제 실패: %s", strerror(result));

    return cnt;
}

/**
 * 현재 실행중인 작업의 개수 카운터를 감소시킨다.
 *
 * @param 없음
 * @return 현재 실행중인 작업의 개수 (방금 감소된 작업 포함)
 */
static int
decrease_request_counter(void)
{
    int result;
    int cnt;

    if ((result = pthread_mutex_lock(&mutex)))
	Log(FATAL, "뮤텍스 잠금 실패: %s", strerror(result));

    cnt = --request_count;

    if ((result = pthread_mutex_unlock(&mutex)))
	Log(FATAL, "뮤텍스 해제 실패: %s", strerror(result));

    return cnt;
}

/**
 * 현재 실행중인 작업의 개수 카운터를 반환한다.
 *
 * @param 없음
 * @return 현재 실행중인 작업의 개수
 */
static int
get_request_counter(void)
{
    int result;
    int cnt;

    if ((result = pthread_mutex_lock(&mutex)))
	Log(FATAL, "뮤텍스 잠금 실패: %s", strerror(result));

    cnt = request_count;

    if ((result = pthread_mutex_unlock(&mutex)))
	Log(FATAL, "뮤텍스 해제 실패: %s", strerror(result));

    return cnt;
}

/**
 * @brief 새로운 요청 처리
 * @param sql - SQL 핸들러 포인터
 * @param pxe - PXE 핸들 포인터
 * @param res - SQL result 포인터
 * @return 없음
 */
static void
process_new_request(sql_t *sql, PXE *pxe, sql_res_t *res)
{
    REGNO regno;
    ssize_t cnt;
    void *ptr = NULL;
    void *err_msg = NULL;
    sql_row_t row;

    ASSERT(sql && pxe && res);

    while ((row = db_fetch_row(res)))
    {
	/* 현재 실행중인 작업이 최대 실행 개수를 넘었으면 해당 작업 SKIP */
	if (get_request_counter() > MAX_REQUEST)
	    continue;

	/* 등록번호 */
	regno = (REGNO) strtoll(row[SID], (char **) NULL, 10);

	/* 요청 분류 */
	if (switchboard(sql, pxe, row, &ptr, &cnt, &err_msg))
	{
	    Log(ERROR, "%s", err_msg);
	    update_status(sql, regno, FAIL_STAT, err_msg);
	    FREE(err_msg);
	    continue;
	}

	/* 명령 전송 */
	pxe_write(pxe, ptr, cnt);	

	update_status(sql, regno, RUN_STAT, NULL);
	(void) increase_request_counter();
	FREE(ptr);
    }
}

/**
 * @brief DB SCAN 주기만큼 sleep
 * @param 없음
 * @return 없음
 */
static void
scan_delay(void)
{
    int sec;
    char *delay = NULL;

    delay = get_config("ScanDelay");
    if (!delay)
    {
	Log(ERROR, "'ScanDelay' 파라메터가 존재하지 않음");
	sleep(DEFAULT_SCAN_DELAY);
	return;
    }

    sec = atoi(delay);
    if (sec <= 0)
	sleep(DEFAULT_SCAN_DELAY);
    else
	sleep(sec);

    FREE(delay);
}

/**
 * @brief 현재 상태가 진행 중으로 되어있는 작업들을 모두 실패로 변경
 * @param sql - SQL 핸들러
 * @return 없음
 */
static void
clean_abnormal_entry(sql_t *sql)
{
    char query[2048];
    char *alias = NULL;
    char *escape_alias = NULL;

    /* 자신에게 할당된 작업들 중에서만 작업을 해야되기 때문에 ServerAlias 값을
     * 가져온다.
     */
    alias = get_config("ServerAlias");
    if (!alias)
    {
	Log(ERROR, "'ServerAlias' 파라메터가 존재하지 않음");
	goto cleanup;
    }
    if (db_escape_string(sql, &escape_alias, alias, strlen(alias)) == -1)
    {
	Log(ERROR, "쿼리문의 예외문자 치환 실패: %s", strerror(errno));
	goto cleanup;
    }

    snprintf(query, sizeof(query), "UPDATE jobs SET status='%c', msg='마스터 "
	    "서버의 비정상 종료로 인해 에러 처리되었음.' WHERE master='%s' "
	    "AND status='%c'", FAIL_STAT, escape_alias, RUN_STAT);
    (void) db_query(sql, query, NULL);

cleanup:
    FREE(alias);
    FREE(escape_alias);
}

/**
 * @brief DBMS 인터페이스 쓰레드 메인
 * @param arg - PXE 쓰레드 로딩시 넘겨받는 파라메터
 * @return NULL 포인터
 */
void *
db_interface(void *arg)
{
    int cnt;
    sql_t sql;
    PXE *pxe = NULL;
    sql_res_t *res = NULL;
    void *err_msg = NULL;

    /* PXE 핸들러 생성 */
    pxe = pxe_open(arg);
    if (!pxe)
	Log(FATAL, "PXE 핸들러 생성 실패: %s", strerror(errno));

    /* 처리할 작업 스캔하는 SQL 핸들러 연결 */
    if (dbms_connect(&sql, &err_msg))
	Log(FATAL, "%s", err_msg);
    
    /* 처리한 작업의 결과를 저장하는 SQL 핸들러 연결 */
    if (dbms_connect(&report_sql, &err_msg))
	Log(FATAL, "%s", err_msg);

    /* 현재 상태가 작업중으로 되어있는 엔트리들을 모두 실패로 변경 
     * 데몬이 새로 시작될때 작업중으로 되어있는 엔트리들은 모두 이전 데몬에서 
     * 정상적으로 종료가 되지 않은 작업들이다.
     */
    clean_abnormal_entry(&sql);

    /* Main loop */
    while (1)
    {
	/* 현재 처리중인 작업이 최대 실행 개수 이하인 경우만 새로운 요청 처리 */
	if (get_request_counter() < MAX_REQUEST)
	{
	    cnt = db_scan(&sql, &res);
	    if (cnt > 0 && res)
	    {
		process_new_request(&sql, pxe, res);
		db_free_result(res);
	    }
	}

	scan_delay();
    }

    /* Never reached */
    pxe_close(pxe);
    db_close(&sql);
    db_close(&report_sql);
    free(arg);

    return NULL;
}

/**
 * @brief DBMS Interface 쓰레드에 처리결과 통보
 * @param regno - 등록번호
 * @param status - 처리결과
 * @param format - 메세지 포맷
 * @param ... - \a format 의 가변인자
 * @return 없음
 */
void
report_result(REGNO regno, unsigned char status, const char *format, ...)
{
    va_list ap;
    char msg[8192];
    int result;

    if (format)
    {
	va_start(ap, format);
	vsnprintf(msg, sizeof(msg), format, ap);
	va_end(ap);
    }

    (void) decrease_request_counter();

    /*
     * update_status()에 전달되는 SQL 핸들러가 여러 쓰레드에 의해 동시에 사용
     * 될 수 있기 때문에 뮤텍스로 임계영역을 만들었음
     */
    if ((result = pthread_mutex_lock(&mutex)))
	Log(FATAL, "뮤텍스 잠금 실패: %s", strerror(result));
    update_status(&report_sql, regno, (char) status, format ? msg : NULL);
    if ((result = pthread_mutex_unlock(&mutex)))
	Log(FATAL, "뮤텍스 해제 실패: %s", strerror(result));
}

/**
 * @brief 백업파일 상세정보를 DB에 기록
 * @param sql - SQL 핸들러 포인터
 * @param regno - 등록번호
 * @param filename - 파일명
 * @param size - 파일 사이즈
 * @param mtime - 수정 시간
 * @param type - 파일 타입
 * @param offset - 백업 파일의 오프셋
 * @return 없음
 */
void
insert_backup_record(sql_t *sql, REGNO regno, const char *filename, 
	int size, long long mtime, unsigned char type, off_t offset)
{
    size_t len;
    char query[8192];
    char escape_filename[PATH_MAX * 2 + 1];
    char *path = NULL;
    char *name = NULL;

    ASSERT(sql && filename);

    (void) mysql_real_escape_string(sql->mysql, escape_filename, filename, 
	    strlen(filename));

    len = strlen(escape_filename);
    /* 디렉토리인 경우 trailing slash 제거 */
    if (escape_filename[len - 1] == '/')
	escape_filename[len - 1] = '\0';

    path = escape_filename;
    if ((name = strrchr(path, '/')) == NULL)
	return;

    if (name != &escape_filename[0])
	*name++ = '\0';
    /* 대상 경로가 최상위 디렉토리인 '/' 인 경우 */
    else
    {
	path = "/";
	name++;
    }

    snprintf(query, sizeof(query),
	    "INSERT INTO `records` (job_id, path, filename, size, mtime, typeflag, "
	    "offset) VALUES ('%" PRIu64 "', '%s', '%s', '%d', '%lld', '%c', '%llu')", 
	    regno, path, name, size, mtime, type, (unsigned long long) offset);
    (void) db_query(sql, query, NULL);
}

/**
 * 백업 대상 에이전트의 이름을 반환한다.
 *
 * @param sql - SQL 핸들러 포인터
 * @param regno - 등록번호
 * @return 백업 대상 에이전트 이름; 에러 발생 시 NULL
 * @warning 리턴값은 호출자가 free해야 한다.
 */
char *
get_agent_name(sql_t *sql, REGNO regno)
{
    char *name = NULL;
    char query[8192];
    sql_row_t row;
    sql_res_t *res = NULL;

    snprintf(query, sizeof(query), "SELECT agent FROM jobs WHERE id = '%" 
	    PRIu64 "'", regno);
    if (db_query(sql, query, &res))
	return NULL;
    if ((row = db_fetch_row(res)))
	name = row[0] ? strdup(row[0]) : NULL;
    db_free_result(res);

    return name;
}

/**
 * @brief DBMS 접속해서 SQL 핸들러 생성
 * @param sql - SQL 핸들러 포인터
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
int
dbms_connect(sql_t *sql, void **msg)
{
    char *host = NULL;
    char *user = NULL;
    char *passwd = NULL;
    char *dbname = NULL;

    ASSERT(sql && msg);

    /* 아래 파라메터들은 로컬 설정 파일에 필수로 등록되어야 하는 파라메터이기
     * 때문에 유효성 검사를 따로 할 필요는 없다.
     */
    host = config_get_value("SQLHost");
    user = config_get_value("SQLUser");
    passwd = config_get_value("SQLPasswd");
    dbname = config_get_value("SQLName");

    if (db_connect(sql, host, user, passwd, dbname, 0, 0) < 0)
    {
	*msg = dup_str("데이터베이스 접속 실패: %s", db_errmsg(sql));
	return -1;
    }

    if (db_set_character_set(sql, "utf8"))
    {
	db_close(sql);
	*msg = dup_str("현재 데이터베이스 연결의 문자셋을 UTF-8로 설정할 수 "
		"없습니다.");
	return -1;
    }

    return 0;
}
