/**
 * @file dbs.c
 * @brief Direct Backup Solution Main
 */

/*
 * Direct Backup Solution Main 
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/resource.h>

#include <pxe.h>
#include "dbs.h"
#include "core.h"
#ifdef HAVE_LIBMYSQLCLIENT
#include "db_interface.h"
#endif
#include "scheduler.h"
#include "socket_interface.h"
#include "dir_navigator.h"
#include "show_status.h"
#include "tar_frontend.h"
#include "configure.h"
#include "support.h"
#include "sysconfdir.h"
#include "../lib/slib.h"

#define DEFAULT_LOG_LEVEL          5
#define DEFAULT_RECONNECT_DELAY    5

/**
 * DBS 구성 쓰레드 리스트 구조체
 */
typedef struct 
{
    int id;                   /**< 쓰레드 식별 번호 */
    void *(*func)(void *);    /**< 쓰레드 함수 포인터 */
    int type;                 /**< 로딩 타입 (COMMON, MASTER, AGENT) */
} ThreadList;

/**
 * DBS 구성 쓰레드 리스트 
 */
static ThreadList tlist[] = {
#ifdef HAVE_LIBMYSQLCLIENT    
    { DI_THREAD, &db_interface, MASTER },
    { SC_THREAD, &scheduler, MASTER },
#endif
    { BK_THREAD, &core, COMMON },
    { SI_THREAD, &socket_interface, MASTER },
    { DN_THREAD, &dir_navigator, AGENT },
    { SS_THREAD, &show_status, COMMON }
};

/*
 * 현재 로딩된 설정파일의 절대경로
 */
static char config_file[PATH_MAX];

/**
 * @brief 시스템 리소스 제한 설정
 * @param 없음
 * @return 없음
 */
static void
set_limit(void)
{
    struct rlimit rlp;

    rlp.rlim_cur = RLIM_INFINITY;
    rlp.rlim_max = RLIM_INFINITY;

    /* Core file size를 시스템 최대치로 설정 */
    if (setrlimit(RLIMIT_CORE, &rlp))
	Log(ERROR, "코어 파일 사이즈 설정 실패: %s", strerror(errno));
}

/**
 * @brief PXE Master 또는 PXE Agent 실행
 * @param type - 현재 실행중인 타입 (마스터 or 에이전트)
 * @return 없음
 */
static void
run_pxe_main(const char *type)
{
    int port;
    char *alias = NULL;
    char *port_ptr = NULL;
    int recon_delay;
    char *delay_ptr = NULL;
    char *ip = NULL;

    port_ptr = get_config("ServerPort");
    if (!port_ptr)
	Log(FATAL, "'ServerPort' 파라메터가 존재하지 않음");
    port = atoi(port_ptr);
    if (port < 0 || port > 65535)
	Log(FATAL, "유효하지 않은 'ServerPort' 파라메터 값: %d", port);

    if (strcasecmp(type, "Master") == 0)
    {
#ifdef HAVE_LIBMYSQLCLIENT
	if (pxe_master(port))
	    Log(FATAL, "PXE 마스터 로딩 실패: %s", strerror(errno));
#else
	Log(FATAL, "소스코드가 MySQL 라이브러리와 함께 컴파일되지 않아서 오직 "
		"에이전트로만 실행 가능함");
#endif
    }
    else if (strcasecmp(type, "Agent") == 0)
    {
	ip = get_config("ServerIP");
	if (!ip)
	    Log(FATAL, "'ServerIP' 파라메터가 존재하지 않음");
	delay_ptr = get_config("ReconnectDelay");
	if (!delay_ptr)
	    Log(FATAL, "'ReconnectDelay' 파라메터가 존재하지 않음");
	recon_delay = atoi(delay_ptr);
	if (recon_delay < 0)
	    recon_delay = DEFAULT_RECONNECT_DELAY;
	alias = get_config("ServerAlias");
	if (!alias)
	    Log(FATAL, "'ServerAlias' 파라메터가 존재하지 않음");
	if (pxe_agent(alias, ip, port, recon_delay))
	    Log(FATAL, "PXE 에이전트 로딩 실패: %s", strerror(errno));
    }
    else
	Log(FATAL, "유효하지 않은 'Type' 파라메터 값: %s", type);

    FREE(ip);
    FREE(port_ptr);
    FREE(delay_ptr);
    FREE(alias);
}

/**
 * @brief PXE 엔진의 에러를 수신해서 로그 기록
 * @param msg - 에러 메세지
 * @return 없음
 */
static void
pxe_logger(const char *msg)
{
    Log(ERROR, "%s", msg);
}

/**
 * @brief Log 기능 활성화
 * @param 없음
 * @return 없음
 */
static void
activate_log(void)
{
    char fullpath[PATH_MAX];
    int log_level;
    char *log_file = NULL;
    char *log_level_ptr = NULL;

    log_file = get_config("LogFile");
    if (!log_file)
    {
	fprintf(stderr, "Not found 'LogFile' parameter\n");
	exit(EXIT_FAILURE);
    }
    log_level_ptr = get_config("LogLevel");
    if (!log_level_ptr)
    {
	fprintf(stderr, "Not found 'LogLevel' parameter\n");
	exit(EXIT_FAILURE);
    }
    log_level = atoi(log_level_ptr);
    if (log_level < FATAL || log_level > DEBUG)
    {
	fprintf(stderr, "Invalid 'LogLevel' parameter value\n");
	exit(EXIT_FAILURE);
    }

#ifdef OS_CYGWIN
    snprintf(fullpath, sizeof(fullpath), "%s", log_file);
#else
    if (log_file[0] != '/')
	snprintf(fullpath, sizeof(fullpath), "/var/log/%s", log_file);
    else
	snprintf(fullpath, sizeof(fullpath), "%s", log_file);
#endif

    if (log_init(fullpath, log_level))
    {
	fprintf(stderr, "Failed to initialize log file '%s': %s\n", 
		fullpath, strerror(errno));
	exit(EXIT_FAILURE);
    }

    FREE(log_level_ptr);
    FREE(log_file);
}

/**
 * @brief PXE 엔진 실행
 * @param 없음
 * @return 없음
 */
static void
launch_pxe(void)
{
    int i;
    int type;
    int result;
    char *mode = NULL;

    if (pxe_init(&pxe_logger))
	Log(FATAL, "PXE 초기화 실패: %s", strerror(errno));

    mode = get_config("Type");
    if (!mode)
	Log(FATAL, "'Type' 파라메터가 존재하지 않음");

    /* 마스터 or 에이전트 실행 */
    run_pxe_main(mode);

    /* 
     * 필수 쓰레드 실행 
     */
    if (strcasecmp(mode, "Master") == 0)
	type = MASTER;
    else
	type = AGENT;

    for (i = 0; i < ARRAY_SIZE(tlist); i++)
    {
	/* 공통 쓰레드이거나 현재 실행타입과 같은 쓰레드만 실행 */
	if (tlist[i].type == COMMON || tlist[i].type == type)
	{
	    result = pxe_load_thread(tlist[i].id, tlist[i].func, 
		    NULL, 0, PXE_DETACH_THREAD);
	    if (result < 0)
		Log(FATAL, "신규 PXE 쓰레드 로딩 실패: %s", strerror(errno));
	}
    }

    FREE(mode);
}

/**
 * @brief 커맨드라인 인자 사용법 출력
 * @param name - 프로그램명
 * @return 없음
 */
static void
show_usage(const char *name)
{
    fprintf(stderr, "usage: %s [options]\n", name);
    fprintf(stderr, " -s, Disable background daemon mode and send all output "
	    "to stderr\n");
    fprintf(stderr, " -c, Specify alternative configuration file\n");
    fprintf(stderr, " -h, Display this help message\n");
}

/**
 * @brief 실행환경 조사
 * @param 없음
 * @return 없음
 *
 * 데몬 실행에 필요한 환경이 갖춰져있지 않으면 프로그램 종료
 */
static void
check_environment(void)
{
    if (geteuid() != 0)
    {
	fprintf(stderr, "You are not a root user. We need super-user "
		"privileges to process some restricted operations.\n");
	exit(EXIT_FAILURE);
    }
}

/**
 * @brief 프로그램을 데몬 모드로 변경
 * @param 없음
 * @return 없음
 */
static void
into_daemon_mode(void)
{
    if (daemonize() < 0)
    {
	fprintf(stderr, "Failed to change program mode to daemon: %s\n", 
		strerror(errno));
	exit(EXIT_FAILURE);
    }
}

/**
 * @brief SIGHUP 시그널 핸들러
 * @param signo - 수신한 시그널 번호
 * @return 없음
 */
static void
sighup_handler(int signo)
{
    void *err_msg = NULL;

    if (config_read(config_file, &config_check, &err_msg))
	Log(ERROR, "설정 파일 리로딩 실패: %s", 
		err_msg ? err_msg : strerror(errno));
    else
	Log(INFO, "설정 파일이 성공적으로 리로드 되었음");

    FREE(err_msg);
}

/**
 * @brief 시그널 핸들러 등록
 * @param 없음
 * @return 없음
 */
static void
signal_register(void)
{
    struct sigaction sa;

    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sa.sa_handler = &sighup_handler;
    if(sigaction(SIGHUP, &sa, NULL) == -1) 
	Log(FATAL, "시그널 핸들러 등록 실패: %s", strerror(errno));

    sa.sa_handler = SIG_IGN;
    if(sigaction(SIGPIPE, &sa, NULL) == -1) 
	Log(FATAL, "시그널 핸들러 등록 실패: %s", strerror(errno));
}

/**
 * @brief PATH를 절대경로로 변환
 * @param path - 변환할 PATH
 * @param len - \a path 의 길이
 * @return 없음
 */
static void
convert_abs_path(char *path, size_t len)
{
    char abs_path[PATH_MAX];

    if (realpath(path, abs_path) == NULL)
    {
	fprintf(stderr, "Failed to convert pathname to absolute path: %s\n", 
		strerror(errno));
	exit(EXIT_FAILURE);
    }
    snprintf(path, len, "%s", abs_path);
}

int
main(int argc, char **argv)
{
    int c;
    int errflg = FALSE;
    int daemon = TRUE;	/* 데몬화가 기본값 */

    memset(config_file, 0, sizeof(config_file));
#ifndef OS_CYGWIN
    check_environment();
#endif
    while ((c = getopt(argc, argv, "shc:")) != -1)
    {
	switch (c)
	{
	    case 's':
		/* Standalone mode */
		daemon = FALSE;
		break;
	    case 'h':
		/* Show help message */
		errflg = TRUE;
		break;
	    case 'c':
		/* Specify configuration file */
		snprintf(config_file, sizeof(config_file), "%s", optarg);
		break;
	    case ':':
		errflg = TRUE;
		fprintf(stderr, "Option -%c requires an operand\n", optopt);
		break;
	    case '?':
		errflg = TRUE;
		fprintf(stderr, "Unrecognized option: -%c\n", optopt);
		break;
	}
    }
    if (errflg)
    {
	show_usage(PACKAGE_TARNAME);
	exit(EXIT_FAILURE);
    }
    if (config_file[0] == '\0')
	/* 커맨드라인 옵션으로 설정파일명이 지정되지 않은 경우 기본값 사용 */
	snprintf(config_file, sizeof(config_file), "%s/%s.conf", SYSCONFDIR, 
		PACKAGE_TARNAME);

    convert_abs_path(config_file, sizeof(config_file));
    init_configure(config_file);
    activate_log();
    signal_register();
    set_limit();
    if (daemon)
	into_daemon_mode();
    launch_pxe();
    init_tar_list();

    Log(INFO, "%s/%s 성공적으로 실행되었음", PACKAGE_NAME, PACKAGE_VERSION);

    /* 필수 쓰레드가 종료할때까지 대기. 그러나 필수 쓰레드는 Critical Error 상태가
     * 아니면 절대로 종료하지 않음. 따라서 실제로는 여기서 무한 대기 
     */
    pxe_sleep();

    /* Never reached */
    return 0;
}
