/**
 * @file dir_navigator.c
 * @brief Directory Navigator
 */

/*
 * Directory Navigator
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <dirent.h>
#include <sys/stat.h>

#include <pxe.h>
#include "dbs.h"
#include "dir_navigator.h"
#include "support.h"
#include "../lib/slib.h" 

#define QUEUE_MAX   256

/**
 * 디렉토리 리스트를 저장하는 내부 큐 
 */
static char *dir_queue[QUEUE_MAX] = {0};

/**
 * 내부 큐 오프셋 
 */
static int dir_queue_offset = 0;

/**
 * 동적메모리로 할당된 리스트에 등록된 엔트리의 갯수 
 */
static size_t list_size = 0;

/**
 * @brief 내부 큐의 내용을 동적 메모리로 플러쉬
 * @param list - (OUT) 생성된 리스트 메모리 포인터 (호출자 사용 후 FREE)
 * @param size - 플러쉬할 사이즈
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
flush_queue(char ***list, size_t size)
{
    int i;
    size_t old_list_size;

    old_list_size = list_size;
    list_size += size;

    /* 처음이면 메모리 할당, 이미 할당되었으면 메모리 확장 
     * +1 is EOF
     */
    if (*list)
	*list = realloc(*list, (list_size + 1) * sizeof(char *));
    else
	*list = malloc((list_size + 1) * sizeof(char *));

    if (!*list)
	return -1;
    else
	(*list)[list_size] = NULL; /* EOF 추가 */

    for (i = 0; i < size; i++)
	(*list)[old_list_size + i] = dir_queue[i];

    /*
     * 초기화 
     */
    dir_queue_offset = 0;
    for (i = 0; i < QUEUE_MAX; i++)
	dir_queue[i] = NULL;

    return 0;
}

/**
 * @brief 디렉토리 리스트에 엔트리 추가
 * @param list - (OUT) 생성된 리스트 메모리 포인터 (호출자 사용 후 FREE)
 * @param name - 추가할 파일 또는 디렉토리 이름
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
add_list(char ***list, const char *name)
{
    if (dir_queue_offset >= QUEUE_MAX)
    {
	if (flush_queue(list, QUEUE_MAX))
	    return -1;
    }

    dir_queue[dir_queue_offset] = strdup(name);
    dir_queue_offset++;

    return 0;
}

/**
 * @brief 디렉토리 리스트 해제
 * @param list - 리스트 메모리 포인터
 * @return 없음
 */
static void
free_list(char ***list)
{
    int i;

    if (!*list)
	return;

    for (i = 0; (*list)[i] != NULL; i++)
	FREE((*list)[i]);

    FREE(*list);
    dir_queue_offset = 0;
    list_size = 0;
}

/**
 * @brief 디렉토리에 포함된 파일 및 디렉토리 리스트 생성
 * @param path - 리스트를 생성할 디렉토리
 * @param list - (OUT) 생성된 리스트 메모리 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 생성된 리스트에 엔트리 갯수,\n
 *  실패 시 -1
 */
static int
gather_dir_info(const char *path, char ***list)
{
    DIR *dir = NULL;
    struct dirent *dp = NULL;

    ASSERT(path);
    
    dir = opendir(path);
    if (!dir)
	return -1;

    /* XXX-
     * 하나의 쓰레드만 실행되기 때문에 현재는 readdir() 함수를 사용했으나 만약
     * 여러 쓰레드가 동시에 실행되도록 수정되면 thread-safe를 위해 readdir_r()
     * 로 변경해야함
     */
    while ((dp = readdir(dir)))
    {
	if (strcmp(dp->d_name, ".") == 0 ||
		strcmp(dp->d_name, "..") == 0)
	    continue;

	if (add_list(list, dp->d_name))
	{
	    closedir(dir);
	    return -1;
	}
    }

    flush_queue(list, dir_queue_offset);
    closedir(dir);

    return list_size;
}

/**
 * @brief 디렉토리 리스트 정렬 
 * @param str1 - 비교할 첫번째 문자열
 * @param str2 - 비교할 두번째 문자열
 * @return
 *  \a str1 이 \a str2 보다 작은 경우 0보다 작은 수,\n
 *  \a str1 과 \a str2 가 같은 경우 0,\n
 *  \a str1 이 \a str2 보다 큰 경우 0보다 큰 수 
 */
static int
sort_list(const void *str1, const void *str2)
{
    return strcasecmp(*((char **) str1), *((char **) str2));
}

/**
 * @brief 지정된 모드의 타입을 리턴
 * @param mode - 타입을 조사할 모드
 * @return 
 *  디렉토리 = D,\n
 *  블럭 디바이스 = B,\n
 *  캐릭터 디바이스 = C,\n
 *  FIFO = F,\n
 *  일반 파일 = R,\n
 *  링크 = L,\n
 *  소켓 = S,\n
 *  알 수 없는 타입 = U
 */
static char
get_file_type(mode_t mode)
{
    char type;

    if (S_ISDIR(mode))
	type = 'D';
    else if (S_ISBLK(mode))
	type = 'B';
    else if (S_ISCHR(mode))
	type = 'C';
    else if (S_ISFIFO(mode))
	type = 'F';
    else if (S_ISREG(mode))
	type = 'R';
    else if (S_ISLNK(mode))
	type = 'L';
    else if (S_ISSOCK(mode))
	type = 'S';
    else
	type = 'U';

    return type;
}

/**
 * @brief 리스트 XML 생성
 * @param path - 리스트를 조사한 디렉토리명
 * @param dir - 디렉토리리 리스트 포인터
 * @param file - 파일 리스트 포인터
 * @return
 *  성공 시 생성된 리스트 포인터,\n
 *  실패 시 NULL 포인터
 */
static char *
make_xml_response(const char *path, const char *dir, const char *file)
{
    char *ptr = NULL;
    char *utf8_str = NULL;
    int len;

    if (encode_str_to_utf8(path, &utf8_str))
	return NULL;

    /* 
     * 현재 디렉토리 경로의 맨 끝이 '/'로 끝나는 경우 제거한다. 이런 경우는
     * 최상위 디렉토리인 '/'를 조회할때만 발생하는데 이것을 제거하지 않고 그대로
     * 전송하면 웹인터페이스에서 '/cygdrive//c/~~~'와 같은 형태로 중복이 된다.
     */
    len = strlen(utf8_str);
    if (utf8_str[len - 1] == '/')
	utf8_str[len - 1] = '\0';

    ptr = dup_str("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<ROOT>\n"
	    "<RET STATUS=\"SUCCESS\"/>\n<LIST PWD=\"%s\">\n", utf8_str);
    FREE(utf8_str);
    if (!ptr)
	return NULL;

    if (dir)
    {
	if (str_append(&ptr, dir))
	    return NULL;
    }

    if (file)
    {
	if (str_append(&ptr, file))
	    return NULL;
    }

    if (str_append(&ptr, "</LIST>\n</ROOT>\r\n"))
	return NULL;

    return ptr;
}

/**
 * HTML 스페셜 태그(&, <, >)를 안전한 문자로 치환한다.
 * 
 * @param in - 원본 문자열
 * @param out - 스페셜 태그가 치환된 문자열
 * @return 0; 실패 시 -1
 * @warning \a out 은 사용 후 FREE()로 해제해야 함
 */
static int
escape_html_tag(const char *in, char **out)
{
    int ret_val = -1;
    char *amp = NULL;
    char *lt = NULL;

    /* & -> &amp; 치환 */
    if (str_replace(in, &amp, "&", "&amp;"))
	goto cleanup;

    /* < -> &lt; 치환 */
    if (str_replace(amp, &lt, "<", "&lt;"))
	goto cleanup;

    /* > -> &gt; 치환 */
    if (str_replace(lt, out, ">", "&gt;"))
	goto cleanup;

    ret_val = 0; /* success */

cleanup:
    FREE(amp);
    FREE(lt);

    return ret_val;
}

/**
 * @brief 리스트에 있는 파일 및 디렉토리의 정보를 수집해서 XML 형태로 전송
 * @param pxe - PXE 핸들러 포인터
 * @param path - 리스트를 조사한 디렉토리명
 * @param list - 파일 및 디렉토리 리스트
 * @return 없음
 */
static void
send_info(PXE *pxe, const char *path, char **list)
{
    int i;
    char type;
    struct stat st;
    char *utf8_str = NULL;
    char *dir = NULL;
    char *file = NULL;
    char *xml = NULL;
    char buf[8192];
    char *escape_str = NULL;

    ASSERT(pxe && list);

    for (i = 0; list[i] != NULL; i++)
    {
	snprintf(buf, sizeof(buf), "%s/%s", path, list[i]);
	if (trim_duplicate(buf, '/') < 0)
	    continue;
	if (trim_duplicate(buf, '\\') < 0)
	    continue;

	/* Device busy error를 감안해서 에러가 발생하면 에러 처리하지 않고
	 * 다음 리스트를 처리하도록 한다.
	 */
	if (lstat(buf, &st))
	    continue;

	/* 
	 * 디렉토리와 파일 리스트 따로 작성 
	 */
	type = get_file_type(st.st_mode);
	if (encode_str_to_utf8(list[i], &utf8_str))
	    continue;
	if (escape_html_tag(utf8_str, &escape_str))
	{
	    FREE(utf8_str);
	    continue;
	}
	snprintf(buf, sizeof(buf), 
		"\t<INFO NAME=\"%s\" TYPE=\"%c\" SIZE=\"%llu\"/>\n", 
		escape_str, type, (unsigned long long) st.st_size);
	FREE(utf8_str);
	FREE(escape_str);
	str_append(type == 'D' ? &dir : &file, buf);
    }

    xml = make_xml_response(path, dir, file);
    if (xml && pxe_write(pxe, xml, strlen(xml)))
	Log(ERROR, "디렉토리 정보를 전송 할 수 없음: %s", 
		strerror(errno));

    FREE(dir);
    FREE(file);
    FREE(xml);
}

/**
 * @brief 로그 기록 후 실패 XML 리턴 메세지 PEER에게 전송
 * @param pxe - PXE 핸들러 포인터
 * @param err_msg - 에러 메세지
 * @return 없음
 */
static void
ret_fail(PXE *pxe, const char *err_msg)
{
    char *msg = NULL;
    char *xml_format = 
	"<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<ROOT>\n"
	"<RET STATUS=\"FAIL\" MSG=\"%s\"/>\n</ROOT>\r\n";

    ASSERT(pxe && err_msg);

    msg = dup_str(xml_format, err_msg);
    if (msg)
	(void) pxe_write(pxe, msg, strlen(msg));

    FREE(msg);
}

/**
 * @brief 디렉토리 네비게이터 쓰레드 메인
 * @param arg - PXE 쓰레드 로딩시 넘겨받는 파라메터
 * @return NULL 포인터
 */
void *
dir_navigator(void *arg)
{
    int cnt;
    PXELength len;
    PXE *pxe = NULL;
    char *encoded_path = NULL;
    void *ptr = NULL;
    void *path = NULL;
    char **list = NULL;
    char err_msg[8192];

    pxe = pxe_open(arg);
    if (!pxe)
	Log(FATAL, "PXE 핸들러 생성 실패: %s", strerror(errno));

    /* Main Loop */
    while (1)
    {
	len = pxe_read(pxe, &ptr);
	if (len > 0)
	{
	    path = calloc(1, len + 1);
	    if (!path)
	    {
		snprintf(err_msg, sizeof(err_msg), 
			"메모리 할당 실패: %s", strerror(errno));
		ret_fail(pxe, err_msg);
		goto cleanup;
	    }

	    memcpy(path, ptr, len);
	    if (decode_str_from_utf8(path, &encoded_path))
		goto cleanup;

#ifdef OS_CYGWIN
	    {
		char *ptr = NULL;

		/*
		 * CYGWIN에서는 경로명이 CYGDRIVE_STRING으로 시작하지 않으면 
		 * CYGWIN의 최상위 경로인 CYGDRIVE_STRING으로 변경한다. 
		 * 이렇게해서 사용자는 시스템에 설치된 모든 드라이브(예: C:\, D:\)를 
		 * 볼 수 있다.
		 */
		if (strncmp(encoded_path, CYGDRIVE_STRING, 
			    strlen(CYGDRIVE_STRING)) != 0)
		{
		    ptr = dup_str("%s%s", CYGDRIVE_STRING, encoded_path);
		    FREE(encoded_path);
		    encoded_path = ptr;
		}
	    }
#endif

	    /* TODO -
	     * 에이전트는 하나의 마스터에 연결되는것을 기준으로 DBS가 설계되었기
	     * 때문에 현재는 디렉토리 리스팅을 하는 루틴을 별도의 쓰레드를 실행
	     * 해서 하지 않고 여기서 처리한다. 만약 하나의 에이전트에 동시에 
	     * 디렉토리 리스팅이 요청되는 상황이 생긴다면 아래 루틴을 별도의
	     * 쓰레드로 실행시키도록 수정 바람.
	     *
	     * 만약 여러 쓰레드로 실행시킬 경우 이 모듈에서 사용한 readdir()
	     * 함수가 non thread-safe 이기 때문에 readdir_r()로 바꿔주어야 함
	     */
	    cnt = gather_dir_info(encoded_path, &list);
	    if (cnt < 0)
	    {
		snprintf(err_msg, sizeof(err_msg), "디렉토리 정보를 수집 할 수 "
			"없음: %s", strerror(errno));
		ret_fail(pxe, err_msg);
		goto cleanup;
	    }
	    
	    qsort(list, cnt, sizeof(char *), &sort_list);
	    send_info(pxe, encoded_path, list);

cleanup:
	    free_list(&list);
	    FREE(path);
	    FREE(encoded_path);
	    pxe_free(ptr);
	} /* End of if (len > 0) */
    } /* End of while (1) */

    /* Never reached */
    pxe_close(pxe);
    return NULL;
}
