/**
 * @file remover.c
 * @brief DB에서 삭제된 작업과 관련된 엔트리 제거
 */

/*
 * DB에서 삭제된 작업과 관련된 엔트리 제거
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#ifdef HAVE_LIBMYSQLCLIENT
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "configure.h"
#include "db_interface.h"
#include "../lib/slib.h"
#include "sysconfdir.h"

/**
 * @brief DBMS 접속해서 SQL 핸들러 생성
 * @param sql - SQL 핸들러 포인터
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
connect_db(sql_t *sql)
{
    char *host = NULL;
    char *user = NULL;
    char *passwd = NULL;
    char *dbname = NULL;

    ASSERT(sql != NULL);

    /* 아래 파라메터들은 로컬 설정 파일에 필수로 등록되어야 하는 파라메터이기
     * 때문에 유효성 검사를 따로 할 필요는 없다.
     */
    host = config_get_value("SQLHost");
    user = config_get_value("SQLUser");
    passwd = config_get_value("SQLPasswd");
    dbname = config_get_value("SQLName");

    if (db_connect(sql, host, user, passwd, dbname, 0, 0) < 0)
	return -1;
    if (db_set_character_set(sql, "utf8"))
    {
	db_close(sql);
	return -1;
    }

    return 0;
}

/**
 * @brief 삭제 상태로 등록된 ROW 체크
 * @param sql - SQL 핸들러
 * @param res - (OUT) 검색된 row list (호출자 db_free_result() 필요)
 * @param cnt - (OUT) 검색된 row의 개수
 * @return 
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
db_scan(sql_t *sql, sql_res_t **res, unsigned long long *cnt)
{
    char query[8192];

    ASSERT(sql != NULL);
    ASSERT(res != NULL);
    snprintf(query, sizeof(query), "SELECT id FROM jobs WHERE status='%c'", 
	    CANCEL_STAT);
    if (db_query(sql, query, res))
	return -1;
    *cnt = db_num_rows(*res);
    return 0;
}

/**
 * @brief 삭제 상태인 데이터 제거
 * @param sql - SQL 핸들러
 * @param res - 삭제 상태인 row 리스트
 * @param cnt - 삭제 대상 row의 개수
 * @return 없음
 */
static void
delete_row(sql_t *sql, sql_res_t *res, unsigned long long cnt)
{
    unsigned long long i;
    sql_row_t row;
    char query[8192];

    ASSERT(sql != NULL);
    ASSERT(res != NULL);

    i = 1;
    while ((row = db_fetch_row(res)))
    {
	printf("[%llu of %llu] Removing data now..\r", i++, cnt);
	fflush(stdout);
	snprintf(query, sizeof(query), "DELETE FROM records WHERE job_id='%s'", 
		row[0]);
	(void) db_query(sql, query, NULL);
	snprintf(query, sizeof(query), "DELETE FROM jobs WHERE id='%s'", 
		row[0]);
	(void) db_query(sql, query, NULL);
    }
    printf("\nDone.\n");
}

/**
 * @brief 단편화된 DB 테이블을 옵티마이징
 * @param sql - SQL 핸들러
 * @return 없음
 */
static void
optimize_table(sql_t *sql)
{
    char query[8192];

    ASSERT(sql != NULL);
    printf("Optimizing table now..\n");
    snprintf(query, sizeof(query), "OPTIMIZE TABLE job, record");
    (void) db_query(sql, query, NULL);
    printf("Done.\n");
}

/**
 * @brief 커맨드라인 인자 사용법 출력
 * @param name - 프로그램명
 * @return 없음
 */
static void
show_usage(const char *name)
{
    fprintf(stderr, "usage: %s [options]\n", name);
    fprintf(stderr, " -o, Optimize DB tables (it may take a long time)\n");
    fprintf(stderr, " -c, Specify alternative configuration file\n");
    fprintf(stderr, " -h, Display this help message\n");
}

int
main(int argc, char **argv)
{
    int c;
    int errflg = FALSE;
    int optimize = FALSE;
    sql_t sql;
    sql_res_t *res = NULL;
    unsigned long long cnt;
    int result;
    char config_file[PATH_MAX];
    char abs_config_file[PATH_MAX];
    void *err_msg = NULL;

    memset(config_file, 0, sizeof(config_file));
    while ((c = getopt(argc, argv, "ohc:")) != -1)
    {
	switch (c)
	{
	    case 'o':
		optimize = TRUE;
		break;
	    case 'h':
		/* Show help message */
		errflg = TRUE;
		break;
	    case 'c':
		/* Specify configuration file */
		snprintf(config_file, sizeof(config_file), "%s", optarg);
		break;
	    case ':':
		errflg = TRUE;
		fprintf(stderr, "Option -%c requires an operand\n", optopt);
		break;
	    case '?':
		errflg = TRUE;
		fprintf(stderr, "Unrecognized option: -%c\n", optopt);
		break;
	}
    }
    if (errflg)
    {
	show_usage(argv[0]);
	exit(EXIT_FAILURE);
    }
    if (config_file[0] == '\0')
	/* 커맨드라인 옵션으로 설정파일명이 지정되지 않은 경우 기본값 사용 */
	snprintf(config_file, sizeof(config_file), "%s/%s.conf", SYSCONFDIR, 
		PACKAGE_TARNAME);
    if (realpath(config_file, abs_config_file) == NULL)
    {
	fprintf(stderr, "Failed to convert pathname to absolute path: %s\n", 
		strerror(errno));
	exit(EXIT_FAILURE);
    }
    if (config_read(abs_config_file, NULL, &err_msg))
    {
	fprintf(stderr, "Failed to load %s configuration file: %s\n", 
		abs_config_file, (char *) (err_msg ? err_msg : strerror(errno)));
	exit(EXIT_FAILURE);
    }

    if (connect_db(&sql))
    {
	fprintf(stderr, "Failed to connect to DBMS: %s\n", db_errmsg(&sql));
	exit(EXIT_FAILURE);
    }

    result = db_scan(&sql, &res, &cnt);
    if (result < 0)
    {
	db_close(&sql);
	fprintf(stderr, "Failed to scan deleted entry: %s\n", db_errmsg(&sql));
	exit(EXIT_FAILURE);
    }
    if (cnt == 0)
    {
	db_free_result(res);
	db_close(&sql);
	fprintf(stderr, "Not found deleted data\n");
	exit(EXIT_SUCCESS);
    }
	
    delete_row(&sql, res, cnt);
    db_free_result(res);
    if (optimize)
	optimize_table(&sql);

    db_close(&sql);
    return 0;
}
#else /* Else of #ifdef HAVE_LIBMYSQLCLIENT */
int
main(void)
{
    printf("This program is not activated on client mode.\n");
    return 0;
}
#endif /* End if #ifdef HAVE_LIBMYSQLCLIENT */
