/**
 * @file scheduler.c
 * @brief Job Scheduler
 */

/*
 * Job Scheduler
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태(superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#define __USE_XOPEN
#define __USE_XOPEN_EXTENDED
#include <time.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <pthread.h>
#include <stdbool.h>

#include <pxe.h>
#include "scheduler.h"
#include "core.h"
#include "db_interface.h"
#include "support.h"
#include "configure.h"
#include "../lib/slib.h"

/** 
 * 스케줄 DB 스켄 주기 (초) 
 *  
 * 스케줄 지정의 최소 단위가 분(min)이다. 따라서 하나의 스케줄이 중복해서 JOB 
 * 테이블에 등록되지 않도록 최소 단위를 주기로해서 DB를 스캔한다.
 */
#define SCHEDULE_SCAN_DELAY    60

/**
 * 스케줄 DB SQL 쿼리 필드 리스트
 *
 * 쿼리에서 리턴된 결과를 사용할때 이 필드 리스트에 있는 멤버 네임을
 * 사용해서 접근한다. 따라서 쿼리문이 변경되는 경우 멤버 리스트도 
 * 같이 변경해줘야 한다.
 */
typedef enum
{
    SID,             /**< 등록번호 */
    BK_SID,          /**< 복구작업시 대상 백업 등록번호 */
    DATETIMES,       /**< 실행 할 년-월-일 시:분 */
    WEEKDAY,         /**< 실행 할 요일 */
    ARGUMENT,        /**< 대상 파일 및 디렉토리 */
    WK_TYPE,         /**< 작업 타입(백업, 복구, 삭제) */
    OPTIONS,         /**< 작업 옵션 */
    IP,              /**< 대상 IP */
    MASTER_EXPIRE,   /**< 마스터에 저장된 백업 파일 삭제할 시간 (상대시간) */
    AGENT_EXPIRE,    /**< 에이전트에 저장된 백업 파일 삭제할 시간 (상대시간) */
    LDATE,           /**< 제한 시간 */
    EMAIL,           /**< 에러 발생시 공지할 이메일 주소 */
    AGENT            /**< 에이전트 이름 */
} SC_FieldName;

/**
 * 스케줄 실행시간을 비교해서 매칭되는 스케줄을 실행시켜주는 쓰레드에 전달할
 * 인자 구조체
 */
typedef struct
{
    sql_res_t *res;     /**< 스케줄 실행시간을 비교할 스케줄 테이블 ROW LIST */
    time_t tm;          /**< 스케줄 실행을 비교할 기준 시간 */
} schedule_arg_t;

/* 스케줄 타이머를 위한 조건변수 */
static pthread_cond_t timer_cond = PTHREAD_COND_INITIALIZER;
/* 스케줄 타이머를 위한 뮤텍스 */
static pthread_mutex_t timer_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * @brief 스케줄 DB를 스캔 
 * @param sql - SQL 핸들러
 * @param res - (OUT) 쿼리 결과 (호출자가 사용 후 FREE)
 * @return
 *  성공 시 등록된 스케줄 갯수 (>= 0)\n
 *  실패 시 -1
 */
static int
db_scan(sql_t *sql, sql_res_t **res)
{
    size_t cnt;
    char *alias = NULL;
    char *escape_alias = NULL;
    char query[8192];

    ASSERT(sql && res);

    /* ServerAlias는 로컬 설정파일의 필수 파라메터라서 유효성 검사 필요없음 */
    alias = get_config("ServerAlias");
    if (db_escape_string(sql, &escape_alias, alias, strlen(alias)) == -1)
    {
	FREE(alias);
	return -1;
    }
    FREE(alias);

    /* XXX: Select filed 순서가 바뀌면 SC_FieldName 타입도 수정요망 */
    snprintf(query, sizeof(query), "SELECT id, bk_id, datetimes, weekday, "
	    "argument, wk_type, options, ip, master_expire, agent_expire, ldate, "
	    "email, agent FROM schedules WHERE master = '%s' ORDER BY id ASC", escape_alias);
    Log(DEBUG, "Scan SQL = [%s]", query);

    if (db_query(sql, query, res))
    {
	Log(ERROR, "신규 스케줄 스캔을 위한 쿼리 실패: %s", db_errmsg(sql));
	FREE(escape_alias);
	return -1;
    }

    cnt = db_num_rows(*res);
    if (cnt == 0)
	db_free_result(*res);

    FREE(escape_alias);
    return cnt;
}

/**
 * @brief 실제로 DELETE 테이블에 레코드 삽입
 * @param sql - SQL 핸들러 포인터
 * @param row - 스케줄 테이블 쿼리 결과
 * @param res - JOB 테이블 쿼리 결과
 * @param cnt - 삭제 해야할 레코드의 개수
 * @param options - DELETE 명령에 넣어줄 옵션
 * @return 없음
 */
static void
insert_delete_record(sql_t *sql, sql_row_t row, sql_res_t *res, int cnt, 
	int options)
{
    int i;
    char query[8192];
    sql_row_t job_row;
    char *alias = NULL;
    char *escape_alias = NULL;

    /* ServerAlias는 로컬 설정파일의 필수 파라메터라서 유효성 검사 필요없음 */
    alias = get_config("ServerAlias");
    if (db_escape_string(sql, &escape_alias, alias, strlen(alias)) == -1)
    {
	Log(ERROR, "쿼리문 예외문자 치환 실패: %s", strerror(errno));
	goto cleanup;
    }

    for (i = 0; i < cnt; i++)
    {
	job_row = db_fetch_row(res);
	snprintf(query, sizeof(query), "INSERT INTO jobs (master, agent, bk_sid, wk_type, "
		"options, status, email, ts) VALUES ('%s', '%s', '%s', 'D', '%d', 'W', "
		"'%s', now())", escape_alias, row[AGENT], job_row[SID], options, row[EMAIL]);
	(void) db_query(sql, query, NULL);
    }

cleanup:
    FREE(escape_alias);
    FREE(alias);
}

/**
 * @brief 마스터 서버에 저장된 백업 파일의 개수를 확인
 * @param sql - SQL 핸들러
 * @param row - SQL row
 * @param res - SQL 결과 (사용 후 FREE)
 * @return
 *  성공 시 저장된 개수,\n
 *  실패 시 -1
 */
static int
count_master_record(sql_t *sql, sql_row_t row, sql_res_t **res)
{
    char query[8192];

    snprintf(query, sizeof(query), "SELECT id FROM jobs WHERE schedule_id = '%s' AND "
	    "status = '%c' ORDER BY id ASC", row[SID], DONE_STAT);
    if (db_query(sql, query, res))
    {
	Log(ERROR, "마스터 서버에 저장된 백업 파일 개수 확인 실패: %s", 
		db_errmsg(sql));
	return -1;
    }

    return db_num_rows(*res);
}

/**
 * @brief 에이전트 호스트에 저장된 백업 파일의 개수를 확인
 * @param sql - SQL 핸들러
 * @param row - SQL row
 * @param res - SQL 결과 (사용 후 FREE)
 * @return
 *  성공 시 저장된 개수,\n
 *  실패 시 -1
 */
static int
count_agent_record(sql_t *sql, sql_row_t row, sql_res_t **res)
{
    char query[8192];

    snprintf(query, sizeof(query), "SELECT id FROM jobs WHERE schedule_id = '%s' AND "
	    "(status = '%c' OR status = '%c') AND options & %d != 0 ORDER BY "
	    "id ASC", row[SID], DONE_STAT, CANCEL_STAT, LOCAL_COPY);
    if (db_query(sql, query, res))
    {
	Log(ERROR, "에이전트 호스트에 저장된 백업 파일 개수 확인 실패: %s", 
		db_errmsg(sql));
	return -1;
    }

    return db_num_rows(*res);
}

/**
 * @brief DELETE 테이블에 삭제 작업 등록
 * @param sql - SQL 핸들러 포인터
 * @param row - SQL row
 * @return 없음
 */
static void
insert_delete(sql_t *sql, sql_row_t row)
{
    int cnt;
    int master_expire;
    int agent_expire;
    sql_res_t *res = NULL;

    master_expire = atoi(row[MASTER_EXPIRE]);
    agent_expire = atoi(row[AGENT_EXPIRE]);

    /*
     * 이 함수가 호출되기 직전에 등록된 백업 작업은 완료 상태가 아닐수가 있기 
     * 때문에 백업 파일의 개수가 expire 값과 같은 경우에도 파일을 지워줘여 한다.
     * 따라서 insert_delete_record를 호출 할 때 +1을 해준다.
     */
    if (master_expire > 0)
    {
	cnt = count_master_record(sql, row, &res);
	if (cnt >= master_expire)
	    insert_delete_record(sql, row, res, cnt - master_expire + 1, 0);
	if (res)
	{
	    db_free_result(res);
	    res = NULL;
	}
    }

    if (agent_expire > 0)
    {
	cnt = count_agent_record(sql, row, &res);
	if (cnt >= agent_expire)
	    insert_delete_record(sql, row, res, cnt - agent_expire + 1, 
		    REMOVE_LOCAL_COPY);
	if (res)
	{
	    db_free_result(res);
	    res = NULL;
	}
    }
}

/**
 * @brief 쿼리문에 문자열을 동적으로 추가
 * @param field - 쿼리문 중 field 리스트 문자열 (e.g. INSERT INTO job (sid))
 * @param value - 쿼리문 중 value 리스트 문자열 (e.g. VALUES (1152))
 * @param record - \a field에 추가할 필드명
 * @param data - \a value에 추가할 데이터
 * @param quote - value를 ' '로 둘러쌀지 여부
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
append_query(char **field, char **value, const char *record, const char *data, 
	bool quote)
{
    char *tmp = NULL;

    tmp = dup_str(", %s", record);
    if (!tmp)
	return -1;
    if (str_append(field, tmp))
	return -1;
    FREE(tmp);

    if (quote)
	tmp = dup_str(", '%s'", data);
    else
	tmp = dup_str(", %s", data);

    if (!tmp)
	return -1;
    if (str_append(value, tmp))
	return -1;
    FREE(tmp);

    return 0;
}

/**
 * @brief JOB 테이블에 작업 등록
 * @param sql - SQL 핸들러 포인터
 * @param row - SQL row
 * @return 없음
 */
static void 
insert_job(sql_t *sql, sql_row_t row)
{
    char *alias = NULL;
    char *query = NULL;
    char *field = NULL;
    char *value = NULL;
    char *escape_alias = NULL;
    char *escape_arg = NULL;

    /* ServerAlias는 로컬 설정파일의 필수 파라메터라서 유효성 검사 필요없음 */
    alias = get_config("ServerAlias");
    if (db_escape_string(sql, &escape_alias, alias, strlen(alias)) == -1)
    {
	Log(ERROR, "쿼리문 예외문자 치환 실패: %s", strerror(errno));
	goto cleanup;
    }

    if (db_escape_string(sql, &escape_arg, row[ARGUMENT], strlen(row[ARGUMENT])) == -1)
    {
	Log(ERROR, "쿼리문 예외문자 치환 실패: %s", strerror(errno));
	goto cleanup;
    }

    field = dup_str("INSERT INTO jobs (master, agent, schedule_id, argument, wk_type, "
	                "options, ip, email, ts");
    value = dup_str("VALUES ('%s', '%s', '%s', '%s', '%s', '%s', '%s', '%s', now()", 
	    escape_alias, row[AGENT], row[SID], escape_arg, row[WK_TYPE], row[OPTIONS], 
	    row[IP], row[EMAIL]);
    if (!field || !value)
    {
	Log(ERROR, "쿼리문 생성 실패: %s", strerror(errno));
	goto cleanup;
    }

    /* 제한시간 */
    if (row[LDATE])
    {
	char ldate[2048];

	snprintf(ldate, sizeof(ldate), "DATE_ADD(NOW(), INTERVAL %s MINUTE)", 
		row[LDATE]);
	if (append_query(&field, &value, "ldate", ldate, false))
	{
	    Log(ERROR, "쿼리문 생성 실패: %s", strerror(errno));
	    goto cleanup;
	}
    }
    /* 백업 시 대상 복구 번호 */
    if (row[BK_SID])
    {
	if (append_query(&field, &value, "bk_sid", row[BK_SID], true))
	{
	    Log(ERROR, "쿼리문 생성 실패: %s", strerror(errno));
	    goto cleanup;
	}
    }

    str_append(&field, ") ");
    str_append(&value, ")");
    query = dup_str("%s %s", field, value);
    if (!query)
    {
	Log(ERROR, "쿼리문 생성 실패: %s", strerror(errno));
	goto cleanup;
    }

    (void) db_query(sql, query, NULL);
    
cleanup:
    FREE(escape_alias);
    FREE(escape_arg);
    FREE(query);
    FREE(field);
    FREE(value);
    FREE(alias);
}

/**
 * @brief 주어진 DATETIME 문자열을 가공 후 기준시간과 비교해서 매칭되는지 여부확인
 * @param datetime - MySQL DATETIME 문자열
 * @param cur_tm - 스케줄 실행여부를 판단할 기준 시간
 * @return 
 *  매칭되는 경우 0,\n
 *  매칭되지 않는 경우 -1
 *
 * DATETIME 문자열에서 년, 월, 일이 0으로 채워져있는 경우, 오늘에 해당하는 값으로
 * 치환한다. 예를들어 0000-01-00 이렇게 문자열이 들어오는 경우, 오늘이 2007년 3월
 * 25일이라고 가정하면 2007-01-25로 치환되는 것이다. 
 *
 * 단, 시, 분, 초는 이렇게 하지 않는다. 따라서 매시, 매분 실행하게 하려면 그 개수
 * 만큼 맞춰서 row를 insert 해줘야 한다.
 */
static int
check_datetime(const char *datetime, time_t cur_tm, int weekday)
{
    time_t t, cond_tm;
    int result;
    char buf[20];
    struct tm *tm = NULL;
    int year, month, day, hour, min, sec;   /* sec는 사용 안함 */

    result = sscanf(datetime, "%d-%d-%d %d:%d:%d", &year, &month, &day, &hour, 
	    &min, &sec);
    if (result != 6)
    {
	Log(ERROR, "DATETIME 문자열 해석 오류: %s", strerror(errno));
	return -1;
    }

    t = time(NULL);
    tm = malloc(sizeof(struct tm));
    if (!tm)
    {
	Log(ERROR, "메모리 할당 실패: %s", strerror(errno));
	return -1;
    }
    if (localtime_r(&t, tm) == NULL)
    {
	Log(ERROR, "로컬타임 확인 실패: %s", strerror(errno));
	FREE(tm);
	return -1;
    }

    /*
     * 년, 월, 일 중 0으로 채워진 값이 있으면 오늘에 해당하는 값으로 치환
     */
//    if (year == 0)
//	year = 1900 + tm->tm_year;  /* Years since 1900 */
//    if (month == 0)
//	month = tm->tm_mon + 1;	/* Jan is 0 */
//    if (day == 0)
//	day = tm->tm_mday;

    // 주간데이터가 세팅되어 있으면 오늘 일자로 설정.
    if (weekday != 0)
    {
	year = 1900 + tm->tm_year;  /* Years since 1900 */
	month = tm->tm_mon + 1;	/* Jan is 0 */
	day = tm->tm_mday;
    }

    snprintf(buf, sizeof(buf), "%04d-%02d-%02d %02d:%02d:%02d", year, month, 
	    day, hour, min, sec); 
    cond_tm = get_tm(buf, "%Y-%m-%d %H:%M:%S");
    Log(DEBUG, "buf=[%s], cond_tem=[%d], weekday=[%d]", buf, cond_tm, weekday);
    if (cond_tm == -1)
    {
	FREE(tm);
	return -1;
    }

    /* 
     * 기준 타임스탬프가 스케줄조건의 타임스탬프와 SCAN_DELAY초 범위내에서 
     * 일치한다면 매칭된것임.
     */
    if (cur_tm == cond_tm ||
	    (cur_tm > cond_tm && cur_tm < cond_tm + SCHEDULE_SCAN_DELAY))
    {
	FREE(tm);
	return 0;
    }

    FREE(tm);
    return -1;
}

/**
 * @brief 지정된 요일 비트 플래그에 오늘이 포함되는지 확인
 * @param weekday - 요일 비트 플래그
 * @return 
 *  오늘이 포함되는 경우 0,\n
 *  오늘이 포함되지 않는 경우 -1
 */
static int
check_weekday(int weekday)
{
    time_t t;
    struct tm *tm = NULL;

    t = time(NULL);
    tm = malloc(sizeof(struct tm));
    if (!tm)
    {
	Log(ERROR, "메모리 할당 실패: %s", strerror(errno));
	return -1;
    }

    if (localtime_r(&t, tm) == NULL)
    {
	Log(ERROR, "로컬타임 확인 실패: %s", strerror(errno));
	FREE(tm);
	return -1;
    }

    /* 1 << 0: 일요일, 1 << 1: 월요일, ... , 1 << 6: 토요일 */
    if (weekday & (1 << tm->tm_wday))
    {
	FREE(tm);
	return 0;
    }

    FREE(tm);
    return -1;
}

/**
 * @brief 현재 실행되어야 할 스케줄을 찾아서 JOB 테이블에 등록
 * @param sql - SQL 핸들러 포인터
 * @param res - SQL result 포인터
 * @param tm - 스케줄 실행여부를 판단할 기준 시간
 * @return 없음
 *
 * WEEKDAY: 1 << 0: 일요일, 1 << 1: 월요일, ..., 1 << 6: 토요일\n
 */
static void *
process_schedule(void *data)
{
    sql_t sql;
    sql_row_t row;
    void *err_msg = NULL;
    schedule_arg_t *arg = data;

    ASSERT(data);
    if (dbms_connect(&sql, &err_msg))
    {
	Log(ERROR, "%s", err_msg);
	return NULL;
    }

    while ((row = db_fetch_row(arg->res)))
    {
	/* 년-월-일 시:분 포멧이 기준 시간과 일치하는지 체크 */
	if (check_datetime(row[DATETIMES], arg->tm, atoi(row[WEEKDAY])))
	    continue;

	/* 요일이 지정되어 있는 경우 오늘이 포함되어 있는지 체크 */
	if (row[WEEKDAY] && atoi(row[WEEKDAY]) != 0)
	{
	    if (check_weekday(atoi(row[WEEKDAY])))
		continue;
	}
	
	/* 모든 조건이 매칭되기 때문에 JOB 테이블에 등록 */
	insert_job(&sql, row);

	/* 마스터 또는 에이전트 백업 파일에 대한 삭제시간이 있는 경우 DELETE 
	 * 명령 등록
	 */
	if (atoi(row[MASTER_EXPIRE]) > 0 || atoi(row[AGENT_EXPIRE]) > 0)
	    insert_delete(&sql, row);
    }

    db_free_result(arg->res);
    db_close(&sql);
    FREE(data);
    return NULL;
}

/**
 * @brief 스케줄 검사 쓰레드에 전달한 인자를 생성
 * @param res - 검사할 쓰레드 ROW 리스트
 * @param tm - 스케줄 검사를 할 기준 시간
 * @return
 *  성공 시 생성된 인자 구조체 메모리 포인터 (사용 후 FREE),\n
 *  실패 시 NULL
 */
static schedule_arg_t *
make_schedule_arg(sql_res_t *res, time_t tm)
{
    schedule_arg_t *arg = NULL;

    ASSERT(res);
    ASSERT(tm > 0);
    arg = malloc(sizeof(schedule_arg_t));
    if (!arg)
    {
	Log(ERROR, "메모리 할당 에러: %s", strerror(errno));
	return NULL;
    }

    arg->res = res;
    arg->tm = tm;
    return arg;
}

/**
 * @brief 입력받은 절대시간까지 delay하는 타이머
 * @param abs - 절대시간 (epoch time)
 * @return 없음
 *
 * XXX - 
 * sleep()을 사용해서 timer를 만들면 시간이 지날수록 시간이 약간씩 뒤로 밀린다.
 * 그래서 처음 스케줄러가 실행된 epoch time을 기준으로 SCHEDULE_SCAN_DELAY 만큼
 * 더한 절대시간을 사용해서 timer를 구현하기 위해 pthread 조건변수를 사용했다.
 */
static void
abs_timer(time_t abs)
{
    struct timespec timeout;

    (void) pthread_mutex_lock(&timer_mutex);
    timeout.tv_sec = abs;
    timeout.tv_nsec = 0;
    (void) pthread_cond_timedwait(&timer_cond, &timer_mutex, &timeout);
    (void) pthread_mutex_unlock(&timer_mutex);
}

/**
 * @brief Job Scheduler
 * @param arg - PXE 쓰레드 로딩시 넘겨받는 파라메터
 * @return NULL 포인터
 */
void *
scheduler(void *arg)
{
    int cnt;
    sql_t sql;
    time_t tm;
    pthread_t tid;
    PXE *pxe = NULL;
    sql_res_t *res = NULL;
    void *err_msg = NULL;
    schedule_arg_t *schedule_arg = NULL;

    /* PXE 핸들러 생성 */
    pxe = pxe_open(arg);
    if (!pxe)
	Log(FATAL, "PXE 핸들러 생성 실패: %s", strerror(errno));

    if (dbms_connect(&sql, &err_msg))
	Log(FATAL, "%s", err_msg);

    /* 스케줄 스캔을 할 최초 기준시간 */
    tm = time(NULL);	

    /* Main loop */
    while (1)
    {
	if ((cnt = db_scan(&sql, &res)) > 0)
	{
	    schedule_arg = make_schedule_arg(res, tm);
	    if (!schedule_arg)
	    {
		db_free_result(res);
		res = NULL;
	    }
	    else
	    {
		if (pthread_create(&tid, NULL, process_schedule, 
			    (void *) schedule_arg))
		    Log(ERROR, "스케줄 검사 쓰레드 로딩 실패: %s", 
			    strerror(errno));
		else
		    (void) pthread_detach(tid);
	    }
	}
    
	/* 어떠한 이유로 위의 루틴이 SCHEDULE_SCAN_DELAY보다 오래 걸려서 tm
	 * 값이 현재시간보다 작으면 abs_timer()가 즉시 리턴된다. 따라서 이런
	 * 경우에도 등록된 스케줄이 조금 늦게 실행되긴 하지만 결과적으로 모두
	 * 실행이 된다는 보장을 할 수 있다.
	 */
	tm += SCHEDULE_SCAN_DELAY;
	abs_timer(tm);
    }

    /* Never reached */
    db_close(&sql);
    pxe_close(pxe);
    FREE(arg);

    return NULL;
}
