/**
 * @file show_status.c
 * @brief 현재 호스트의 상태를 XML 형태로 출력
 */

/*
 * 현재 호스트의 상태를 XML 형태로 출력
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <pxe.h>
#include "show_status.h"
#include "support.h"
#include "../lib/slib.h" 

/**
 * @brief 로그 기록 후 실패 XML 리턴 메세지 PEER에게 전송
 * @param pxe - PXE 핸들러 포인터
 * @param err_msg - 에러 메세지
 * @return 없음
 */
static void
ret_fail(PXE *pxe, const char *err_msg)
{
    char *msg = NULL;
    char *xml_format = 
	"<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<ROOT>\n"
	"<RET STATUS=\"FAIL\" MSG=\"%s\"/>\n</ROOT>\r\n";

    ASSERT(pxe && err_msg);

    msg = dup_str(xml_format, err_msg);
    if (msg)
	(void) pxe_write(pxe, msg, strlen(msg));

    FREE(msg);
}

/**
 * @brief 설정내용을 XML 형태로 변경
 * @param config - 설정 내용
 * @return 
 *  성공 시 생성된 XML (호출자 사용 후 FREE),\n
 *  실패 시 NULL
 */
static char *
config_to_xml(config_list_t *config)
{
    int i;
    char *entry = NULL;
    char *xml = NULL;
    
    xml = dup_str("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<ROOT>\n"
	    "<RET STATUS=\"SUCCESS\"/>\n<CONFIG>\n");
    if (!xml)
	return NULL;

    for (i = 0; config[i].parameter; i++)
    {
	entry = dup_str("\t<PARAM NAME=\"%s\" VALUE=\"%s\"/>\n", 
		config[i].parameter, config[i].value); 
	if (!entry)
	{
	    FREE(xml);
	    return NULL;
	}
	if (str_append(&xml, entry))
	{
	    FREE(entry);
	    FREE(xml);
	    return NULL;
	}
	FREE(entry);
    }
    if (str_append(&xml, "</CONFIG>\n</ROOT>\r\n"))
    {
	FREE(xml);
	return NULL;
    }

    return xml;
}

/**
 * @brief 현재 상태 정보를 XML 형식으로 전송
 * @param pxe - PXE 핸들러
 * @param config - 현재 설정 정보
 * @return 없음
 */
static void
send_status(PXE *pxe, config_list_t *config)
{
    char *xml = NULL;

    xml = config_to_xml(config);
    if (!xml)
    {
	Log(ERROR, "상태정보 XML 생성 실패: %s", strerror(errno));
	return;
    }
    if (pxe_write(pxe, xml, strlen(xml)))
    {
	Log(ERROR, "상태정보 XML 전송 실패: %s", strerror(errno));
	return;
    }
    FREE(xml);
}

/**
 * @brief 호스트 상태 출력 쓰레드 메인
 * @param arg - PXE 쓰레드 로딩시 넘겨받는 파라메터
 * @return NULL 포인터
 */
void *
show_status(void *arg)
{
    char msg[8192];
    PXE *pxe = NULL;
    void *ptr = NULL;
    PXELength len;
    config_list_t *config = NULL;

    pxe = pxe_open(arg);
    if (!pxe)
	Log(FATAL, "PXE 핸들러 생성 실패: %s", strerror(errno));

    /* Main Loop */
    while (1)
    {
	len = pxe_read(pxe, &ptr);
	if ((config = config_get_list()) == NULL)
	{
	    snprintf(msg, sizeof(msg), "설정파일 정보 확인 실패: %s", 
		    strerror(errno));
	    ret_fail(pxe, msg);    
	    goto cleanup;
	}

	send_status(pxe, config);
	
cleanup:
	if (config)
	    config_free_list(config);
	pxe_free(ptr);
    }

    /* Never reached */
    pxe_close(pxe);
    return NULL;
}
