/**
 * @file socket_interface.c
 * @brief Socket Interface
 */

/*
 * Socket Interface
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <sys/stat.h>

#include <pxe.h>
#include "socket_interface.h"
#include "support.h"
#include "dbs.h"
#include "configure.h"
#include "../lib/slib.h"

/**
 * 소켓 인터페이스로 수신된 명령어의 인자 리스트 구조체
 */
typedef struct
{
    int argc;          /**< 인자의 갯수 */
    char **argv;       /**< 인자 문자열 배열 */
} ArgumentList;

/**
 * Function Prototypes
 */
static void cmd_dirlist(int fd, PXE *pxe, ArgumentList *list);
static void cmd_agentlist(int fd, PXE *pxe, ArgumentList *list);
static void cmd_showstatus(int fd, PXE *pxe, ArgumentList *list);
static void cmd_getsize(int fd, PXE *pxe, ArgumentList *list);

/**
 * 소켓 인터페이스 명령어 콜백 함수 테이블 구조체
 */
typedef struct 
{
    char *cmd;                                     /**< 명령어 문자열 */
    void (*func)(int fd, PXE *, ArgumentList *);   /**< 콜백 함수 포인터 */
} CallbackTable;

/**
 * 소켓 인터페이스 명령어 콜백 함수 테이블 
 */
CallbackTable callback[] = {
    { "DIRLIST", cmd_dirlist },      /* 디렉토리 내용 보기 */
    { "AGENTLIST", cmd_agentlist },  /* 현재 연결된 에이전트 리스트 보기 */
    { "SHOWSTATUS", cmd_showstatus }, /* 마스터 또는 에이전트의 현재 상태 보기 */
    { "GETSIZE", cmd_getsize }  /* 백업 파일의 사이즈를 반환 */
};

/**
 * @brief 소켓에서 패킷 수신
 * @param fd - 소켓 디스크립터
 * @param packet - (OUT) 수신한 패킷을 저장할 메모리 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 수신한 패킷 사이즈,\n
 *  EOF 수신 시 0,\n
 *  실패 시 -1
 *
 * 수신한 패킷에서 \\r\\n 문자열이 발견될때까지 계속 수신
 */
static int
recv_packet(int fd, char **packet)
{
    size_t cnt;
    size_t size;
    size_t old_size;
    unsigned char buf[8192];

    ASSERT(fd >= 0 && packet);

    /* 현재 할당된 메모리 사이즈 */
    size = 0;
    old_size = 0;

    while (1)
    {
	cnt = read(fd, buf, sizeof(buf));
	if (cnt < 0)
	    return -1;
	else if (cnt == 0)
	    return 0;
	else
	{
	    if (size == 0)
	    {
		/* 
		 * 처음 패킷을 수신해서 메모리 할당 
		 */
		size = old_size = cnt; 

		/* +1 is NULL character */
		*packet = malloc(size + 1);
		if (!*packet)
		    return -1;
		memcpy(*packet, buf, cnt);
		(*packet)[size] = '\0';
	    }
	    else
	    {
		/*
		 * 기존 메모리를 확장
		 */
		size += cnt;
		
		/* +1 is NULL character */
		*packet = realloc(*packet, size + 1);
		if (!*packet)
		    return -1;
		memcpy(&(*packet)[old_size], buf, cnt);
		(*packet)[size] = '\0';
		old_size = size;
	    }

	    /* EOF 수신 */
	    if (strstr(*packet, "\r\n"))
		break;
	}
    }

    /* 수신한 패킷 사이즈 */
    return size;
}

/**
 * @brief 문자열을 공백으로 구분해서 문자열 배열을 생성
 * @param list - 인자 리스트 구조체 포인터
 * @param packet - 수신한 문자열 포인터
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
make_arg_list(ArgumentList *list, const char *packet)
{
    int i;
    int cnt;
    int ret_val;
    char *ptr = NULL;
    char *end = NULL;
    char *dup_str = NULL;

    ASSERT(list && packet);

    /* for cleanup routine */
    ret_val = -1;

    dup_str = strdup(packet);
    if (!dup_str)
	goto cleanup;

    end = strstr(dup_str, "\r\n");
    if (!end)
    {
	errno = EINVAL;
	goto cleanup;
    }
    *end = '\0';

    /* 
     * 문자열에 TAB이 몇개 있는지 확인 
     * TAB이 없어도 적어도 하나의 문자열이 있기 때문에 1부터 시작
     */
    cnt = 1;
    for (i = 0; i < strlen(dup_str); i++)
    {
	if (dup_str[i] == '\t')
	    cnt++;
    }

    list->argc = cnt;
    list->argv = (char **) malloc(sizeof(unsigned char *) * cnt);
    if (!list->argv)
	goto cleanup;

    ptr = dup_str;
    for (i = 0; (end = strchr(ptr, '\t')) != NULL; i++)
    {
	*end++ = '\0';
	list->argv[i] = strdup(ptr);
	ptr = end;
    }
    list->argv[i] = strdup(ptr);

    /* Success */
    ret_val = 0;

cleanup:
    FREE(dup_str);

    return ret_val;
}

/**
 * @brief make_arg_list에서 생성된 문자열 배열을 해제
 * @param list - 문자열 배열 포인터
 * @return 없음
 */
static void
free_arg_list(ArgumentList *list)
{
    int i;

    ASSERT(list);

    for (i = 0; i < list->argc; i++)
	FREE(list->argv[i]);

    FREE(list->argv);
}

/**
 * @brief 로그 기록 후 실패 XML 리턴 메세지 소켓으로 전송
 * @param fd - 소켓 디스크립터
 * @param err_msg - 에러 메세지
 * @return 없음
 */
static void
ret_fail(int fd, const char *err_msg)
{
    char *msg = NULL;
    char *xml_format = 
	"<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<ROOT>\n"
	"<RET STATUS=\"FAIL\" MSG=\"%s\"/>\n</ROOT>\r\n";

    ASSERT(err_msg);

    msg = dup_str(xml_format, err_msg);
    if (msg)
	(void) write(fd, msg, strlen(msg));

    FREE(msg);
}

/**
 * @brief DIRLIST 명령 처리
 * @param fd - 소켓 디스크립터
 * @param pxe - PXE 핸들러 포인터
 * @param list - 인자 배열 포인터
 * @return 없음
 *
 * list->argv[0]: 명령어\n
 * list->argv[1]: IP\n
 * list->argv[2]: 대상 디렉토리
 */
static void 
cmd_dirlist(int fd, PXE *pxe, ArgumentList *list)
{
    PXELength len;
    void *ptr = NULL;
    char err_msg[8192];

    ASSERT(pxe && list);
    
    if (list->argc != 3)
    {
	snprintf(err_msg, sizeof(err_msg), 
		"Invalid argument list of DIRLIST command");
	ret_fail(fd, err_msg);
	return;
    }

    /* 절대경로 체크에서 Cygwin의 C:\ or D:/ 형식도 고려 */
    if (list->argv[2][0] != '/' &&
	    !(list->argv[2][1] == ':' && 
		(list->argv[2][2] == '/' || list->argv[2][2] == '\\')))
    {
	snprintf(err_msg, sizeof(err_msg), 
		"Target directory name must to be absolute pathname");
	ret_fail(fd, err_msg);
	return;
    }

    if (pxe_set_destination(pxe, list->argv[1], DN_THREAD))
    {
	snprintf(err_msg, sizeof(err_msg), "Disconnected agent IP address: %s", 
		list->argv[1]);
	ret_fail(fd, err_msg);
	return;
    }

    if (pxe_write(pxe, list->argv[2], strlen(list->argv[2])))
    {
	snprintf(err_msg, sizeof(err_msg), "Failed to send DIRLIST command to "
		"%s: %s", list->argv[1], list->argv[2]);
	ret_fail(fd, err_msg);
	return;
    }

    len = pxe_read_timedwait(pxe, &ptr, DBS_READ_TIMEOUT);
    if (len > 0)
	(void) writen(fd, ptr, len);

    pxe_free(ptr);
}

/**
 * @brief AGENTLIST 명령 처리
 * @param fd - 소켓 디스크립터
 * @param pxe - PXE 핸들러 포인터
 * @param list - 인자 배열 포인터
 * @return 없음
 *
 * list->argv[0]: 명령어
 */
static void 
cmd_agentlist(int fd, PXE *pxe, ArgumentList *list)
{
    int i, cnt;
    char buf[8192];
    char err_msg[8192];
    char *ptr = NULL;
    PXE_DCE_LIST **result = NULL;

    if (list->argc > 1)
    {
	snprintf(err_msg, sizeof(err_msg), 
		"Invalid argument list of AGENTLIST command");
	ret_fail(fd, err_msg);
	return;
    }

    cnt = pxe_get_dce_list(&result);
    ptr = dup_str("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<ROOT>\n"
	    "<RET STATUS=\"SUCCESS\" COUNT=\"%d\"/>\n", cnt);
    if (cnt > 0)
    {
	if (str_append(&ptr, "<LIST>\n"))
	{
	    FREE(ptr);
	    return;
	}
	for (i = 0; result[i] != NULL; i++)
	{
	    snprintf(buf, sizeof(buf), "\t<INFO ALIAS=\"%s\" IP=\"%s\"/>\n", 
		    result[i]->alias, result[i]->ip);
	    if (str_append(&ptr, buf))
	    {
		FREE(ptr);
		return;
	    }
	}
	if (str_append(&ptr, "</LIST>\n"))
	{
	    FREE(ptr);
	    return;
	}
    }
    if (str_append(&ptr, "</ROOT>\r\n"))
    {
	FREE(ptr);
	return;
    }
    (void) writen(fd, ptr, strlen(ptr));
    FREE(ptr);
    pxe_free_dce_list(result);
}

/**
 * @brief SHOWSTATUS 명령 처리
 * @param fd - 소켓 디스크립터
 * @param pxe - PXE 핸들러 포인터
 * @param list - 인자 배열 포인터
 * @return 없음
 *
 * 해당 호스트의 현재 상태를 XML 형태로 돌려준다. 현재 상태에는 해당 호스트이 
 * 설정파일의 내용이나 현재 작업하고 있는 내역등이 포함 될 수 있다.
 *
 * list->argv[0]: 명령어\n
 * list->argv[1]: 대상 호스트의 IP (없는 경우 마스터로 간주)
 */
static void 
cmd_showstatus(int fd, PXE *pxe, ArgumentList *list)
{
    PXELength len;
    void *ptr = NULL;
    char err_msg[8192];

    ASSERT(pxe && list);
    
    if (list->argc > 2)
    {
	snprintf(err_msg, sizeof(err_msg), 
		"Invalid argument list of SHOWSTATUS command");
	ret_fail(fd, err_msg);
	return;
    }

    if (pxe_set_destination(pxe, list->argc == 2 ? list->argv[1] : PXE_MASTER, 
		SS_THREAD))
    {
	snprintf(err_msg, sizeof(err_msg), "Disconnected agent IP address: %s", 
		list->argv[1]);
	ret_fail(fd, err_msg);
	return;
    }

    if (pxe_write(pxe, NULL, 0))
    {
	snprintf(err_msg, sizeof(err_msg), "Failed to send SHOWSTATUS command to "
		"%s", list->argc == 2 ? list->argv[1] : "localhost");
	ret_fail(fd, err_msg);
	return;
    }

    len = pxe_read_timedwait(pxe, &ptr, DBS_READ_TIMEOUT);
    if (len > 0)
	(void) writen(fd, ptr, len);

    pxe_free(ptr);
}

/**
 * @brief GETSIZE 명령 처리
 * @param fd - 소켓 디스크립터
 * @param pxe - PXE 핸들러 포인터
 * @param list - 인자 배열 포인터
 * @return 없음
 *
 * 해당 백업 작업에서 생성된 파일의 사이즈를 반환한다.
 *
 * list->argv[0]: 명령어\n
 * list->argv[1]: 백업 등록번호
 */
static void 
cmd_getsize(int fd, PXE *pxe, ArgumentList *list)
{
    char err_msg[8192];
    char *data_dir = NULL;
    char pathname[PATH_MAX];
    struct stat st;
    BOOL err_flag;
    char *ptr = NULL;

    ASSERT(fd >= 0);
    ASSERT(pxe != NULL);
    ASSERT(list != NULL);

    if (list->argc != 2)
    {
	snprintf(err_msg, sizeof(err_msg), 
		"Invalid argument list of GETSIZE command");
	ret_fail(fd, err_msg);
	return;
    }

    data_dir = get_config("DataDir");
    if (!data_dir)
    {
	snprintf(err_msg, sizeof(err_msg), "Failed to get backup repository "
		"information");
	ret_fail(fd, err_msg);
	return;
    }

    err_flag = FALSE;
    snprintf(pathname, sizeof(pathname), "%s/%s.%s", data_dir, list->argv[1], 
	    TAR_EXT);
    while (lstat(pathname, &st) == -1)
    {
	if (errno != ENOENT || err_flag == TRUE)
	{
	    snprintf(err_msg, sizeof(err_msg), "Failed to get backup file "
		    "information: %s", strerror(errno));
	    ret_fail(fd, err_msg);
	    FREE(data_dir);
	    return;
	}
	else
	{
	    snprintf(pathname, sizeof(pathname), "%s/%s.%s", data_dir, 
		    list->argv[1], GZ_EXT);
	    err_flag = TRUE;
	}
    }

    ptr = dup_str("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n<ROOT>\n<RET "
	    "STATUS=\"SUCCESS\"/>\n<INFO NAME=\"%s\" SIZE=\"%llu\"/>\n</ROOT>"
	    "\r\n", pathname, (unsigned long long) st.st_size);
    (void) writen(fd, ptr, strlen(ptr));
    FREE(ptr);
    FREE(data_dir);
}

/**
 * @brief 소켓 인터페이스 스위치보드
 * @param arg - PXE 쓰레드 로딩시 넘겨받는 파라메터 (소켓 디스크립터)
 * @return NULL 포인터
 */
static void *
switchboard(void *arg)
{
    int i;
    int fd;
    int len;
    PXE *pxe = NULL;
    char *packet = NULL;
    ArgumentList arg_list;

    /* PXE 핸들러 생성 */
    pxe = pxe_open(arg);
    if (!pxe)
	Log(FATAL, "PXE 핸들러 생성 실패: %s", strerror(errno));
    
    /* 소켓디스크립터 획득 */
    fd = *((int *) pxe_get_argument(arg));

    len = recv_packet(fd, &packet); 
    if (len <= 0)
    {
	if (len < 0)
	    Log(ERROR, "제어 패킷 수신 실패: %s", strerror(errno));
	else
	    Log(ERROR, "컨트롤 커넥션 종료되었음");

	goto cleanup;
    }

    /* 인자 배열 생성 */
    if (make_arg_list(&arg_list, packet))
    {
	Log(ERROR, "인자 배열 생성 실패: %s", strerror(errno));
	goto cleanup;
    }

    /* 명령어 처리 */
    for (i = 0; i < ARRAY_SIZE(callback); i++)
    {
	if (strcasecmp(callback[i].cmd, arg_list.argv[0]) == 0)
	{
	    callback[i].func(fd, pxe, &arg_list);
	    break;
	}
    }
    if (i == ARRAY_SIZE(callback))
	Log(ERROR, "유효하지 않은 제어 명령어: %s", arg_list.argv[0]);

    free_arg_list(&arg_list);

cleanup:
    pxe_close(pxe);
    FREE(packet);
    FREE(arg);
    close(fd);

    return NULL;
}

/**
 * @brief 소켓 인터페이스 쓰레드 메인
 * @param arg - PXE 쓰레드 로딩시 넘겨받는 파라메터
 * @return NULL 포인터
 */
void *
socket_interface(void *arg)
{
    int listenfd;
    int connfd;
    int port;
    int result;
    socklen_t len;
    char *port_ptr = NULL;
    struct sockaddr_in cliaddr;

    port_ptr = get_config("ControlPort");
    if (!port_ptr)
	Log(FATAL, "'ControlPort' 파라메터가 존재하지 않음");
    port = atoi(port_ptr);
    FREE(port_ptr);
    if (port < 1 || port > 65535)
	Log(FATAL, "유효하지 않은 'ControlPort' 파라메터 값: %d", port);
    listenfd = socket_listen(port);
    if (listenfd < 0)
	Log(FATAL, "제어포트(%d) 리슨 실패: %s", port, strerror(errno));

    while (1)
    {
	connfd = Accept(listenfd, (struct sockaddr *) &cliaddr, &len);
	if (connfd < 0)
	{
	    Log(ERROR, "새로운 컨트롤 커넥션 연결 실패: %s", strerror(errno));
	    continue;
	}

	result = pxe_load_thread(PXE_AUTO_ASSIGN, switchboard, &connfd, 
		sizeof(connfd), PXE_DETACH_THREAD);
	if (result < 0)
	{
	    Log(ERROR, "제어 명령 처리 쓰레드 로딩 실패: %s", strerror(errno));
	    continue;
	}
    }

    /* Never reached */
    return NULL;
}
