/* Copyright 2007 Samjung Data Service, Inc. All rights reserved. */

/**
 * @file support.c
 * @brief Support 루틴 모듈
 * @author 김기태 (superkkt@sds.co.kr)
 * @defgroup support_module Support 루틴 모듈
 * 각종 유틸리티 함수들을 모아놓은 모듈
 * @{
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <errno.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <iconv.h>
#define __USE_XOPEN
#include <time.h>

#include "support.h"
#include "types.h"
#include "configure.h"
#include "../lib/slib.h"

/**
 * @brief 문자열로 된 IP를 정수형으로 변환
 * @param ip - 문자열 IP 포인터
 * @param buf - 정수형 IP를 저장할 메모리 포인터
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
int
convert_ip_to_int(const char *ip, unsigned char *buf)
{
    int i;
    char *dup_ip = NULL;
    char *offset = NULL;
    char *ptr = NULL;

    ASSERT(ip && buf);

    dup_ip = strdup(ip);
    if (!dup_ip)
	return -1;

    offset = dup_ip;
    for (i = 0; i < 3; i++)
    {
	ptr = strchr(offset, '.');
	if (!ptr)
	{
	    errno = EINVAL;
	    free(dup_ip);
	    return -1;
	}
	*ptr++ = '\0';

	buf[i] = atoi(offset);
	offset = ptr;
    }
    buf[3] = atoi(offset);

    free(dup_ip);

    return 0;
}

/**
 * @brief 정수형으로 된 IP를 문자열로 변환
 * @param ip - 정수형 IP 포인터
 * @param buf - 문자열 IP를 저장할 메모리 포인터
 * @param len - \a buf 의 길이
 * @return
 *  \a buf 의 포인터
 */
char *
convert_ip_to_str(const unsigned char *ip, char *buf, size_t len)
{
    ASSERT(ip && buf);

    snprintf(buf, len, "%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);

    return buf;
}

/**
 * @brief 특정 문자열로 된 시간을 epoch time으로 변환
 * @param str - 문자열로 된 시간 (예 - 2006/10/02-08:35:12)
 * @param format - 시간 형식 문자열
 * @param tm - Epoch time 포인터
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 *
 * 시간을 표현하는 문자열 형식은 %Y-%m-%d %H:%M:%S
 */
int
convert_time_to_epoch(const char *str, 
	const char *format, time_t *tm)
{
    struct tm t;

    ASSERT(str && format && tm);

    if (strptime(str, format, &t) == NULL)
	return -1;

    t.tm_isdst = -1;
    *tm = mktime(&t);

    return 0;
}

/**
 * @brief 동적 메모리에 문자열 복사
 * @param format - 문자열 포멧
 * @param ... - \a format 의 가변인자
 * @return 
 *  성공 시 생성된 문자열 메모리 포인터 (호출자 사용 후 FREE),\n
 *  실패 시 NULL 포인터
 */
char *
dup_str(const char *format, ...)
{
    va_list ap;
    ssize_t len;
    char *ptr = NULL;

    va_start(ap, format);

    /* 필요한 메모리 사이즈 확인 */
#if __STDC_VERSION__ < 199901L
    FILE *nulldev = fopen("/dev/null", "w");
    if (!nulldev)
	return NULL;
    len = vfprintf(nulldev, format, ap);
    fclose(nulldev);
#else
    len = vsnprintf(NULL, 0, format, ap);
#endif
    if (len <= 0)
	return NULL;

    ptr = malloc(len + 1);
    if (!ptr)
	return NULL;

    va_start(ap, format);
    vsprintf(ptr, format, ap);
    va_end(ap);

    return ptr;
}

/**
 * @brief 지정된 제한시간이 지났는지 체크
 * @param timeout - 제한시간 (UNIX time stamp)
 * @return 
 *  제한시간이 지나지 않았으면 0,\n
 *  제한시간이 지났으면 -1
 */
int
check_timeout(TIME timeout)
{
    time_t tm;

    tm = time(NULL);
    if ((TIME) tm > timeout)
	return -1;

    return 0;
}

/**
 * @brief 기존 문자열에 새로운 문자열 추가
 * @param ptr - (IN/OUT) 기존 문자열 포인터
 * @param str - 추가할 문자열
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 *
 * 기존 문자열이 NULL인 경우 새로운 메모리를 할당하고, NULL이 아닌 경우
 * 새로운 문자열을 추가하기 위한 공간만큼 기존 문자열 메모리를 확장한 후
 * 새로운 문자열을 추가한다.
 */
int
str_append(char **ptr, const char *str)
{
    ASSERT(str);

    /* +1 is terminating null */
    if (!*ptr)
	*ptr = calloc(1, strlen(str) + 1);
    else
	*ptr = realloc(*ptr, strlen(*ptr) + strlen(str) + 1);

    if (!*ptr)
	return -1;

    strcat(*ptr, str);

    return 0;
}

/**
 * @brief 문자열안에 어떤 문자가 중복되서 나올경우 한개로 줄여주는 코드 
 * @param string - 대상 문자열
 * @param c - 찾을 문자
 * @return
 *  성공 시 제거된 문자의 갯수,\n
 *  실패 시 -1
 */
int
trim_duplicate(char *string, char c)
{
    int i, len;

    len = strlen(string);
    for (i = 0; i < len; i++)
    {
	if (string[i] == c)
	{
	    if (i > 0 && string[i - 1] == c)
	    {
		memmove(&string[i - 1], &string[i], len - i);
		len--;
		i--;
	    }
	}
    }
    string[len] = '\0';
    return len;
}

/**
 * @brief 문자열의 앞에 존재하는 공백을 제거
 * @param str - 공백을 제거할 문자열
 * @return 없음
 */
void
trim_post_space(char *str)
{
    int i;
    size_t len;

    len = strlen(str);
    for (i = 0; i < len; i++)
    {
	if (str[i] != ' ')
	    break;
    }

    if (i > 0)
    {
	memmove(&str[0], &str[i], len - i);
	str[len - i] = '\0';
    }
}

/**
 * 시스템 인코딩에서 UTF-8로 문자열 변환한다. 설정파일에 저장된 인코딩이 UTF-8
 * 이라면 변환 없이 새로 할당된 동적 메모리에 문자열을 복사하고 성공을 리턴한다.
 *
 * @param str - 변환할 문자
 * @param ptr - UTF-8로 문자열이 저장된 메모리 포인터
 * @return 성공 시 0, 실패 시 -1
 * @warning \a ptr 은 호출자가 사용 후 FREE() 함수를 통해 메모리를 해제해야 한다.
 */
int
encode_str_to_utf8(const char *str, char **ptr)
{
    iconv_t cd;
    size_t in_len;
    size_t out_len;
    char *out_str = NULL;
    char *charset = NULL;

    /* 시스템의 캐릭터셋 */
    charset = get_config("CharSet");
    if (!charset)
    {
	Log(ERROR, "'CharSet' 파라메터가 존재하지 않음");
	errno = EINVAL;
	return -1;
    }

    /* 시스템 인코딩이 UTF-8이면 변환없이 그대로 문자열 복사 */
    if (strcasecmp(charset, "UTF-8") == 0)
    {
	FREE(charset);
	*ptr = dup_str("%s", str);
	if (!*ptr)
	    return -1;
	
	return 0;
    }

    /* 
     * UTF-8로 변환시 필요한 메모리 상한선 
     */
    in_len = strlen(str);
    out_len = in_len * 2 + 1;

    *ptr = calloc(1, out_len);
    if (!*ptr)
    {
	FREE(charset);
	return -1;
    }

    /* iconv() 실행 후 결과 포인터의 위치가 변하기 때문에 미리 저장 */
    out_str = *ptr;

    /* 설정파일에 지정된 인코딩에서 UTF-8로 변환 */
    cd = iconv_open("UTF-8", charset);
    if (cd == (iconv_t) -1)
    {
	FREE(*ptr);
	FREE(charset);
	return -1;
    }

    if (iconv(cd, (char **) &str, &in_len, 
		(char **) &out_str, &out_len) == (size_t) -1)
    {
	FREE(*ptr);
	FREE(charset);
	return -1;
    }

    iconv_close(cd);
    FREE(charset);
    return 0;
}

/**
 * @brief UTF-8에서 시스템 인코딩으로 문자열 변환
 * @param str - 변환할 UTF-8 문자열
 * @param ptr - (OUT) 변환된 문자열이 저장된 메모리 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 *
 * 설정파일에 저장된 인코딩이 UTF-8이라면 변환 없이 새로 할당된 동적 메모리에
 * 문자열을 복사하고 성공을 리턴한다.
 */
int
decode_str_from_utf8(const char *str, char **ptr)
{
    iconv_t cd;
    size_t in_len;
    size_t out_len;
    char *out_str = NULL;
    char *charset = NULL;

    /* 시스템의 캐릭터셋 */
    charset = get_config("CharSet");
    if (!charset)
    {
	Log(ERROR, "'CharSet' 파라메터가 존재하지 않음");
	errno = EINVAL;
	return -1;
    }

    /* 시스템 인코딩이 UTF-8이면 변환없이 그대로 문자열 복사 */
    if (strcasecmp(charset, "UTF-8") == 0)
    {
	FREE(charset);
	*ptr = dup_str("%s", str);
	if (!*ptr)
	    return -1;
	
	return 0;
    }

    /* 
     * 시스템 인코딩으로 변환시 필요한 메모리 상한선 
     */
    in_len = strlen(str);
    out_len = in_len + 1;

    *ptr = calloc(1, out_len);
    if (!*ptr)
    {
	FREE(charset);
	return -1;
    }

    /* iconv() 실행 후 결과 포인터의 위치가 변하기 때문에 미리 저장 */
    out_str = *ptr;

    /* UTF-8에서 설정파일에 지정된 인코딩으로 변환 */
    cd = iconv_open(charset, "UTF-8");
    if (cd == (iconv_t) -1)
    {
	FREE(charset);
	FREE(*ptr);
	return -1;
    }

    if (iconv(cd, (char **) &str, &in_len, 
		(char **) &out_str, &out_len) == (size_t) -1)
    {
	FREE(charset);
	FREE(*ptr);
	return -1;
    }

    iconv_close(cd);
    FREE(charset);
    return 0;
}

/**
 * @brief 문자열로 된 시간을 Unix Timestamp로 변환
 * @param str - 시간을 표현하는 문자열
 * @param format - 문자열에 표현된 시간의 포맷
 * @return
 *  성공 시 계산된 Unix Timestamp,\n
 *  실패 시 -1
 */
time_t
get_tm(const char *str, const char *format)
{
    time_t t;
    struct tm tm;
 
    if (strptime(str, format, &tm) == NULL)
	return -1;

    tm.tm_isdst = -1;
    t = mktime(&tm);
    if (t == -1)
	return -1;

    return t;
}

/**
 * 대상 문자열에서 \a search 문자열을 찾아서 모두 \a replace 문자열로 치환한다.
 *
 * @param in - 대상 문자열
 * @param out - 치환이 완료된 문자열
 * @param search - 찾을 문자열
 * @param replace - 치환할 문자열
 * @return 0; 실패 시 -1
 * @warning \a out 은 사용 후 FREE()로 해제해야 함
 */
int
str_replace(const char *in, char **out, const char *search, const char *replace)
{
    int i, j, cnt;
    size_t in_len, search_len, replace_len, out_len;

    in_len = strlen(in);
    search_len = strlen(search);
    replace_len = strlen(replace);

    if (in_len <= 0 || search_len <= 0)
    {
	errno = EINVAL;
	return -1;
    }

    for (i = 0, cnt = 0; i < in_len; i++)
    {
	if (in[i] == search[0])
	{
	    if (strncmp(&in[i], search, search_len) == 0)
		cnt++;
	}
    }

    if (search_len < replace_len)
	out_len = in_len + cnt * (replace_len - search_len) + 1;
    else
	out_len = in_len - cnt * (search_len - replace_len) + 1;

    *out = calloc(1, out_len);
    if (!*out)
	return -1;

    for (i = 0, j = 0; i < in_len; i++)
    {
	if (in[i] == search[0])
	{
	    if (strncmp(&in[i], search, search_len) == 0)
	    {
		memcpy(&(*out)[j], replace, replace_len);
		j += replace_len;
		/* for loop에서 i++가 다시 발생하기 때문에 -1 */
		i += search_len - 1;
	    }
	}
	else
	    (*out)[j++] = in[i];
    }

    return 0;
}
/**
 * @} 
 * End of support_module
 */
