/**
 * @file tar_frontend.c
 * @brief Frontend of TAR library
 */

/*
 * Frontend of TAR library
 *
 * Copyright 2007 Samjung Data Service, Inc.  All rights reserved.
 * 김기태 (superkkt@sds.co.kr)
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <glob.h>
#include <inttypes.h>
#include <limits.h>
#include <fcntl.h>
#include <pthread.h>

#include <zlib.h>
#include <libtar.h>
#include "tar_frontend.h"
#ifdef HAVE_LIBMYSQLCLIENT
#include "db_interface.h"
#endif
#include "core.h"
#include "core_restore.h"
#include "dbs.h"
#include "support.h"
#include "configure.h"
#include "../lib/slib.h"

/** 
 * 백업용 Libtar wrapping functions 
 */
static int b_open(const char *path, int oflag, ...);
static ssize_t b_read(int fd, void *buf, size_t len);
static ssize_t b_write(int fd, const void *buf, size_t len);
static ssize_t b_gz_write(int fd, const void *buf, size_t len);
static int b_close(int fd);

/** 
 * 복구용 Libtar wrapping functions 
 */
static int r_open(const char *path, int oflag, ...);
static ssize_t r_read(int fd, void *buf, size_t len);
static ssize_t r_gz_read(int fd, void *buf, size_t len);
static ssize_t r_write(int fd, const void *buf, size_t len);
static int r_close(int fd);

/**
 * TAR 내부 버퍼 구조체 
 */
typedef struct 
{
    off_t total;                       /**< 버퍼에 있는 총 데이터 사이즈 */
    off_t cur_offset;                  /**< 현재 오프셋 */
    unsigned char buf[INTERNAL_BUF];   /**< 내부 버퍼 */
} TAR_BUF;

/**
 * TAR 리스트 구조체 
 */
typedef struct 
{
    int active;     /**< 활성 여부 */
    int finish;     /**< 마지막 패킷인지 여부 */
    int fd;         /**< Local Copy 사용시 파일 디스크립터 */
    z_stream strm;  /**< libz 스트림 */
    PXE *pxe;       /**< PXE 핸들러 */
    TAR_BUF tbuf;   /**< TAR 버퍼 구조체 */
    void *info;     /**< 정보 구조체 포인터 (백업 or 복구 정보) */
    BOOL long_flag; /**< 연속적인 롱네임/롱링크 헤더 플래그 */
} TAR_List;

/**
 * 실행중인 TAR 작업 리스트 
 * 최대 동시 실행수만큼 TAR 리스트 생성 
 */
static TAR_List tlist[MAX_REQUEST];

/**
 * TAR 리스트를 위한 뮤텍스
 */
static pthread_mutex_t tlist_mutex;

/**
 * 백업용 Wrapping Functions 
 */
static tartype_t backup_type = 
{
    (openfunc_t) b_open,
    (closefunc_t) b_close,
    (readfunc_t) b_read,
    (writefunc_t) b_write
};

/**
 * 백업용 gzip Wrapping Functions 
 */
static tartype_t backup_gz_type = 
{
    (openfunc_t) b_open,
    (closefunc_t) b_close,
    (readfunc_t) b_read,
    (writefunc_t) b_gz_write
};

/**
 * 복구용 Wrapping Functions 
 */
static tartype_t restore_type = 
{
    (openfunc_t) r_open,
    (closefunc_t) r_close,
    (readfunc_t) r_read,
    (writefunc_t) r_write
};

/**
 * 복구용 gzip Wrapping Functions 
 */
static tartype_t restore_gz_type = {
    (openfunc_t) r_open,
    (closefunc_t) r_close,
    (readfunc_t) r_gz_read,
    (writefunc_t) r_write
};

/**
 * @brief TAR 리스트 초기화
 * @param 없음
 * @return 
 *  성공 시 0,\n
 *  실패 시 -1
 */
int
init_tar_list(void)
{
    int i;
    int result;

    result = pthread_mutex_init(&tlist_mutex, NULL);
    if (result) {
	errno = result;
	return -1;
    }

    for (i = 0; i < ARRAY_SIZE(tlist); i++) {
	tlist[i].active = FALSE;
	tlist[i].finish = FALSE;
	tlist[i].fd = -1;
	tlist[i].pxe = NULL;
	memset(tlist[i].tbuf.buf, 0, sizeof(tlist[i].tbuf.buf));
	tlist[i].info = NULL;
	tlist[i].long_flag = FALSE;
    }

    return 0;
}

/**
 * @brief 빈 리스트 검색
 * @param 없음
 * @return 
 *  빈 리스트 검색 시 리스트 인덱스,\n
 *  비어있는 리스트가 없으면 -1
 *
 * 리스트 접근 시 뮤텍스로 임계영역 설정을 하지 않기 때문에
 * 호출자가 임계영역 처리를 직접 해줘야 한다.
 */
static int
search_empty_list(void)
{
    int i;

    for (i = 0; i < ARRAY_SIZE(tlist); i++) {
	if (tlist[i].active == FALSE)
	    return i;
    }

    return -1;
}

/**
 * @brief TAR 리스트 설정
 * @param pxe - PXE 핸들러 포인터
 * @param info - 정보구조체 포인터 (백업 or 복구 구조체)
 * @return
 *  성공 시 TAR 리스트 인덱스,\n
 *  실패 시 -1
 */
static int
activate_tar_list(PXE *pxe, void *info)
{
    int idx;
    int result;

    ASSERT(pxe && info);

    result = pthread_mutex_lock(&tlist_mutex);
    if (result) {
	errno = result;
	return -1;
    }

    idx = search_empty_list();
    if (idx < 0)
	goto cleanup;

    tlist[idx].active = TRUE;
    tlist[idx].pxe = pxe;
    tlist[idx].info = info;
    
cleanup:
    result = pthread_mutex_unlock(&tlist_mutex);
    if (result)
    {
	errno = result;
	return -1;
    }

    return idx;
}

/**
 * @brief TAR 리스트 제거
 * @param idx - TAR 리스트 인덱스
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
inactivate_tar_list(int idx)
{
    int result;

    ASSERT(idx >= 0);

    result = pthread_mutex_lock(&tlist_mutex);
    if (result)
    {
	errno = result;
	return -1;
    }

    tlist[idx].active = FALSE;
    tlist[idx].finish = FALSE;
    tlist[idx].fd = -1;
    tlist[idx].pxe = NULL;
    tlist[idx].tbuf.cur_offset = 0;
    tlist[idx].tbuf.total = 0;
    memset(tlist[idx].tbuf.buf, 0, sizeof(tlist[idx].tbuf.buf));
    tlist[idx].info = NULL;
    tlist[idx].long_flag = FALSE;

    result = pthread_mutex_unlock(&tlist_mutex);
    if (result)
    {
	errno = result;
	return -1;
    }

    return 0;
}

/**
 * @brief 백업용 open wrapping function
 * @param path - 할당받은 TAR 리스트 인덱스 문자열
 * @param oflag - 옵션 플래그 (사용 안함)
 * @param ... - 가변인자 (사용 안함)
 * @return 정수형으로 변환된 \a path
 */
static int
b_open(const char *path, int oflag, ...)
{
    ASSERT(path);

    /* 파일네임은 할당받은 TAR 리스트 인덱스 */
    return atoi(path);
}

/**
 * @brief 백업용 read wrapping function (사용 안함)
 */
static ssize_t
b_read(int fd, void *buf, size_t len)
{
    return 0;
}

/**
 * @brief 주어진 패킷이 TAR 헤더인지 판별
 * @param tar - TAR 헤더 포인터
 * @return 
 *  TAR 헤더일 경우 0,\n
 *  아닐 경우 -1
 */
static inline int
is_tar_header(struct tar_header *tar)
{
    ASSERT(tar);

    /* FIXME -
     * 매직스트링과 파일 타입 플래그를 체크해서 TAR 헤더인지 판별한다. 이렇게
     * 판별하는 방법은 정확한 방법은 아니다. 제일 정확한 방법은 체크섬을
     * 계산해서 헤더의 체크섬과 비교하는 방법인데 시도해본 결과 엄청난 성능
     * 저하를 유발시켰다.
     */
    if (memcmp(tar->magic, DBS_MAGIC, MAGIC_LEN) == 0 &&
            ((tar->typeflag >= '0' && tar->typeflag <= '7') ||
                tar->typeflag == 'L' || tar->typeflag == 'K' || tar->typeflag == '\0'))
        return 0;

    return -1;
}

/**
 * @brief DBS 전용 매직스트링을 USTAR 매직 스트링으로 변경
 * @param ptr - 변환할 데이터
 * @param len - \a ptr의 길이
 * @return 없음
 */
static void
change_magic(void *ptr, size_t len)
{
    off_t offset;
    struct tar_header *tar = NULL;

    for (offset = 0; offset < len; offset += T_BLOCKSIZE)
    {
	/* 압축모드인 경우 T_BLOCKSIZE의 배수로 길이가 맞춰진다는 보장이 없다 */
	if (offset + T_BLOCKSIZE > len)
	    break;
	tar = (struct tar_header *) ((char *) ptr + offset);
	if (is_tar_header(tar) == 0)
	    memcpy(tar->magic, USTAR_MAGIC, strlen(USTAR_MAGIC));
    }
}

/**
 * @brief 로컬호스트에 백업 복사본 저장
 * @param idx - TAR List Index
 * @param buf - 데이터
 * @param len - \a buf 의 길이
 * @return 
 *  성공 시 실제로 기록한 바이트 수,\n
 *  실패 시 -1
 */
static int
write_local_copy(int idx, const void *buf, size_t len)
{
    int result;
    void *dup_ptr = NULL;
    
    dup_ptr = malloc(len);
    if (!dup_ptr)
	return -1;
    memcpy(dup_ptr, buf, len);

    /* 로컬복사본은 마스터로 전송되지 않기 때문에 DBS 전용 매직 스트링이 그대로
     * 저장되게 된다. 따라서 여기서 USTAR 매직 스트링으로 변경 후 저장한다.
     */
    change_magic(dup_ptr, len);
    result = write(tlist[idx].fd, dup_ptr, len);
    free(dup_ptr);

    return result;
}

/**
 * @brief 내부버퍼에 적재된 데이터를 전송 후 버퍼 비움
 * @param idx - TAR 리스트 인덱스
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
flush_buffer(int idx)
{
    int ret_val;
    size_t len;
    PXE *pxe = NULL;
    void *ptr = NULL;
    void *buf = NULL;
    off_t *offset = NULL;
    BackupInfo *info = NULL;
    BackupHeader header;

    ASSERT(idx >= 0);

    /* for cleanup routine */
    ret_val = -1;

    pxe = tlist[idx].pxe;
    offset = &tlist[idx].tbuf.cur_offset;
    info = tlist[idx].info;
    buf = tlist[idx].tbuf.buf;

    if (*offset > 0)
    {
	memcpy(&header, &info->header, sizeof(BackupHeader));
	header.cmd = BACKUP_DATA;

	len = create_backup_packet(&ptr, info->regno, &header, buf, *offset);
	if (len == 0 || !ptr)
	    goto cleanup;

	if (pxe_write(pxe, ptr, len))
	    goto cleanup;

	/* 압축 백업 시 로컬 복사본은 압축하지 않는 옵션이 사용된 경우에는 이미
	 * 로컬 복사본을 디스크에 저장한 상태이기 때문에 여기서 저장하지 않음
	 * 단, 압축 옵션이 지정되지 않은 상태에서는 IS_LOCAL_NOCOMPRESS 옵션에
	 * 상관없이 로컬 복사본 옵션 여부에 따라 복사본을 저장한다.
	 */
	if (IS_LOCAL_COPY(info->header.options) &&
		(!IS_COMPRESS(info->header.options) || 
		 !IS_LOCAL_NOCOMPRESS(info->header.options)))
	{
	    if (write_local_copy(idx, buf, *offset) < 0)
		goto cleanup;
	}

	*offset = 0;
    }

    /* Success */
    ret_val = 0;

cleanup:
    FREE(ptr);
    return ret_val;
}

/**
 * @brief 내부 버퍼에 송신 데이터 적재
 * @param idx - TAR 리스트 인덱스
 * @param buf - 데이터
 * @param len - \a buf 의 길이
 * @return
 *  성공 시 0,\n
 *  실패(내부버퍼 공간 부족) 시 -1
 */
static int
push_buffer(int idx, const void *buf, size_t len)
{
    off_t *offset = NULL;

    ASSERT(idx >= 0 && buf);

    offset = &tlist[idx].tbuf.cur_offset;

    /* 내부 버퍼에 공간이 부족한 경우 */
    if (*offset + len > INTERNAL_BUF)
	return -1;

    memcpy(&tlist[idx].tbuf.buf[*offset], buf, len);
    *offset += len;

    return 0;
}

/**
 * Longlink/Longname 헤더가 중간에 잘리지 않고 안전하게 저장될만한 공간이 있는지
 * 확인하고 공간이 부족하면 버퍼를 flush해서 여유 공간을 만든다. 
 *
 * 일반적인 Longname/Longlink는 아래와 같은 구조로 되어있다.
 * 롱 헤더 | 롱 네임 | 리얼 헤더 | 데이터
 *
 * 그러나 롱링크과 롱헤더가 동시에 적용된 파일이 있을 수 있는데 그 경우 아래와
 * 같은 구조를 가진다.
 *
 * 롱 헤더 | 롱 네임 | 롱 헤더 | 롱 네임 | 리얼 헤더 | 데이터
 *
 * 이런 특성을 이용해서 롱 헤더가 나온 후에 리얼 헤더가 나오지 않고 바로 롱헤더가
 * 다시 나온 경우에는 연결되는 헤더들이 잘리지 않도록 flush를 하지 않고, 반대의
 * 경우에는 flush를 한다. 즉, 롱헤더가 나오고 다음에 리얼 헤더가 나온 경우 다음번
 * 롱헤더가 나올때 flush를 해버리면 된다.
 *
 * @param idx - TAR 리스트 인덱스
 * @param tar - TAR 헤더 포인터
 * @return 0; 실패 시 -1
 */
static int
check_safe_block(int idx, struct tar_header *tar)
{
    if (tar->typeflag == GNU_LONGNAME_TYPE || 
	    tar->typeflag == GNU_LONGLINK_TYPE)
    {
	if (tlist[idx].long_flag == FALSE)
	{
	    tlist[idx].long_flag = TRUE;
	    return flush_buffer(idx);
	}
    }
    else
	tlist[idx].long_flag = FALSE;

    return 0;
}

/**
 * @brief 백업용 write wrapping function
 * @param fd - 할당받은 TAR 리스트 인덱스
 * @param buf - 데이터
 * @param len - \a buf 의 길이
 * @return 
 *  성공 시 내부 버퍼에 적재한 데이터의 길이,\n
 *  실패 시 -1
 */
static ssize_t
b_write(int fd, const void *buf, size_t len)
{
    struct tar_header *tar = NULL;

    ASSERT(buf && len <= INTERNAL_BUF);

    tar = (struct tar_header *) buf;
    if (is_tar_header(tar) == 0)
    {
	/* Longname/Longlink Extension이 나오면 헤더가 잘리는것을 예방 */
	if (check_safe_block(fd, tar))
	    return -1;
    }

    /* 버퍼에 데이터가 모두 적재되면 리턴 */
    if (push_buffer(fd, buf, len) == 0)
	return len;

    /* 
     * 버퍼에 공간이 없기 때문에 비우고 다시 적재 
     */
    if (flush_buffer(fd))
	return -1;
    push_buffer(fd, buf, len);

    return len;
}

/**
 * @brief zlib 핸들러 압축 초기화
 * @param strm - zlib 핸들러 포인터
 * @return
 *  성공 시 Z_OK,\n
 *  실패 시 zlib's 에러코드
 */
static int
zlib_stream_deflate_init(z_stream *strm)
{
    ASSERT(strm);

    strm->zalloc = Z_NULL;
    strm->zfree = Z_NULL;
    strm->opaque = Z_NULL;

    return deflateInit2(strm, Z_BEST_SPEED, Z_DEFLATED, 
	    MAX_WBITS + 16, MAX_MEM_LEVEL, Z_DEFAULT_STRATEGY);
}

/**
 * @brief zlib 핸들러 복구 초기화
 * @param strm - zlib 핸들러 포인터
 * @return
 *  성공 시 Z_OK,\n
 *  실패 시 zlib's 에러코드
 */
static int
zlib_stream_inflate_init(z_stream *strm)
{
    ASSERT(strm);

    strm->zalloc = Z_NULL;
    strm->zfree = Z_NULL;
    strm->opaque = Z_NULL;
    strm->avail_in = 0;
    strm->next_in = Z_NULL;

    return inflateInit2(strm, MAX_WBITS + 16);
}

/**
 * @brief zlib 에러코드를 분석해서 적절한 errno 셋팅
 * @param err - zlib 에러코드
 * @return -1
 */
static int
zerr(int err)
{
    switch (err) {
	case Z_ERRNO:
	    errno = Z_ERRNO;
	    break;
	case Z_STREAM_ERROR:
	case Z_DATA_ERROR:
	    errno = EINVAL;
	    break;
	case Z_MEM_ERROR:
	    errno = ENOMEM;
	    break;
	default:
	    errno = EPERM;
	    break;
    }

    return -1;
}

/**
 * @brief gzip으로 압축해서 내부버퍼에 적재
 * @param strm - zlib 핸들러 포인터
 * @param idx - TAR 리스트 인덱스
 * @param buf - 데이터
 * @param len - \a buf 의 길이
 * @return 
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
gz_push_buffer(z_stream *strm, int idx, const void *buf, size_t len)
{
    int result;
    size_t avail, have;
    off_t *offset = NULL;
    BackupInfo *info = NULL;

    ASSERT(strm && idx >= 0 && buf);

    offset = &tlist[idx].tbuf.cur_offset;
    strm->avail_in = len;
    strm->next_in = (unsigned char *) buf;

    /* 압축 백업 & 로컬 복사본 저장이 지정되어있고, 로컬 복사본은 압축하지 않는
     * 옵션이 같이 지정되어 있으면 데이터를 압축하기 전에 여기서 미리 복사본을
     * 디스크에 저장한다.
     */
    info = tlist[idx].info;
    if (IS_LOCAL_COPY(info->header.options) &&
	    IS_LOCAL_NOCOMPRESS(info->header.options))
    {
	if (write_local_copy(idx, buf, len) < 0)
	    return -1;
    }

    do {
	avail = INTERNAL_BUF - *offset;
	strm->avail_out = avail;
	strm->next_out = &tlist[idx].tbuf.buf[*offset];

	/* TAR 아카이브의 마지막 패킷이 들어올때만 Z_FINISH */
	result = deflate(strm, tlist[idx].finish ? Z_FINISH : Z_NO_FLUSH);
	if (result == Z_STREAM_ERROR)
	    return zerr(result);

	have = avail - strm->avail_out;
	*offset += have;

	/* 내부 버퍼가 꽉 찬 경우 */
	if (have == avail) {
	    if (flush_buffer(idx))
		return -1;
	}
    } while (strm->avail_out == 0);

    return len;
}

/**
 * @brief 백업용 gzip write wrapping function
 * @param fd - 할당받은 TAR 리스트 인덱스
 * @param buf - 데이터
 * @param len - \a buf 의 길이
 * @return 
 *  성공 시 내부 버퍼에 적재한 데이터의 길이,\n
 *  실패 시 -1
 */
static ssize_t
b_gz_write(int fd, const void *buf, size_t len)
{
    struct tar_header *tar = NULL;

    ASSERT(buf && len <= INTERNAL_BUF);

    tar = (struct tar_header *) buf;
    if (is_tar_header(tar) == 0)
    {
	/* 압축된 데이터는 마스터서버에서 헤더를 파싱해서 기록을 남기지 않기 때문에
	 * DBS 전용 MAGIC STRING을 패킷을 전송하기 전에 여기서 USTAR MAGIC STRING
	 * 으로 변경해줘야 한다.
	 */
	memcpy(tar->magic, USTAR_MAGIC, strlen(USTAR_MAGIC));

	/* Logname Extension이 나오면 헤더가 잘리는것을 예방하기 위해 내부버퍼를 
	 * 모두 비운 후에 데이터를 적재한다.
	 */
	if (tar->typeflag == GNU_LONGNAME_TYPE)
	{
	    if (flush_buffer(fd))
		return -1;
	}
    }

    return gz_push_buffer(&tlist[fd].strm, fd, buf, len);
}

/**
 * @brief 백업용 close wrapping function (사용 안함)
 */
static int
b_close(int fd)
{
    return 0;
}

/**
 * @brief 복구용 open wrapping function
 * @param path - 할당받은 TAR 리스트 인덱스
 * @param oflag - 옵션 플래그 (사용 안함)
 * @param ... - 가변인자 (사용 안함)
 * @return 정수형으로 변환된 \a path
 */
static int
r_open(const char *path, int oflag, ...)
{
    ASSERT(path);

    /* 파일네임은 할당받은 TAR 리스트 인덱스 */
    return atoi(path);
}

/**
 * @brief PXE 핸들에서 패킷을 수신해서 내부버퍼를 채움
 * @param idx - TAR List Index
 * @return
 *  성공 시 수신한 패킷에서 사용자 데이터 사이즈의 크기,\n
 *  실패 시 -1
 */
static ssize_t
fill_buffer(int idx)
{
    int ret_val;
    PXE *pxe = NULL;
    void *ptr = NULL;
    void *buf = NULL;
    void *data = NULL;
    off_t *total = NULL;
    off_t *offset = NULL;
    RestoreHeader header;

    ASSERT(idx >= 0);

    /* for cleanup routine */
    ret_val = -1;

    pxe = tlist[idx].pxe;
    total = &tlist[idx].tbuf.total;
    offset = &tlist[idx].tbuf.cur_offset;
    buf = tlist[idx].tbuf.buf;

    /* XXX - 
     * Peer가 INTERNAL BUF에 충분히 저장 가능한 사이즈로 패킷을
     * 보낸다고 가정하고 구현된 루틴이다.
     */
    (void) pxe_read(pxe, &ptr);
    parse_restore_header(&header, (char *) ptr + BK_HEADER_SIZE);
    data = (char *) ptr + BK_HEADER_SIZE + RESTORE_HEADER_SIZE;

    if (header.cmd == RESTORE_ERROR) {
	errno = ENODATA;
	Log(ERROR, "[원격호스트에서 전송된 에러 메세지] %s", data);
	goto cleanup;
    }

    if (header.len > 0 && data)
	memcpy(buf, data, header.len);
    *total = header.len;
    *offset = 0;

    /* Success */
    ret_val = header.len;

cleanup:
    pxe_free(ptr);
    return ret_val;
}

/**
 * @brief 내부 버퍼에서 지정된 사이즈만큼 데이터 추출
 * @param idx - TAR List Index
 * @param buf - 추출한 데이터를 저장할 버퍼 포인터
 * @param len - \a buf 의 길이
 * @return
 *  성공 시 추출한 데이터의 길이,\n
 *  실패 시 -1
 */
static ssize_t
pop_buffer(int idx, void *buf, size_t len)
{
    size_t size;
    off_t *total= NULL;
    off_t *offset = NULL;

    ASSERT(buf);

    total = &tlist[idx].tbuf.total;
    offset = &tlist[idx].tbuf.cur_offset;

    /* 최초 실행이거나 내부버퍼에 있는 데이터를 전부 사용한 경우 */
    if (*offset == *total) {
	if (fill_buffer(idx) < 0)
	    return -1;
    }

    /* 내부버퍼에 있는 데이터보다 많은 양을 요구하면 버퍼에
     * 있는 데이터 만큼만 pop 시켜준다.
     */
    if (len > *total - *offset)
	size = *total - *offset;
    else
	size = len;

    memcpy(buf, &tlist[idx].tbuf.buf[*offset], size);
    *offset += size;

    return size;
}

/**
 * @brief 내부 버퍼에서 gzip으로 압축된 데이터를 지정된 사이즈만큼 추출
 * @param strm - zlib 핸들러 포인터
 * @param idx - TAR List Index
 * @param buf - 추출한 데이터를 저장할 버퍼 포인터
 * @param len - \a buf 의 길이
 * @return
 *  성공 시 추출한 데이터의 길이,\n
 *  실패 시 -1
 */
static ssize_t
gz_pop_buffer(z_stream *strm, int idx, void *buf, size_t len)
{
    int result;
    size_t size;
    size_t have;
    off_t *total= NULL;
    off_t *offset = NULL;

    ASSERT(strm && buf);

    total = &tlist[idx].tbuf.total;
    offset = &tlist[idx].tbuf.cur_offset;

    /* 최초 실행이거나 내부버퍼에 있는 데이터를 전부 사용한 경우 */
    if (*offset == *total)
    {
	if (fill_buffer(idx) < 0)
	    return -1;
    }

    /* 내부버퍼에 있는 데이터보다 많은 양을 요구하면 버퍼에
     * 있는 데이터 만큼만 pop 시켜준다.
     */
    if (len > *total - *offset)
	size = *total - *offset;
    else
	size = len;

    strm->avail_in = size;
    strm->next_in = &tlist[idx].tbuf.buf[*offset];
    strm->avail_out = len;
    strm->next_out = buf;

    result = inflate(strm, Z_NO_FLUSH);
    switch (result)
    {
	case Z_NEED_DICT:
	    result = Z_DATA_ERROR;
	case Z_DATA_ERROR:
	case Z_MEM_ERROR:
	    return zerr(result);
    }

    have = len - strm->avail_out;
    *offset += size - strm->avail_in;

    /* Libtar는 TAR_BLOCKSIZE 만큼 read를 못하면 에러로 처리한다.
     * 압축을 풀때 strm->avail_in이 작아서 요청된 len을 꽉 채우지
     * 못하는 경우가 생기는데 이 경우에는 재귀호출로 len만큼 채워서
     * 리턴하도록 한다.
     */
    return have == len ? have : 
	have + gz_pop_buffer(strm, idx, (char *) buf + have, len - have);
}

/**
 * @brief 복구용 read wrapping function
 */
static ssize_t
r_read(int fd, void *buf, size_t len)
{
    ASSERT(buf);
    return pop_buffer(fd, buf, len);
}

/**
 * @brief 복구용 gzip read wrapping function
 */
static ssize_t
r_gz_read(int fd, void *buf, size_t len)
{
    ASSERT(buf);
    return gz_pop_buffer(&tlist[fd].strm, fd, buf, len);
}

/**
 * @brief 복구용 write wrapping function (사용 안함)
 */
static ssize_t
r_write(int fd, const void *buf, size_t len)
{
    return 0;
}

/**
 * @brief 복구용 close wrapping function (사용 안함)
 */
static int
r_close(int fd)
{
    return 0;
}

/**
 * @brief TAR 작업할 파일 및 디렉토리 등록
 * @param list - Libtar 리스트 포인터
 * @param path - 작업할 파일 및 디렉토리 포인터
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
add_list(libtar_list_t *list, const char *path)
{
    int i;
    glob_t pglob;
    char *pathname = NULL;
    char *dup_string = NULL;
    char *ptr = NULL;
    char *null_ptr = NULL;

    ASSERT(list && path);

    dup_string = strdup(path);
    if (!dup_string)
	return -1;

    ptr = dup_string;
    while (ptr)
    {
	null_ptr = strchr(ptr, '\n');
	if (null_ptr)
	    *null_ptr++ = '\0';

#ifdef OS_CYGWIN
	/*
	 * CYGWIN에서는 경로명이 CYGDRIVE_STRING으로 시작하지 않으면 CYGWIN의 
	 * 최상위 경로인 CYGDRIVE_STRING으로 변경한다. 이렇게해서 사용자는 시스템에
	 * 설치된 모든 드라이브(예: C:\, D:\)를 볼 수 있다.
	 */
	if (strncmp(ptr, CYGDRIVE_STRING, strlen(CYGDRIVE_STRING)) != 0)
	    pathname = dup_str("%s%s", CYGDRIVE_STRING, ptr);
	else
	    pathname = dup_str("%s", ptr);
#else
	pathname = dup_str("%s", ptr);
#endif
	if (!pathname)
	{
	    FREE(dup_string);
	    errno = EINVAL;
	    return -1;
	}

	if (glob(pathname, GLOB_NOCHECK, NULL, &pglob) != 0)
	{
	    FREE(dup_string);
	    FREE(pathname);
	    return -1;
	}
	
	/* 매칭되는 파일이 하나도 없으면 에러처리 */
	if (pglob.gl_pathc == 0)
	{
	    globfree(&pglob);
	    FREE(dup_string);
	    FREE(pathname);
	    errno = ENOENT;
	    return -1;
	}

	for (i = 0; i < pglob.gl_pathc; i++)	
	{
	    /* 이 함수가 종료되어도 리스트에 추가된 스트링이
	     * 유지되도록 스트링을 복사해서 추가한다.
	     */
	    if (libtar_list_add(list, strdup(pglob.gl_pathv[i])))
	    {
		FREE(dup_string);
		FREE(pathname);
		globfree(&pglob);
		return -1;
	    }
	}
	globfree(&pglob);
	
	FREE(pathname);
	ptr = null_ptr;
    }

    FREE(dup_string);
    return 0;
}

/**
 * @brief Libtar 백업 초기 설정
 * @param tar - (OUT) Libtar 헨들러 포인터 (호출자 사용 후 tar_close())
 * @param list - (OUT) Libtar 리스트 포인터 (호출자 사용 후 libtar_list_free())
 * @param idx - TAR 리스트 인덱스
 * @param path - 작업 대상 파일 및 디렉토리
 * @param hdr - 백업모듈 헤더 포인터
 * @return  
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
init_backup(TAR **tar, libtar_list_t **list, 
	int idx, const char *path, BackupHeader *hdr)
{
    int option;
    int result;
    char *charset = NULL;
    char *decoded_str = NULL;
    char buf[BUFSIZ];

    ASSERT(tar && list && idx >= 0 && path);

    *list = libtar_list_new(LIST_QUEUE, NULL);
    if (!*list)
	return -1;

    /*
     * 마스터에서 전송한 백업 대상 리스트(UTF-8)를 에이전트 시스템의
     * 캐릭터셋으로 변경 한 후 TAR 대상 파일 리스트에 추가한다.
     */
    if (decode_str_from_utf8(path, &decoded_str))
	return -1;
    if (add_list(*list, decoded_str))
    {
	FREE(decoded_str);
	return -1;
    }
    FREE(decoded_str);

    option = TAR_GNU;
    if (IS_IGNORE_LOCK_FILE(hdr->options))
	option |= TAR_IGNORE_LOCK_FILE;

    charset = get_config("CharSet");
    if (!charset)
    {
	Log(ERROR, "'CharSet' 파라메터가 존재하지 않음");
	errno = EINVAL;
	return -1;
    }

    snprintf(buf, sizeof(buf), "%d", idx);
    result = tar_open(tar, buf, 
	    IS_COMPRESS(hdr->options) ? &backup_gz_type : &backup_type, 
	    O_WRONLY | O_CREAT, 0644, option, hdr->mtime, charset);
    FREE(charset);
    if (result < 0)
	return -1;

    return 0;
}

/**
 * @brief Libtar 복구 초기 설정
 * @param tar - (OUT) Libtar 핸들러 포인터 (호출자 사용 후 tar_close())
 * @param idx - TAR 리스트 인덱스
 * @param path - 작업 대상 파일 및 디렉토리
 * @param hdr - 복구모듈 헤더 포인터
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
init_restore(TAR **tar, int idx, const char *path, RestoreHeader *hdr)
{
    int result;
    char buf[BUFSIZ];
    char *charset = NULL;

    ASSERT(tar && hdr);

    charset = get_config("CharSet");
    if (!charset)
    {
	Log(ERROR, "'CharSet' 파라메터가 존재하지 않음");
	errno = EINVAL;
	return -1;
    }

    snprintf(buf, sizeof(buf), "%d", idx);
    result = tar_open(tar, buf, 
	    IS_COMPRESS(hdr->options) ? &restore_gz_type : &restore_type, 
	    O_RDONLY, 0, TAR_GNU 
#ifdef OS_CYGWIN
	    | TAR_FORCE_WRITE_PERM 
#endif
	    , 0, charset);
    FREE(charset);
    if (result < 0)
	return -1;

    return 0;
}

/**
 * @brief TAR 아카이브가 완료되었을때 EOF 기록
 * @param tar - TAR 핸들러 포인터
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
append_eof(TAR *tar, void **msg)
{
    int i;

    ASSERT(tar && msg);

    for (i = 0; i < 2; i++)
    {
	/* TAR 아카이브의 마지막 패킷이라는것을 알림 */
	if (i == 1)
	    tlist[tar->fd].finish = TRUE;

	if (tar_append_eof(tar) != 0)
	{
	    *msg = dup_str("TAR EOF 문자열 추가 실패: %s", 
		    strerror(errno));
	    return -1;
	}
    }

    return 0;
}

/**
 * @brief 백업 에러 통보 패킷인지 확인
 * @param packet - 수신한 패킷 포인터
 * @param msg - (OUT) 에러 메세지 포인터 (호출자 사용 후 FREE)
 * @return 
 *  에러 통보 패킷인 경우 TRUE,\n
 *  아닌 경우 FALSE
 */
static BOOL
is_backup_error(void *packet, void **msg)
{
    char *data = NULL;
    BackupHeader backup_hdr;

    ASSERT(packet && msg);

    parse_backup_header(&backup_hdr, (char *) packet + BK_HEADER_SIZE);
    data = (char *) packet + BK_HEADER_SIZE + BACKUP_HEADER_SIZE;

    if (backup_hdr.cmd != BACKUP_ERROR)
	return FALSE;

    *msg = calloc(1, backup_hdr.len + 1);
    if (!*msg)
	memcpy(*msg, data, backup_hdr.len);

    return TRUE;
}

/**
 * @brief Libtar를 사용해서 백업 실행
 * @param pxe - PXE 핸들러 포인터
 * @param tar - Libtar 핸들러 포인터
 * @param list - Libtar 리스트 포인터
 * @param hdr - 백업모듈 헤더 포인터
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 DBS_ERR_OK,\n
 *  로컬호스트에서 에러 발생 시 DBS_ERR_LOCAL,\n
 *  리모트호스트에서 에러 발생 시 DBS_ERR_PEER
 */
static int
run_backup(PXE *pxe, TAR *tar, libtar_list_t *list, 
	BackupHeader *hdr, void **msg)
{
    int pxe_options;
    void *packet = NULL;
    libtar_listptr_t lp;
    char *pathname = NULL;

    ASSERT(tar && list);

    libtar_listptr_reset(&lp);
    while (libtar_list_next(list, &lp) != 0)
    {
	/* 
	 * 자신한테 수신된 패킷이 있는지 확인하고 에러 패킷이면 에러 처리
	 */
	pxe_options = pxe_get_options(pxe);
	pxe_set_options(pxe, pxe_options | PXE_NON_BLOCK);
	if (pxe_read(pxe, &packet))
	{
	    if (is_backup_error(packet, msg))
	    {
		pxe_free(packet);
		/* 상대방 호스트에서 에러 발생 */
		return DBS_ERR_PEER;
	    }
	    pxe_free(packet);
	}
	pxe_set_options(pxe, pxe_options);

	if (IS_TIME_LIMIT(hdr->options))
	{
	    if (check_timeout(hdr->ltime))
	    {
		*msg = strdup("작업 취소!! 제한시간 만료되었음");
		return DBS_ERR_LOCAL;
	    }
	}

	pathname = libtar_listptr_data(&lp);
	if (tar_append_tree(tar, pathname, pathname) != 0)
	{
	    *msg = dup_str("%s 백업 실패: %s", pathname, strerror(errno));
	    return DBS_ERR_LOCAL;
	}
    }

    if (append_eof(tar, msg))
	return DBS_ERR_LOCAL;

    return DBS_ERR_OK;
}

/**
 * @brief Libtar를 사용해서 복구 실행
 * @param tar - Littar 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터 (호출자 사용 후 FREE)
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
run_restore(TAR *tar, REGNO regno, RestoreHeader *hdr, void **msg)
{
    char *datadir = NULL;
    char *prefix = NULL;

    ASSERT(tar && hdr &&msg);

    datadir = get_config("DataDir");
    if (!datadir)
    {
	*msg = dup_str("'DataDir' 파라메터가 존재하지 않음");
	return -1;
    }

    /* Overwrite 옵션이 켜져있으면 원본을 덮어쓰면서 복구 */
    prefix = IS_OVERWRITE(hdr->options) ? NULL : datadir;
    if (tar_extract_all(tar, prefix) != 0)
    {
	*msg = dup_str("%" PRIu64 "번 복구 실패: %s", regno, 
		strerror(errno));
	FREE(datadir);
	return -1;
    }
    FREE(datadir);
    return 0;
}

/**
 * @brief 백업 종료 패킷 전송
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 백업모듈 헤더 포인터
 * @return
 *  성공 시 0,\n
 *  실패 시  -1
 */
static int
send_backup_finish(PXE *pxe, REGNO regno, BackupHeader *hdr)
{
    size_t len;
    void *ptr = NULL;
    BackupHeader header;

    ASSERT(pxe && hdr);

    memcpy(&header, hdr, sizeof(BackupHeader));
    header.cmd = BACKUP_END;

    len = create_backup_packet(&ptr, regno, &header, NULL, 0);
    if (len == 0)
	return -1;

    if (pxe_write(pxe, ptr, len))
	return -1;

    FREE(ptr);

    return 0;
}

/**
 * @brief 로컬 백업파일 오픈
 * @param idx - TAR List Index
 * @param regno - 등록번호
 * @param options - 백업 옵션
 * @param msg - (OUT) 에러 발생 시 에러메세지 포인터
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
open_local_file(int idx, REGNO regno, int options, void **msg)
{
    char *datadir = NULL;
    char pathname[PATH_MAX];

    ASSERT(msg);

    datadir = get_config("DataDir");
    if (!datadir)
    {
	*msg = dup_str("'DataDir' 파라메터가 존재하지 않음");
	return -1;
    }

    /* 압축 옵션이 켜져 있어도 로컬복사본은 압축하지 않는 옵션이 같이 설정되어
     * 있다면 파일 확장자를 일반 tar 파일로 설정한다.
     */
    snprintf(pathname, sizeof(pathname), "%s/%" PRIu64 ".%s", datadir, regno, 
	    IS_COMPRESS(options) && !IS_LOCAL_NOCOMPRESS(options) ? GZ_EXT : TAR_EXT);
    FREE(datadir);

    tlist[idx].fd = open(pathname, 
	    O_RDWR | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
    if (tlist[idx].fd < 0)
    {
	*msg = dup_str("복사본 파일(%s) 생성 실패: %s", pathname, strerror(errno));
	return -1;
    }

    return 0;
}

/**
 * @brief 로컬 백업파일 삭제
 * @param regno -  등록번호
 * @param options - 백업 옵션
 * @return 없음
 */
static void
delete_local_file(REGNO regno, int options)
{
    char *datadir = NULL;
    char pathname[PATH_MAX];

    datadir = get_config("DataDir");
    if (!datadir)
    {
	Log(ERROR, "'DataDir' 파라메터가 존재하지 않음");
	return;
    }
    snprintf(pathname, sizeof(pathname), "%s/%" PRIu64 ".%s", datadir, regno, 
	    IS_COMPRESS(options) ? GZ_EXT : TAR_EXT);
    (void) unlink(pathname);
    FREE(datadir);
}

/**
 * @brief 백업용 Libtar 쓰레드
 * @param arg - PXE 쓰레드 로딩시 넘겨받는 파라메터
 * @return NULL 포인터
 */
void *
ltar_backup(void *arg)
{
    int idx;
    int status;
    int result;
    TAR *tar = NULL;
    PXE *pxe = NULL;
    void *err_msg = NULL;
    libtar_list_t *list = NULL;
    BackupInfo *info = NULL;

    /* For cleanup routine */
    idx = -1;
    status = -1;

    pxe = pxe_open(arg);
    if (!pxe)
    {
	Log(ERROR, "PXE 핸들러 생성 실패: %s", strerror(errno));
	goto cleanup;
    }

    info = (BackupInfo *) pxe_get_argument(arg);
    ASSERT(info);

    /* 마스터의 담당 쓰레드 아이디 설정 */
    pxe_set_destination(pxe, PXE_MASTER, info->src);
    
    idx = activate_tar_list(pxe, info);
    if (idx < 0) {
	ret_backup_error(pxe, info->regno, &info->header, 
		"TAR 리스트 활성화 실패: %s", strerror(errno));
	goto cleanup;
    }

    if (init_backup(&tar, &list, idx, info->data, &info->header)) {
	ret_backup_error(pxe, info->regno, &info->header, 
		"TAR 라이브러리 초기화 실패: %s", strerror(errno));
	goto cleanup;
    }

    if (IS_LOCAL_COPY(info->header.options)) {
	if (open_local_file(idx, info->regno, info->header.options, &err_msg)) {
	    ret_backup_error(pxe, info->regno, &info->header, "%s", err_msg);
	    goto cleanup;
	}
    }

    if (IS_COMPRESS(info->header.options)) {
	int result;
	if ((result = zlib_stream_deflate_init(&tlist[idx].strm)) != Z_OK) {
	    ret_backup_error(pxe, info->regno, &info->header, "zlib 초기화 실패: "
		    "%s", zError(result));
	    goto cleanup;
	}
    }

    result = run_backup(pxe, tar, list, &info->header, &err_msg);
    if (result != DBS_ERR_OK)
    {
	if (IS_COMPRESS(info->header.options))
	    (void) deflateEnd(&tlist[idx].strm);
	if (result == DBS_ERR_LOCAL)
	    ret_backup_error(pxe, info->regno, &info->header, "%s", err_msg);
	goto cleanup;
    }

    if (IS_COMPRESS(info->header.options))
	(void) deflateEnd(&tlist[idx].strm);

    /* 내부버퍼에 남아있는 데이터 전송 */
    flush_buffer(idx);
    Log(INFO, "%" PRIu64 "번 백업 종료", info->regno);
    send_backup_finish(pxe, info->regno, &info->header);

    /* Success */
    status = 0;

cleanup:
    if (tar)
	tar_close(tar);
    if (list)
	libtar_list_free(list, free);
    if (idx >= 0)
    {
	if (tlist[idx].fd >= 0)
	{
	    close(tlist[idx].fd);
	    if (status && IS_LOCAL_COPY(info->header.options))
		delete_local_file(info->regno, info->header.options);
	}
	inactivate_tar_list(idx);
    }
    if (pxe)
	pxe_close(pxe);
    FREE(err_msg);
    FREE(arg);

    return NULL;
}

/**
 * @brief 백업정보 구조체 생성
 * @param ptr - (OUT) 생성된 백업정보 구조체의 포인터 (호출자 사용 후 FRE)
 * @param regno - 등록번호
 * @param src - Peer 호스트의 쓰레드 인덱스
 * @param hdr - 백업모듈 헤더 포인터
 * @param buf - 데이터
 * @param len - \a buf 의 길이
 * @return 
 *  성공 시 생성된 백업정보 구조체의 길이,\n
 *  실패 시 0
 */
size_t
create_ltar_backup_info(void **ptr, REGNO regno, PXEThread src, 
	BackupHeader *hdr, const void *buf, size_t len)
{
    size_t total_size;
    BackupInfo *info = NULL;

    ASSERT(ptr && hdr && buf);

    total_size = sizeof(BackupInfo) + len;
    *ptr = calloc(1, total_size);
    if (!*ptr)
	return 0;

    info = (BackupInfo *) *ptr;
    info->regno = regno;
    info->src = src;
    memcpy(&info->header, hdr, sizeof(BackupHeader));
    info->len = len;
    memcpy(info->data, buf, len);

    return total_size;
}

/**
 * @brief 마스터의 BK_THREAD에 특정 복구명령 전송
 * @param pxe - PXE 핸들러 포인터
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param cmd - 전송할 명령어
 * @param data - 사용자 데이터
 * @return
 *  성공 시 0,\n
 *  실패 시 -1
 */
static int
send_cmd_to_bk(PXE *pxe, REGNO regno, 
	RestoreHeader *hdr, int cmd, void *data)
{
    int ret_val;
    size_t size;
    void *ptr = NULL;
    RestoreHeader header;

    ASSERT(pxe && hdr);

    /* 기본 리턴값 */
    ret_val = -1;

    /* 
     * 복구모듈 헤더 복사 후 명령어 변경
     */
    memcpy(&header, hdr, sizeof(RestoreHeader));
    header.cmd = cmd;

    size = create_restore_packet(&ptr, regno, 
	    &header, data, hdr->len);
    if (size == 0)
	goto cleanup;
    
    pxe_set_destination(pxe, PXE_MASTER, BK_THREAD);
    if (pxe_write(pxe, ptr, size))
	goto cleanup;

    ret_val = 0;
    
cleanup:
    FREE(ptr);

    return ret_val;
}

/**
 * @brief 복구용 Libtar 쓰레드
 * @param arg - PXE 쓰레드 로딩시 넘겨받는 파라메터
 * @return NULL 포인터
 */
void *
ltar_restore(void *arg)
{
    int idx;
    PXE *pxe = NULL;
    TAR *tar = NULL;
    void *err_msg = NULL;
    RestoreInfo *info = NULL;

    /* for cleanup routine */
    idx = -1;

    pxe = pxe_open(arg);
    if (!pxe)
    {
	Log(ERROR, "PXE 핸들러 생성 실패: %s", strerror(errno));
	goto cleanup;
    }

    info = (RestoreInfo *) pxe_get_argument(arg);
    ASSERT(info);

    idx = activate_tar_list(pxe, info);
    if (idx < 0)
    {
	ret_restore_error(pxe, info->regno, &info->header, "TAR 리스트 초기화 "
		"실패: %s", strerror(errno));
	goto cleanup;
    }

    if (init_restore(&tar, idx, info->data, &info->header))
    {
	ret_restore_error(pxe, info->regno, &info->header, "TAR 라이브러리 초기화 "
		"실패: %s", strerror(errno));
	goto cleanup;
    }

    if (send_cmd_to_bk(pxe, info->regno, &info->header, RESTORE_START, info->data))
    {
	ret_restore_error(pxe, info->regno, &info->header, "복구 시작 명령을 전송 "
		"할 수 없음: %s", strerror(errno));
	goto cleanup;
    }

    if (IS_COMPRESS(info->header.options))
    {
	int result;
	if ((result = zlib_stream_inflate_init(&tlist[idx].strm)) != Z_OK)
	{
	    ret_restore_error(pxe, info->regno, &info->header, "zlib 초기화 실패: "
		    "%s", zError(result));
	    goto cleanup;
	}
    }

    if (run_restore(tar, info->regno, &info->header, &err_msg))
    {
	if (IS_COMPRESS(info->header.options))
	    (void) inflateEnd(&tlist[idx].strm);

	/* 상대방 호스트에서 에러가 리턴되면 ENODATA를 errno로 셋팅한다.
	 * errno를 확인해서 로컬호스트에서 발생한 에러만 상대방에게 통보한다.
	 */
	if (errno != ENODATA)
	    ret_restore_error(pxe, info->regno, &info->header, "%s", err_msg);
	goto cleanup;
    }
    if (IS_COMPRESS(info->header.options))
	(void) inflateEnd(&tlist[idx].strm);

    if (send_cmd_to_bk(pxe, info->regno, &info->header, RESTORE_END, info->data))
    {
	ret_restore_error(pxe, info->regno, &info->header, "복구 종료 명령을 전송 "
		"할 수 없음: %s", strerror(errno));
	goto cleanup;
    }

cleanup:
    if (tar)
	tar_close(tar);
    if (idx >= 0)
	inactivate_tar_list(idx);
    if (pxe)
	pxe_close(pxe);
    FREE(arg);

    return NULL;
}

/**
 * @brief 복구정보 구조체 생성
 * @param ptr - (OUT) 생성된 복구정보 구조체의 포인터 (호출자 사용 후 FRE)
 * @param regno - 등록번호
 * @param hdr - 복구모듈 헤더 포인터
 * @param buf - 데이터
 * @param len - \a buf 의 길이
 * @return 
 *  성공 시 생성된 복구정보 구조체의 길이,\n
 *  실패 시 0
 */
size_t
create_ltar_restore_info(void **ptr, REGNO regno, 
	RestoreHeader *hdr, const void *buf, size_t len)
{
    size_t total_size;
    RestoreInfo *info = NULL;

    ASSERT(ptr && hdr && buf);

    total_size = sizeof(RestoreInfo) + len;
    *ptr = calloc(1, total_size);
    if (!*ptr)
	return 0;

    info = (RestoreInfo *) *ptr;
    info->regno = regno;
    memcpy(&info->header, hdr, sizeof(RestoreHeader));
    info->len = len;
    memcpy(info->data, buf, len);

    return total_size;
}

/*
 * 아래 함수들은 MySQL 라이브러리가 링크된 경우 마스터로 동작하기 위해 포함되는 
 * 함수들로서 MySQL 라이브러리가 링크되지 않은 경우에는 컴파일에 포함되지 않는다.
 */
#ifdef HAVE_LIBMYSQLCLIENT
/**
 * @brief GNU Longname Extension 처리 
 * @param tar - (IN/OUT) TAR 헤더 포인터
 * @param filename - (OUT) 실제 파일 이름 포인터
 * @return 
 *  Longname Extension 관련 데이터를 건너뛰어서 실제 파일 내용을 
 *  포인팅하기 위해 skip한 TAR BLOCK 수
 *
 * GNU Longname Extension을 사용할 경우, 실제 파일이름과 파일에 대한 
 * 헤더를 찾아서 포인팅 해준다.
 */
static size_t
process_longname_extension(struct tar_header **tar, char **filename)
{
    int blk;
    int size;

    ASSERT(tar && filename);

    /* LONGNAME 헤더 자체에 있는 DBS 전용 매직스트링을 다시 USTAR 매직 
     * 스트링으로 복구시켜서 정상적인 TAR 헤더로 만든다.
     */
    memcpy((*tar)->magic, USTAR_MAGIC, strlen(USTAR_MAGIC));

    /* Longname Extension이 사용하는 블럭 수 측정 */
    size = oct_to_int((*tar)->size);
    blk = (size / T_BLOCKSIZE) + (size % T_BLOCKSIZE ? 1 : 0);

    if ((*tar)->typeflag == GNU_LONGNAME_TYPE)
	*filename = ((char *) *tar) + T_BLOCKSIZE;
    else if ((*tar)->typeflag == GNU_LONGLINK_TYPE)
	*filename = (*tar)->name;
    else
	*filename = NULL;

    /* Longname Extension 블럭 뒤에 실제 파일에 대한 정보를 가지고있는
     * 헤더 부분으로 포인터를 이동
     */
    *tar = (struct tar_header *) ((char *) *tar + T_BLOCKSIZE * (blk + 1));

    /* 실제 파일에 대한 정보를 가지고 있는 헤더가 있기 때문에
     * +1을 해준다.
     */
    return blk + 1;
}

/**
 * @brief TAR 헤더 기록
 * @param sql - SQL 핸들러 포인터
 * @param regno - 등록번호
 * @param packet - 수신한 패킷
 * @param len - \a packet 의 길이
 * @param file_offset - 현재 백업 파일의 오프셋
 * @return 없음
 */
void
write_tar_header(sql_t *sql, REGNO regno, char *packet, PXELength len, 
	off_t file_offset)
{
    size_t blk;
    PXELength offset;
    struct tar_header *tar = NULL;
    char *filename = NULL;
    off_t archive_offset;

    ASSERT(sql && packet);

    /* 성능을 위해 8KB 단위로 버퍼링된 데이터블럭이 수신되기
     * 때문에 TAR 블럭사이즈 단위로 점프하면서 헤더를 분석한다
     */
    for (offset = 0; offset < len; offset += T_BLOCKSIZE)
    {
	tar = (struct tar_header *) (packet + offset);
	if (is_tar_header(tar))
	    continue;

	/* TAR 아카이브에 해당 파일이 위치해있는 offset 계산 */
	archive_offset = file_offset + offset;

	/* GNU Longname/Longlink Extension을 사용하는 경우 실제 파일네임 또는 
	 * 링크의 대상 파일 이름 및 파일 정보를 헤더의 뒷 부분에서 추출한다.
	 */
	if (tar->typeflag != GNU_LONGNAME_TYPE && 
		tar->typeflag != GNU_LONGLINK_TYPE)
	    filename = tar->name;
	else
	{
	    /* Longname/Longlink 헤더 뒤에 실제 파일의 헤더가 나오지 않고 다시
	     * Longname/Longlink가 나올 수 있다. 이런 경우 실제 파일의 헤더까지
	     * 매직스트링을 바꾸면서 점프한다.
	     */
	    while (offset < len && 
		    (tar->typeflag == GNU_LONGNAME_TYPE || 
		     tar->typeflag == GNU_LONGLINK_TYPE))
	    {
		blk = process_longname_extension(&tar, &filename);
		offset += blk * T_BLOCKSIZE;
	    }
	}

	/* DBS 전용 매직스트링을 다시 USTAR 매직 스트링으로 복구시켜서 정상적인 
	 * TAR 헤더로 만든다.
	 */
	memcpy(tar->magic, USTAR_MAGIC, strlen(USTAR_MAGIC));

	insert_backup_record(sql, regno, filename, oct_to_int(tar->size), 
		strtoll(tar->mtime, (char **) NULL, 8), tar->typeflag, 
		archive_offset);
    }
}
/*
 * 여기까지가 MySQL 라이브러리가 링크된 경우에만 컴파일 되는 함수들
 */
#endif
