/*
**  Copyright 1998-2003 University of Illinois Board of Trustees
**  Copyright 1998-2003 Mark D. Roth
**  All rights reserved.
**
**  decode.c - libtar code to decode tar header blocks
**
**  Mark D. Roth <roth@uiuc.edu>
**  Campus Information Technologies and Educational Services
**  University of Illinois at Urbana-Champaign
*/

#include <internal.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <pwd.h>
#include <grp.h>
#include <iconv.h>

#ifdef STDC_HEADERS
# include <string.h>
#endif

/**
 * decode_str_from_utf8 - 
 *  UTF-8에서 시스템 인코딩으로 문자열 변환 (동적 메모리 생성)
 *
 * @parm str: 변환할 UTF-8 문자열
 * @parm ptr(OUT): 변환된 문자열이 저장된 메모리 포인터 (호출자 사용 후 FREE)
 * @parm charset: 시스템 캐릭터셋
 * @return: 성공 시 0, 실패 시 -1
 *
 * 설정파일에 저장된 인코딩이 UTF-8이라면 변환 없이 새로 할당된 동적 메모리에
 * 문자열을 복사하고 성공을 리턴한다.
 */
static int
decode_str_from_utf8(const char *str, char **ptr, char *charset)
{
    iconv_t cd;
    size_t in_len;
    size_t out_len;
    char *out_str = NULL;

    /* 시스템 인코딩이 UTF-8이면 변환없이 그대로 문자열 복사 */
    if (strcasecmp(charset, "UTF-8") == 0)
    {
	*ptr = strdup(str);
	if (!*ptr)
	    return -1;
	
	return 0;
    }

    /* 
     * 시스템 인코딩으로 변환시 필요한 메모리 상한선 
     */
    in_len = strlen(str);
    out_len = in_len + 1;

    *ptr = calloc(1, out_len);
    if (!*ptr)
	return -1;

    /* iconv() 실행 후 결과 포인터의 위치가 변하기 때문에 미리 저장 */
    out_str = *ptr;

    /* UTF-8에서 설정파일에 지정된 인코딩으로 변환 */
    cd = iconv_open(charset, "UTF-8");
    if (cd == (iconv_t) -1)
    {
	if (*ptr)
	{
	    free(*ptr);
	    *ptr = NULL;
	}
	return -1;
    }

    if (iconv(cd, (char **) &str, &in_len, 
		(char **) &out_str, &out_len) == (size_t) -1)
    {
	if (*ptr)
	{
	    free(*ptr);
	    *ptr = NULL;
	}
	return -1;
    }

    iconv_close(cd);
    return 0;
}

/* determine full path name */
char *
th_get_pathname(TAR *t)
{
	char *decoded_str = NULL;
	char *ret_str = NULL;
	char filename[MAXPATHLEN];

	if (t->th_buf.gnu_longname)
	{
		(void) decode_str_from_utf8(t->th_buf.gnu_longname, 
			&decoded_str, t->charset);
		ret_str = strdup(decoded_str);
		if (decoded_str)
		    free(decoded_str);
		return ret_str;
	}

	if (t->th_buf.prefix[0] != '\0')
	{
		snprintf(filename, sizeof(filename), "%.155s/%.100s",
			 t->th_buf.prefix, t->th_buf.name);
		(void) decode_str_from_utf8(filename, &decoded_str, t->charset);
		ret_str = strdup(decoded_str);
		if (decoded_str)
		    free(decoded_str);
		return ret_str;
	}

	snprintf(filename, sizeof(filename), "%.100s", t->th_buf.name);
	(void) decode_str_from_utf8(filename, &decoded_str, t->charset);
	ret_str = strdup(decoded_str);
	if (decoded_str)
	    free(decoded_str);
	return ret_str;
}


uid_t
th_get_uid(TAR *t)
{
	int uid;
	struct passwd *pw;

	pw = getpwnam(t->th_buf.uname);
	if (pw != NULL)
		return pw->pw_uid;

	/* if the password entry doesn't exist */
	sscanf(t->th_buf.uid, "%o", &uid);
	return uid;
}


gid_t
th_get_gid(TAR *t)
{
	int gid;
	struct group *gr;

	gr = getgrnam(t->th_buf.gname);
	if (gr != NULL)
		return gr->gr_gid;

	/* if the group entry doesn't exist */
	sscanf(t->th_buf.gid, "%o", &gid);
	return gid;
}


mode_t
th_get_mode(TAR *t)
{
	mode_t mode;

	mode = (mode_t)oct_to_int(t->th_buf.mode);
	if (! (mode & S_IFMT))
	{
		switch (t->th_buf.typeflag)
		{
		case SYMTYPE:
			mode |= S_IFLNK;
			break;
		case CHRTYPE:
			mode |= S_IFCHR;
			break;
		case BLKTYPE:
			mode |= S_IFBLK;
			break;
		case DIRTYPE:
			mode |= S_IFDIR;
			break;
		case FIFOTYPE:
			mode |= S_IFIFO;
			break;
		case AREGTYPE:
			if (t->th_buf.name[strlen(t->th_buf.name) - 1] == '/')
			{
				mode |= S_IFDIR;
				break;
			}
			/* FALLTHROUGH */
		case LNKTYPE:
		case REGTYPE:
		default:
			mode |= S_IFREG;
		}
	}

	/* Cygwin에서 디렉토리의 owner 퍼미션에 write 권한이 빠져있을 경우 강제로
	 * write 퍼미션을 추가한다. 복구 작업의 마지막에서 ACLs를 복구하면 강제로
	 * 변경한 퍼미션이 원래대로 조정된다.
	 */
	if ((t->options & TAR_FORCE_WRITE_PERM) && 
		t->th_buf.typeflag == '5')
	{
	    if ((mode & 0x80) == 0)
		mode |= 0x80;
	}

	/* TAR 헤더에 8바이트 사이즈의 mode가 저장되는데 0100644 같이 앞에 쓸모없는
	 * mode 비트가 같이 저장된다. 따라서 0xFFF와 AND 연산을 해서 뒤에 0644 부분만
	 * 때어내서 mode로 사용한다. 참고로 GNU TAR 커맨드로 생성된 TAR 아카이브에는
	 * 0000644와 같이 앞에 쓸모없는 mode 비트가 추가되지 않는다.
	 */
	return mode & 0xFFF;
}


