/*
**  Copyright 1998-2003 University of Illinois Board of Trustees
**  Copyright 1998-2003 Mark D. Roth
**  All rights reserved.
**
**  encode.c - libtar code to encode tar header blocks
**
**  Mark D. Roth <roth@uiuc.edu>
**  Campus Information Technologies and Educational Services
**  University of Illinois at Urbana-Champaign
*/

#include <internal.h>

#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <iconv.h>

#ifdef STDC_HEADERS
# include <string.h>
# include <stdlib.h>
#endif

#define DBS_MAGIC    "DBSHD"

/* magic, version, and checksum */
void
th_finish(TAR *t)
{
	int i, sum = 0;

	if (t->options & TAR_GNU)
		strncpy(t->th_buf.magic, "ustar  ", 8);
	else
	{
		strncpy(t->th_buf.version, TVERSION, TVERSLEN);
		strncpy(t->th_buf.magic, TMAGIC, TMAGLEN);
	}

	for (i = 0; i < T_BLOCKSIZE; i++)
		sum += ((unsigned char *)(&(t->th_buf)))[i];
	for (i = 0; i < 8; i++)
		sum += (' ' - (unsigned char)t->th_buf.chksum[i]);
	int_to_oct(sum, t->th_buf.chksum, 8);

	/* using DBS MAGIC STRING, instead of USTAR MAGIC STRING */
	memcpy(t->th_buf.magic, DBS_MAGIC, strlen(DBS_MAGIC));
}


/* map a file mode to a typeflag */
void
th_set_type(TAR *t, mode_t mode)
{
	if (S_ISLNK(mode))
		t->th_buf.typeflag = SYMTYPE;
	if (S_ISREG(mode))
		t->th_buf.typeflag = REGTYPE;
	if (S_ISDIR(mode))
		t->th_buf.typeflag = DIRTYPE;
	if (S_ISCHR(mode))
		t->th_buf.typeflag = CHRTYPE;
	if (S_ISBLK(mode))
		t->th_buf.typeflag = BLKTYPE;
	if (S_ISFIFO(mode) || S_ISSOCK(mode))
		t->th_buf.typeflag = FIFOTYPE;
}

/**
 * encode_str_to_utf8 - 
 *  시스템 인코딩에서 UTF-8로 문자열 변환 (동적 메모리 생성)
 *
 * @parm str: 변환할 문자열
 * @parm ptr(OUT): UTF-8로 문자열이 저장된 메모리 포인터 (호출자 사용 후 FREE)
 * @parm charset: 시스템 캐릭터셋
 * @return: 성공 시 0, 실패 시 -1
 *
 * 설정파일에 저장된 인코딩이 UTF-8이라면 변환 없이 새로 할당된 동적 메모리에
 * 문자열을 복사하고 성공을 리턴한다.
 */
static int
encode_str_to_utf8(const char *str, char **ptr, const char *charset)
{
    iconv_t cd;
    size_t in_len;
    size_t out_len;
    char *out_str = NULL;

    /* 시스템 인코딩이 UTF-8이면 변환없이 그대로 문자열 복사 */
    if (strcasecmp(charset, "UTF-8") == 0)
    {
	*ptr = strdup(str);
	if (!*ptr)
	    return -1;
	
	return 0;
    }

    /* 
     * UTF-8로 변환시 필요한 메모리 상한선 
     */
    in_len = strlen(str);
    out_len = in_len * 2 + 1;

    *ptr = calloc(1, out_len);
    if (!*ptr)
	return -1;

    /* iconv() 실행 후 결과 포인터의 위치가 변하기 때문에 미리 저장 */
    out_str = *ptr;

    /* 설정파일에 지정된 인코딩에서 UTF-8로 변환 */
    cd = iconv_open("UTF-8", charset);
    if (cd == (iconv_t) -1)
    {
	if (*ptr)
	{
	    free(*ptr);
	    *ptr = NULL;
	}
	return -1;
    }

    if (iconv(cd, (char **) &str, &in_len, 
		(char **) &out_str, &out_len) == (size_t) -1)
    {
	if (*ptr)
	{
	    free(*ptr);
	    *ptr = NULL;
	}
	return -1;
    }

    iconv_close(cd);
    return 0;
}

/* encode file path */
void
th_set_path(TAR *t, char *pathname)
{
	char suffix[2] = "";
	char *tmp;
	char *utf8_str = NULL;
	char buf[T_NAMELEN + 1];

#ifdef DEBUG
	printf("in th_set_path(th, pathname=\"%s\")\n", pathname);
#endif

	free(t->th_buf.gnu_longname);
	t->th_buf.gnu_longname = NULL;

	/* TAR 헤더의 파일네임은 시스템 인코딩에서 UTF-8로 변경해서 기록 */
	if (encode_str_to_utf8(pathname, &utf8_str, t->charset))
	    /* 에러 발생 시 원본 path를 그대로 사용 */
	    utf8_str = strdup(pathname);

	if (pathname[strlen(utf8_str) - 1] != '/' && TH_ISDIR(t))
		strcpy(suffix, "/");

	if (strlen(utf8_str) > T_NAMELEN && (t->options & TAR_GNU))
	{
		/* GNU-style long name */
		t->th_buf.gnu_longname = strdup(utf8_str);
		strncpy(t->th_buf.name, t->th_buf.gnu_longname, T_NAMELEN);
	}
	else if (strlen(utf8_str) > T_NAMELEN)
	{
		/* POSIX-style prefix field */
		tmp = strchr(&(utf8_str[strlen(utf8_str) - T_NAMELEN - 1]), '/');
		if (tmp == NULL)
		{
			printf("!!! '/' not found in \"%s\"\n", utf8_str);
			if (utf8_str)
			    free(utf8_str);
			return;
		}
		snprintf(t->th_buf.name, T_NAMELEN, "%s%s", &(tmp[1]), suffix);
		snprintf(t->th_buf.prefix,
			 ((tmp - utf8_str + 1) <
			  155 ? (tmp - utf8_str + 1) : 155), "%s", utf8_str);
	}
	else
	{
		/* classic tar format 
		 * 
		 * 원래는 t->th_buf.name에 snprintf() 직접 파일명을 기록했으나
		 * 파일명의 길이가 정확히 T_NAMELEN인 경우 마지막 한글자가 잘리는
		 * 버그가 발견되었다. 그래서 임시 버퍼에 파일명을 생성한 후에 
		 * 정확히 T_NAMELEN 만큼 복사하도록 수정했다.
		 */
		snprintf(buf, sizeof(buf), "%s%s", utf8_str, suffix);
		memcpy(t->th_buf.name, buf, T_NAMELEN);
	}

	if (utf8_str)
	    free(utf8_str);

#ifdef DEBUG
	puts("returning from th_set_path()...");
#endif
}


/* encode link path */
void
th_set_link(TAR *t, char *linkname)
{
	char *utf8_str = NULL;

#ifdef DEBUG
	printf("==> th_set_link(th, linkname=\"%s\")\n", linkname);
#endif

	/* TAR 헤더의 파일네임은 시스템 인코딩에서 UTF-8로 변경해서 기록 */
	if (encode_str_to_utf8(linkname, &utf8_str, t->charset))
	    /* 에러 발생 시 원본 path를 그대로 사용 */
	    utf8_str = strdup(linkname);

	if (strlen(utf8_str) > T_NAMELEN && (t->options & TAR_GNU))
	{
		/* GNU longlink format */
		t->th_buf.gnu_longlink = strdup(utf8_str);
		strcpy(t->th_buf.linkname, "././@LongLink");
	}
	else
	{
		/* classic tar format */
		strlcpy(t->th_buf.linkname, utf8_str, sizeof(t->th_buf.linkname));
		free(t->th_buf.gnu_longlink);
		t->th_buf.gnu_longlink = NULL;
	}

	if (utf8_str)
	    free(utf8_str);
}


/* encode device info */
void
th_set_device(TAR *t, dev_t device)
{
#ifdef DEBUG
	printf("th_set_device(): major = %d, minor = %d\n",
	       major(device), minor(device));
#endif
	int_to_oct(major(device), t->th_buf.devmajor, 8);
	int_to_oct(minor(device), t->th_buf.devminor, 8);
}


/* encode user info */
void
th_set_user(TAR *t, uid_t uid)
{
	struct passwd *pw;

	pw = getpwuid(uid);
	if (pw != NULL)
		strlcpy(t->th_buf.uname, pw->pw_name, sizeof(t->th_buf.uname));

	int_to_oct(uid, t->th_buf.uid, 8);
}


/* encode group info */
void
th_set_group(TAR *t, gid_t gid)
{
	struct group *gr;

	gr = getgrgid(gid);
	if (gr != NULL)
		strlcpy(t->th_buf.gname, gr->gr_name, sizeof(t->th_buf.gname));

	int_to_oct(gid, t->th_buf.gid, 8);
}


/* encode file mode */
void
th_set_mode(TAR *t, mode_t fmode)
{
	if (S_ISSOCK(fmode))
	{
		fmode &= ~S_IFSOCK;
		fmode |= S_IFIFO;
	}
	int_to_oct(fmode, (t)->th_buf.mode, 8);
}


void
th_set_from_stat(TAR *t, struct stat *s)
{
	th_set_type(t, s->st_mode);
	if (S_ISCHR(s->st_mode) || S_ISBLK(s->st_mode))
		th_set_device(t, s->st_rdev);
	th_set_user(t, s->st_uid);
	th_set_group(t, s->st_gid);
	th_set_mode(t, s->st_mode);
	th_set_mtime(t, s->st_mtime);
	if (S_ISREG(s->st_mode))
		th_set_size(t, s->st_size);
	else
		th_set_size(t, (off_t) 0);
}


