/*
**  Copyright 1998-2003 University of Illinois Board of Trustees
**  Copyright 1998-2003 Mark D. Roth
**  All rights reserved.
**
**  util.c - miscellaneous utility code for libtar
**
**  Mark D. Roth <roth@uiuc.edu>
**  Campus Information Technologies and Educational Services
**  University of Illinois at Urbana-Champaign
*/

#include <internal.h>

#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <errno.h>
#include <inttypes.h>

#ifdef STDC_HEADERS
# include <string.h>
#endif


/* hashing function for pathnames */
int
path_hashfunc(char *key, int numbuckets)
{
	char buf[MAXPATHLEN];
	char *p;

	strcpy(buf, key);
	p = basename(buf);

	return (((unsigned int)p[0]) % numbuckets);
}


/* matching function for dev_t's */
int
dev_match(dev_t *dev1, dev_t *dev2)
{
	return !memcmp(dev1, dev2, sizeof(dev_t));
}


/* matching function for ino_t's */
int
ino_match(ino_t *ino1, ino_t *ino2)
{
	return !memcmp(ino1, ino2, sizeof(ino_t));
}


/* hashing function for dev_t's */
int
dev_hash(dev_t *dev)
{
	return *dev % 16;
}


/* hashing function for ino_t's */
int
ino_hash(ino_t *inode)
{
	return *inode % 256;
}

/* 문자열안에 어떤 문자가 중복되서 나올경우 한개로 줄여주는 코드 */
int
trim_duplicated(char *string, char c)
{
   int i, cnt = 0;
   char *ptr = NULL;

   if((ptr = malloc((strlen(string) + 1) * sizeof(char))) == NULL) {
      return -1;
   }
   for(i = 0; i < strlen(string); i++) {
      if(string[i] == c) {
         if(i > 0) {
            if(string[i-1] == c) {
               cnt++;       
               continue;
            }
         }
      }
      
      ptr[i - cnt] = string[i];      
   }   
   ptr[i - cnt] = '\0';
   
   if(cnt > 0) {
      strcpy(string, ptr);
   }
   if(ptr) free(ptr);

   return cnt;
}

/*
** mkdirhier() - create all directories in a given path
** returns:
**	0			success
**	1			all directories already exist
**	-1 (and sets errno)	error
*/
int
mkdirhier(char *path)
{
	char src[MAXPATHLEN], dst[MAXPATHLEN] = "";
	char *dirp, *nextp = src;
	int retval = 1;
	struct stat st;	

	if (strlcpy(src, path, sizeof(src)) > sizeof(src))
	{
		errno = ENAMETOOLONG;
		return -1;
	}

	if (path[0] == '/')
		strcpy(dst, "/");

	while ((dirp = strsep(&nextp, "/")) != NULL)
	{
		if (*dirp == '\0')
			continue;

		if (dst[0] != '\0')
			strcat(dst, "/");
		strcat(dst, dirp);

		if (trim_duplicated(dst, '/') < 0)
		    return -1;
		if (trim_duplicated(dst, '\\') < 0)
		    return -1;

		/* 디렉토리가 존재하는지 확인 */
		if (lstat(dst, &st))
		{
		    if (errno != ENOENT)
			return -1;

		    if (mkdir(dst, 0777) == -1)
		    {
			    if (errno != EEXIST)
				    return -1;
		    }
		    else
			    retval = 0;
		}
		else
		    retval = 0;
	}

	return retval;
}


/* calculate header checksum */
int
th_crc_calc(TAR *t)
{
	int i, sum = 0;

	for (i = 0; i < T_BLOCKSIZE; i++)
		sum += ((unsigned char *)(&(t->th_buf)))[i];
	for (i = 0; i < 8; i++)
		sum += (' ' - (unsigned char)t->th_buf.chksum[i]);

	return sum;
}


/* string-octal to integer conversion */
int
oct_to_int(char *oct)
{
	int i;

	sscanf(oct, "%o", &i);

	return i;
}

static int
is_little_endian(void) 
{
   union {
        short a;
        char  b[2];
   } endian;
  
   endian.a = 0x0102;
   if(endian.b[0] == 0x02)
       return 1; /* little endian */

   return 0;
}

static void
change_endian(void *var, size_t size)
{
    size_t head, tail;
    unsigned char swap;
    unsigned char *num = var;

    for (head = 0, tail = size - 1; head < size / 2; head++, tail--)
    {
        swap = num[tail];
        num[tail] = num[head];
        num[head] = swap;
    }
}

static off_t
get_size_binary(char *oct)
{
#define OCT_LEN	12
    off_t size;
    off_t *size_ptr = NULL;

    size_ptr = (off_t *) (oct + OCT_LEN - sizeof(off_t));
    size = *size_ptr;
    if (is_little_endian())
	change_endian(&size, sizeof(off_t));
    
    return size;
}

off_t
oct_to_offt(char *oct)
{
	off_t i;

	if (((unsigned char *) oct)[0] != 0x80)
		/* GNU 확장이 아닌 경우 */
		sscanf(oct, "%llo", &i);
	else
		i = get_size_binary(oct);

	return i;
}

/* integer to string-octal conversion, no NULL */
void
int_to_oct_nonull(int num, char *oct, size_t octlen)
{
	snprintf(oct, octlen, "%*lo", (int) octlen - 1, (unsigned long)num);
	oct[octlen - 1] = ' ';
}

static void
set_size_binary(off_t num, char *oct, size_t octlen)
{
    off_t *size_ptr = NULL;

    memset(oct, 0, octlen);
    size_ptr = (off_t *) (oct + octlen - sizeof(off_t));
    if (is_little_endian())
	change_endian(&num, sizeof(off_t));
    *size_ptr = num;
    oct[0] = 0x80;
}

void
offt_to_oct_nonull(off_t num, char *oct, size_t octlen)
{
#define MAXOCT11  017777777777L
	if (num > MAXOCT11)
	    /* 파일 사이즈가 OCT 11자리로 표현이 불가능할때 GNU 확장 사용 */
	    set_size_binary(num, oct, octlen);
	else
	{
	    snprintf(oct, octlen, "%*llo", (int) octlen - 1, 
		    (unsigned long long) num);
	    oct[octlen - 1] = ' ';
	}
}
