/*
**  Copyright 1998-2003 University of Illinois Board of Trustees
**  Copyright 1998-2003 Mark D. Roth
**  All rights reserved.
**
**  wrapper.c - libtar high-level wrapper code
**
**  Mark D. Roth <roth@uiuc.edu>
**  Campus Information Technologies and Educational Services
**  University of Illinois at Urbana-Champaign
*/

#include <internal.h>

#include <stdio.h>
#include <sys/param.h>
#include <dirent.h>
#include <stdlib.h>
#include <errno.h>

#ifdef STDC_HEADERS
# include <string.h>
#endif


/**
 * 파일 복구 시 패턴 매칭되는 파일만 복구하려고 할때 사용하려고 
 * 수정하였으나 적용하기가 복잡해서 사용하지 않기로 했음.
 * DBS에서 직접 TAR 파일을 핸들링해서 패킷을 전송하는것이 더 쉬울
 * 것으로 판단됨.
 */
int
tar_extract_glob(TAR *t, char **globname, char *prefix)
{
    return 0;
}


int
tar_extract_all(TAR *t, char *prefix)
{
	char *filename = NULL;
	char buf[MAXPATHLEN];
	int i;

#ifdef DEBUG
	printf("==> tar_extract_all(TAR *t, \"%s\")\n",
	       (prefix ? prefix : "(null)"));
#endif

	while ((i = th_read(t)) == 0)
	{
#ifdef DEBUG
		puts("    tar_extract_all(): calling th_get_pathname()");
#endif
		filename = th_get_pathname(t);
		if (t->options & TAR_VERBOSE)
			th_print_long_ls(t);
		if (prefix != NULL)
		{
		    /* Cygwin에서 C:\data\C:\filename 처럼 드라이브명이 중복되지
		     * 않도록 하기위해 아래 체크 루틴을 사용한다.
		     */
		    snprintf(buf, sizeof(buf), "%s/%s", prefix, 
			    filename[1] == ':' && (filename[2] == '/' || filename[2] == '\\') 
			    ? &filename[3] 
			    : filename);
		}
		else
		{
			strlcpy(buf, filename, sizeof(buf));
		}

		if (filename != NULL)
		{
		    free(filename);
		    filename = NULL;
		}

		if (trim_duplicated(buf, '/') < 0)
		    return -1;
		if (trim_duplicated(buf, '\\') < 0)
		    return -1;
#ifdef DEBUG
		printf("    tar_extract_all(): calling tar_extract_file(t, "
		       "\"%s\")\n", buf);
#endif
		if (tar_extract_file(t, buf, prefix) != 0)
			return -1;
	}

	return (i == 1 ? 0 : -1);
}

 	
int
tar_append_tree(TAR *t, char *realdir, char *savedir)
{
	int ret_val;
	char realpath[MAXPATHLEN];
	char savepath[MAXPATHLEN];
	struct dirent *dent;
	DIR *dp = NULL;
	struct stat s;

#ifdef DEBUG
	printf("==> tar_append_tree(0x%lx, \"%s\", \"%s\")\n",
	       t, realdir, (savedir ? savedir : "[NULL]"));
#endif

	ret_val = -1;

	if (tar_append_file(t, realdir, savedir) != 0)
	    goto cleanup;

#ifdef DEBUG
	puts("    tar_append_tree(): done with tar_append_file()...");
#endif

	dp = opendir(realdir);
	if (dp == NULL)
	{
	    if (errno == ENOTDIR)
		return 0;
	    return -1;
	}
	while ((dent = readdir(dp)) != NULL)
	{
		if (strcmp(dent->d_name, ".") == 0 ||
		    strcmp(dent->d_name, "..") == 0)
			continue;

		/*
		 * Cygwin에서는 절대경로가 "//" 이렇게 시작하는 경우 네트웍 
		 * 주소로 인식하는 문제가 발견되었다. 그래서 중복되는 /를 제거한다.
		 */
		snprintf(realpath, MAXPATHLEN, "%s/%s", realdir,
			 dent->d_name);
		if (trim_duplicated(realpath, '/') < 0)
		    goto cleanup;
		if (trim_duplicated(realpath, '\\') < 0)
		    goto cleanup;
		if (savedir)
		{
			snprintf(savepath, MAXPATHLEN, "%s/%s", savedir,
				 dent->d_name);
			if (trim_duplicated(savepath, '/') < 0)
			    goto cleanup;
			if (trim_duplicated(savepath, '\\') < 0)
			    goto cleanup;
		}

		if (lstat(realpath, &s) != 0)
		    goto cleanup;

		if (S_ISDIR(s.st_mode))
		{
			if (tar_append_tree(t, realpath,
					    (savedir ? savepath : NULL)) != 0)
			    goto cleanup;
			continue;
		}

		if (tar_append_file(t, realpath,
				    (savedir ? savepath : NULL)) != 0)
		    goto cleanup;
	}

	ret_val = 0;

cleanup:
	if (dp)
	    closedir(dp);

	return ret_val;
}


